/**
 * \file fixflo.c Trancodage virgule fixe en virgule flottante.
 *
 * fixflo transcode une valeur en codage  virgule fixe en codage  virgule
 * flottante. Ce programme gnralise le programme carflo aux codages fixes 
 * quelconques.
 *
 * Usage : fixflo [options de codage] -- valeur [valeur ...]
 *         options de format : voir le manuel de create pour une liste 
 *                             exhaustive de codage (-f,-r,-s, ...)
 *         -- : ce champs sert de sparateur entre les options de codage
 *              et les valeurs  transcoder. Il est obligatoire : s'il n'est
 *              pas trouv, le programme ne fait rien.
 *
 * Exemples :
 *
 *  Transcodage 1 octet non sign (identique  carflo) :
 *  $ fixflo -f -- 0 127 255
 *  0.000000 0.498039 1.000000
 *
 *  Transcodage 1 octet sign  :
 *  $ fixflo -s -- -128 0 127
 *  -1.000000 0.000000 1.000000
 *
 *  Transcodage 2 bits sign :
 *  $ fixflo -b 2 -s -- -2 0 1
 *  -1.000000 0.000000 1.000000
 *
 *  La valeur d'exposant est pris en compte dans le transcodage :
 *  $ fixflo -f -e 1 -- 255
 *  2.000000
 *  $ fixflo -f -e -1 -- 255
 *  0.500000
 *
 * Parceque ce programme n'utilise que la fonction cnvtbg(), les valeurs de biais 
 * et d'chelle (-imbi -imsc) ne sont pas pris en compte.
 *
 * \seealso carflo, flofix
 * $Id$
 */

#include <inrimage/image.h>
#include <string.h>
#include <stdlib.h>


static char version[] = "1.00";
static char detail[] = "[options de format] -- valeur [valeur ...]";
static char options[] = 
  "Transcode une valeur en codage a virgule fixe en valeur a virgule flottante.\n"
  "Les options sont : \n"
  "  [options de format] : voir manuel de 'create'\n"
  "  -- : separateur avant la liste des valeurs a convertir (obligatoire)\n"
  "  valeurs ... : liste des valeurs a convertir \n";

extern int debug_;

int main( int argc, char **argv) {
  Fort_int lfmt[9]={0,0,0,0,0,0,0}, iopt;
  int val, argp ;
  Fort_int icodo[2] = {sizeof(float),REELLE}, iexpo = 0;
  float out;
  int lgbit;

  inr_init( argc, argv, version, detail, options);
  iopt = 1;
  optfmt_((void*)lfmt,&iopt);
  if( debug_) wrfmg_(lfmt,&iopt);

  for( argp = 0;
       argp < argc && strcmp( argv[argp], "--");
       argp ++);

  for( argp ++; argp < argc; argp ++) {
    unsigned char uc;
    unsigned short us;
    unsigned int ui;
    float f;
    void *in;

    switch( TYPE) {
    case PACKEE:
      /* on force a FIXE, cela ne change rien  l'affaire  */
      TYPE=FIXE;
    case FIXE:
      /* on traite a la fois les valeurs signes et non signes
       * si on entre une valeur signe, il y a une conversion en unsigned par
       * le compilateur */
      
      /* si codage en bit, faire tenir sur un multiple de 8 bits */
      lgbit = (BSIZE>0)?BSIZE:((-BSIZE+7)/8);
      switch( lgbit) {
      case 1:
	uc = atoi( argv[argp]);
	in = &uc;
	break;
      case 2:
	us = atoi( argv[argp]);
	in = &us;
	break;
      case 4:
	ui = atol( argv[argp]);
	in = &ui;
	break;
      }
      break;
    case REELLE:
      f = atof( argv[argp]);
      in = &f;
    }

    c_cnvtbg( (void*)in, (void*)&out, 1, &BSIZE, icodo, EXP, iexpo);
    printf( "%f ", out);    
  }
  printf( "\n");
  return 0;
}
