/**
 * \file flofix.c Trancodage virgule flottante en virgule fixe.
 *
 * flofix transcode une valeur en codage  virgule flottante en codage  virgule
 * fixe. Ce programme ralise l'opration inverse de fixflo. Leur utilisation
 * conjointe permet de transcoder n'importe quel codage en n'importe quel autre
 * codage.
 *
 * Usage : flofix [options de codage] -- valeur [valeur ...]
 *         options de format : voir le manuel de create pour une liste 
 *                             exhaustive de codage (-f,-r,-s, ...)
 *         -- : ce champs sert de sparateur entre les options de codage
 *              et les valeurs  transcoder. Il est obligatoire : s'il n'est
 *              pas trouv, le programme ne fait rien.
 *
 * Exemples :
 *
 *  Transcodage vers 1 octet non sign (identique  carflo) :
 *  $ flofix -f -- 0 0.5 1
 *  0 128 255
 *
 *  Transcodage vers 1 octet sign :
 *  $ flofix -s -- -1 -0.5 0 0.5 1
 *  -128 -64 0 64 127
 *
 *  Transcodage 4 bits sign :
 *  $ flofix -b 4 -s -- -1 0 1
 *  -8 0 7
 *
 *  La valeur d'exposant est pris en compte dans le transcodage :
 *  $ flofix -f -e 1 -- 1
 *  128
 *
 *  Transcoder un codage 1 octet vers 2 octet (avec fixflo) :
 *  $ flofix -o 2 -- `fixflo -o 1 -- 127 `
 *  32639
 *
 * Parceque ce programme n'utilise que la fonction cnvtbg(), les valeurs de biais 
 * et d'chelle (-imbi -imsc) ne sont pas pris en compte.
 *
 * \seealso carflo, fixflo
 *
 * $Id$
 */

#include <inrimage/image.h>
#include <string.h>
#include <stdlib.h>


static char version[] = "1.00";
static char detail[] = "[options de format] -- valeur [valeur ...]";
static char options[] = 
  "Transcode une valeur en codage a virgule flottante en valeur a virgule fixe.\n"
  "Les options sont : \n"
  "  [options de format] : voir manuel de 'create'\n"
  "  -- : separateur avant la liste des valeurs a convertir (obligatoire)\n"
  "  valeurs ... : liste des valeurs a convertir \n";

extern int debug_;

int main( int argc, char **argv) {
  Fort_int lfmt[9]={0,0,0,0,0,0,0}, iopt;
  int argp ;
  Fort_int icodi[2] = {sizeof(float),REELLE}, iexpi = 0;
  int lgbit;

  inr_init( argc, argv, version, detail, options);
  iopt = 1;
  optfmt_((void*)lfmt,&iopt);
  if( debug_) wrfmg_(lfmt,&iopt);

  for( argp = 0;
       argp < argc && strcmp( argv[argp], "--");
       argp ++);

  for( argp ++; argp < argc; argp ++) {
    long out;
    float f,outf;

    f = atof( argv[argp]);
     
    switch( TYPE) {
    case PACKEE:
    case FIXE:
      c_cnvtbg( (void*)&f, (void*)&out, 1, icodi, &BSIZE, iexpi, EXP);
      /* on traite a la fois les valeurs signes et non signes
       * si on entre une valeur signe, il y a une conversion en unsigned par
       * le compilateur */
      
      /* si codage en bit, faire tenir sur un multiple de 8 bits */
      lgbit = (BSIZE>0)?BSIZE:((-BSIZE+7)/8);
      switch( lgbit) {
      case 1:
	printf( "%d ", EXP<0?(char)out:(unsigned char)out);
	break;
      case 2:
	printf( "%d ", EXP<0?(short)out:(unsigned short)out);
	break;
      case 4:
	printf( "%ld ", EXP<0?(int)out:(unsigned int)out);
	break;
      }
      break;
    case REELLE:
      c_cnvtbg( (void*)&f, (void*)&outf, 1, icodi, &BSIZE, iexpi, EXP);
      printf( "%f ", outf);
    }
  }
  printf( "\n");
  return 0;
}
