/* 
 * imap: pipeline plan par plan 
 *
 * (c) 2017 D.Béréziat 
 */

#include <inrimage/image.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/wait.h>
#include <string.h>
extern int debug_;

char **parse_command(char *cmd) {
  char **outcmd = NULL;
  char name[128];
  int count = 1;

  for(;; count++) {
    while( *cmd == ' ') cmd++;

    outcmd = (char**)realloc(outcmd,sizeof(char*)*count);
    outcmd[count-1]=cmd;
    
    cmd = strchr(cmd, ' ');
    if( !cmd ) break;
    else *cmd++ = 0;
  }
  count++;
  outcmd = (char**)realloc(outcmd,sizeof(char*)*count);
  outcmd[count-1] = NULL;

  return outcmd;
}


int main(int argc, char **argv) {
  struct image *nfin, *nfout;
  float *buf;
  Fort_int lfmt[9], lfmt1z[9], lfmt1zout[9];
  char name[128];
  int dimz, iz;
  int flg_out; 
  char command[256], command2[256];
  
  inr_init( argc, argv, "1.0",
	    "[image-in] -pipe 'command' [-no|image-out] [options-de-format]",
	    "Filtrer chaque plan de image-in pour former en sortie image-out. "
	    "Par défaut image-out a le format de image-in, mais on peut le modifier via les options de "
	    "format habituelles.\n"
	    "\t-no: (no-ouput) ne pas écrire d'image de sortie. Attention si la commande imprime, l'ordre "
	    "d'affichage peut ne pas être respecté,\n"
	    "\t-pipe: commande à appliquer en pipeline à chaque plan. La commande voie un plan comme une image mono-plan. "
	    "La chaîne donnée peut contenir jusqu'à cinq directives de format %d (ou équivalent) qui "
	    "seront remplacées par le numéro du plan courant. "
	    "La commande est lancée par le shell courant.\n\n"
	    "Exemples:\n"
	    "\tnormaliser, puis appliquer un médian plan à plan:\n"
	    "\t\timap image-in -pipe 'norma|med' image-out\n"
	    "\tcalculer des statistiques plan par plan:\n"
	    "\t\timap image-in -pipe 'echo \"plan %d: $(ical)\"'\n"
	    "\tséparer les plans d'une séquence:\n"
	    "\t\timap image-in -pipe 'cat > plan-%03d.inr' -no\n"
	    )
    ;

  flg_out = 1-igetopt0("-no");
  if( ! igetopt1("-pipe","%s",command))
    imerror(33,"option -pipe requise\n");
  // fprintf(stderr,"%s\n",command);
  infileopt(name);
  nfin = image_(name,"e","",lfmt);
  dimz = NDIMZ;

  for( iz = 0; iz < 9; iz++) lfmt1z[iz]=lfmt[iz];
  lfmt1z[I_NDIMZ] = 1;
  lfmt1z[I_DIMY] = NDIMY;
  buf = (float*)malloc(sizeof(float)*DIMX*NDIMY);
  
  if( flg_out) {
    outfileopt(name);

    /* Format de l'image de sortie, par défaut, le même que l'entrée.
       c_optfmt() réinitialise lfmt, alors que l'on souhaite garder
       les valeurs en cas d'absence d'options */
    if( opt_modfmt((NF_fmt*)lfmt,3))
      ver_gfmt((NF_fmt*)lfmt,3);
    
    nfout = image_(name,"c"," ",lfmt);
    if( debug_) c_prtnf(nfout,3,stderr);

    for( iz = 0; iz < 9; iz++) lfmt1zout[iz]=lfmt[iz];
    lfmt1zout[I_NDIMZ] = 1;
    lfmt1zout[I_DIMY] = NDIMY;

    /* Peut-être que ça n'a pas vraiment d'intérêt de pouvoir changer
       le format de la sortie car on peut le faire avec extg et cco en
       amont ! à réfléchir. mais ça ne coûte pas cher de le faire.
    */
  }

  for( iz = 1; iz <= dimz; iz++) {
    int pid, cpsto, cpsti, fd[4], status;
    
    char **cmd;
    struct image *nfpip;
#if 0
    int l;

    for(l=0;l<NDIMY;l++)
      c_lecflt( nfin, 1, buf+l*DIMX);
#else
      c_lecflt( nfin, NDIMY, buf);
#endif

    /* un tube pour envoi de données du père au fils */
    pipe(fd); /* fd[1]: entrée du tube, fd[0]: sortie */
    /* un second tube pour récupérer les données traitées par le fils */
    if(flg_out) pipe(fd+2);

    pid = fork();

    switch (pid) {
    case -1:
      imerror(33,"invalid fork\n");
      break;
      
    case 0:
      /* parent */
      cpsto = dup(STDOUT_FILENO); /* faire une copie de l'id de stdout */
      dup2(fd[1],STDOUT_FILENO);  /* rediriger la sortie standard du
				     père (il ecrit les plans d'image)
				     dans l'entrée du tube */      
      close(fd[0]); /* il faut fermer la sortie du tube inutilisée */

      if( flg_out) {
	cpsti = dup(STDIN_FILENO); /* copie de stdin ,*/ 
	dup2(fd[2],STDIN_FILENO);  /* puis redirection sortie second tube */
	close(fd[3]);
      }
      
      nfpip = image_(">","c","",lfmt1z);
      c_ecrflt(nfpip,NDIMY,buf);  /* on écrit dans la sortie standard
				     mais celle-ci a été redirigée
				     dans l'entrée du tube */
      fermnf_(&nfpip);
      dup2(cpsto,STDOUT_FILENO);   /* annuler la redirection de stdout */
      close(fd[1]); 
      
      if( flg_out) {
	nfpip = image_("<","e","",lfmt1zout);
#if 0	
	for( l=0; l<NDIMY; l++)
	  c_lecflt(nfpip,1,buf+l*DIMX);
#else
	c_lecflt(nfpip,NDIMY,buf);
#endif
	fermnf_(&nfpip);
	dup2(cpsti,STDIN_FILENO);   /* annuler la redirection de stdin */
	close(fd[2]);
	c_ecrflt(nfout,NDIMY,buf);
      }

      wait(&status);
      break;

    default:
      /* child */

      dup2(fd[0],STDIN_FILENO); /* dupliquer l'entrée standard du fils
				   (il lance une commande qui lit sur
				   son entrée standard) dans la sortie
				   du tube */
      close(fd[1]); /* obligatoire */

      if( flg_out) {
	dup2(fd[3],STDOUT_FILENO);
	close(fd[2]);
      }
      
      sprintf( command2, command, iz, iz, iz, iz, iz);
      // fprintf(stderr,"%s\n",command2);
      /*
      cmd = parse_command(command2);
      
      execvp(cmd[0],cmd); */ /* le processus fils est remplacé par une
			     commande, le break n'est jamais atteint,
			     la commande lit l'entrée standard,
			     c'est-à-dire la sortie du tube */
      system(command2);
      fflush(stdout);
      return 0;
      break;
    }

  }
  fermnf_(&nfin);
  if( flg_out) fermnf_(&nfout);
  return 0;  
}
