/*
 * Comparaison entre deux formats d'image (entete)
 * On peut cumuler les tests (faire des AND) avec autant
 * d'image que souhait. Pour faire des OR ou des AND, on 
 * combinera avec la commande test (a moins de faire mieux).
 *
 * TODO : faire un message de difference plus explicite.
 *        voir pour la maille.
 */

#include <inrimage/image.h>
#include <stdio.h>

/* masque retourne par optfmt */
#define T_SIZE 0x4
#define T_TYPE 0X8
#define T_X    0x10
#define T_Y    0x20
#define T_V    0x40
#define T_Z    0x80
#define T_EXP 0x100
#define T_X0  0x200
#define T_Y0  0x400
#define T_Z0  0x800
#define T_BI 0x2000
#define T_SC 0x4000


char version[] = "1.0";
char command[] = 
  "[coding format switches without argument] [-S] image1 image2 [...]\n"
  "             [coding format switches with argument] [-S] image";
char details[] = "Switches are:\n"
"\t-S: silent option\n"
"\t-x,-y,-z,-v: switches to test an image size\n"
"\t-e: image exponent\n"
"\t-o,-b: value size (in byte or in bit)\n"
"\t-r,-f,-p: format of coding value (real,fix,packed)\n"
"\t-x0,-y0,-z0: image origin\n"
"\t-imsc,-imbi: image scale and image bias\n"
;


/* test l'egalite des deux formats, option par option */

int test_fmt( NF_fmt *lfmt_s, NF_fmt *gfmt_s, Fort_int iopt){
  Fort_int *lfmt = lfmt_s -> lfmt;
  Fort_int *gfmt = gfmt_s -> lfmt;
  int res = 1;
 
  if( iopt & T_X && !(lfmt[I_NDIMX] == gfmt[I_NDIMX]))
    res = 0;
  if( iopt & T_Y && !(lfmt[I_NDIMY] == gfmt[I_NDIMY]))
    res = 0;
  if( iopt & T_Z && !(lfmt[I_NDIMZ] == gfmt[I_NDIMZ]))
    res = 0;
  if( iopt & T_V && !(lfmt[I_NDIMV] == gfmt[I_NDIMV]))
    res = 0;
  if( iopt & T_EXP && !(lfmt[I_EXP] == gfmt[I_EXP]))
    res = 0;
  if( iopt & T_SIZE && !(lfmt[I_BSIZE] == gfmt[I_BSIZE]))
    res = 0;
  if( iopt & T_TYPE && !(lfmt[I_TYPE] == gfmt[I_TYPE]))
    res = 0;
  if( iopt & T_X0 && !(lfmt_s->offsets[0] == gfmt_s->offsets[0]))
    res = 0;
  if( iopt & T_Y0 && !(lfmt_s->offsets[0] == gfmt_s->offsets[1]))
    res = 0;
  if( iopt & T_Z0 && !(lfmt_s->offsets[0] == gfmt_s->offsets[2]))
    res = 0;
  if( iopt & T_BI && !(lfmt_s->bias == gfmt_s->bias))
    res = 0;
  if( iopt & T_SC && !(lfmt_s->scale == gfmt_s->scale))
    res = 0;
  return res;
}

extern int debug_;

int main( int argc, char **argv) {
  NF_fmt lfmt_s, gfmt_s;
  Fort_int *lfmt = lfmt_s.lfmt;
  Fort_int *gfmt = gfmt_s.lfmt;
  struct image *nf;
  char fname[256];
  int res = 0;
  Fort_int iopt = 0;

  inr_init( argc, argv, version, command, details);

  /* il faut au moins uneimage */
  infileopt(fname);
  nf = imagex_( fname, "ex", "", &lfmt_s);
  if(debug_) c_prtnf( nf, 1+2+4+8, stdout);
  fermnf_(&nf);

  
  if( infileopt(fname)) {

    /* cas o on teste deux images ou plus */

    /* lecture des options */    
    iopt |= T_X*igetopt0("-x");
    iopt |= T_Y*igetopt0("-y");
    iopt |= T_Z*igetopt0("-z");
    iopt |= T_V*igetopt0("-v");
    iopt |= T_EXP*(igetopt0("-e")|igetopt0("-s"));
    iopt |= T_SIZE*(igetopt0("-b")|igetopt0("-o"));
    iopt |= T_TYPE*(igetopt0("-f")|igetopt0("-r")|igetopt0("-p"));
    iopt |= T_X0*igetopt0("-x0");
    iopt |= T_Y0*igetopt0("-y0");
    iopt |= T_Z0*igetopt0("-z0");
    iopt |= T_BI*igetopt0("-imbi");
    iopt |= T_SC*igetopt0("-imsc");

    /* todo: format etendu */
    
    do {
      
      nf = imagex_( fname, "ex", "", &gfmt_s);
      if(debug_) c_prtnf( nf, 1+2+4+8, stdout);
      fermnf_(&nf);

      res = test_fmt( &lfmt_s, &gfmt_s, iopt);

      if( res == 0) 
	break;
    
    } while( infileopt(fname));
  } else {
    /* cas o on teste une image contre un format prcis */


    /* lecture du format 
     * non document : optfmt retourne un masque de bit
     * pour les options rencontres
     *  10 -x 
     *  20 -y
     *  40 -v
     *  80 -z
     *   4 -o,-b
     *   8 -r,-f,-p
     * 100 -s,-e
     * 200 -x0
     * 400 -y0
     * 800 -z0
     *2000 -imbi
     *4000 -imsc
     */
    
    iopt = 1+2+4+8;
    iopt = optfmt_( &gfmt_s, &iopt);

    /* pb avec test maille : voir le code de optfmt() */
    res = test_fmt( &lfmt_s, &gfmt_s, iopt);

  }
    
  if( !igetopt0( "-S"))
    fprintf( stdout, "format %s\n", res ? "identique":"different");
  return !res;
}

