/*!!! ATTENTION PB avec -p */
/* genere Postscript a partir d'une ou plusieurs images
im2ps [-i] [-D] [-k octets] [-p pas] [inputfile] [-x xoff] [-y yoff] [-L] >out
centrage horiz. ou vert. si xoff ou yoff == 'c'
genere une BoundingBox pour Encapsulated PostScript 
*/


/*
1cm = 28.3 unites Ps
taille image standard en cm : DIMX * .016933
*/

#include <stdio.h>
#include <inrimage/image.h>
#include <inrimage/error.h>

static char Ucmd[]= "[-D] [-k octets] [-conv] [-wr postscript-file-out] [-p pas] [-x xoff] [-y yoff] [-L | -up] [-sc n | [-xsc n][-ysc n]] [-i]  [file ... file]";
static char Udetail[] =
"\t-wr : fichier de sortie (defaut stdout)\n\
\t-conv : conversion des images couleur en NB\n\
\t-x, -y : offsets (centrage avec -x c -y c)\n\
\t-sc ou -xsc, -xsc : facteurs d'echelle\n\
\t-L : landscape: rotation de 90 degres\n\
\t-up : upsidedown = symetrie par rapport axe X\n\
\t-p : pas d'echantillonage\n\
\t-i : pour insertion LaTeX avec \\image ( 1 seul fichier dans ce cas)\n\
";

/* echange de 2 valeurs 32 bits */
#define EXCHG(x,y) {x ^= y; y^= x; x ^= y;}

#define echan(x) ((x -1)/ pas + 1)
/* coeff de conversion de pixel image en cm */
#define COEFCM .016933
/* 72 pts Postscript par pouce = 28.3/cm */
#define IN2PS 72
#define CM2PS 28.346
#define PAGE_WIDTH ((int) ((21 * CM2PS) + .5))
#define PAGE_HEIGHT ((int) ((29.7 * CM2PS) + .5))

#define OUTLINE 39

#define MAXCOL 256
static char tabr[MAXCOL], tabg[MAXCOL], tabb[MAXCOL];
static int ntabcou;
static char *pcol[3] = {tabr,tabg,tabb};
extern FILE *inr_PSFILE;
extern int khoros_mode;

int i_out= 0;
static unsigned char *tmpbuf;
static unsigned char hxtab[]= "0123456789abcdef";
static int pg_ctr;

int g_size;
int lglig;
int ix_min, ix_max, iy_min, iy_max;
int iwidth, iheight;  /* taille image en coord Ps */
int imwidth, imheight;  /* taille image en coord Ps */

float x_bot= 0, y_bot= 0;  /* position en cm du coin bas gauche 0 au lieu de .5 .5*/
float width, height;   /* taille image en cm */
float scale= 1.;  /* scale facteur */
float xscale= 1., yscale= 1.;


int xoff0, yoff0;  /* offsets globaux centrage si < 0 */
int xoff, yoff;  /* offsets de page centrage si < 0 */
int ix_off, iy_off;  /* offsets en points PS */
int pas= 1;


static Fort_int zero= 0;
static int landscape= 0, updown= 0;
static int ins_mode = 0;
static int h_pixels, v_pixels;  /* horiz & vert. bitmap sizes */
static int h_pg, v_pg;      /* horiz & vert size of page in PS points */


extern errno;
int sim2ps();
static Fort_int ier, nvar, nbv;
/*extern unsigned char *i_malloc(); */

static FILE *outf;

main(argc,argv)
int argc;
char **argv;
{
	register i, nbim;
	register nx, ny;
	struct image *nf;
	int ier;
	register unsigned char *ptb;
	char in[512];
	register next;
	int tobw;

	inr_init(argc,argv,"2.1",Ucmd,Udetail);

	g_size = 0;
	igetopt("-p","",0,"%d",&pas,"",0);
	if(pas <= 0)
		imerror(9,"le pas doit etre >= 1\n");
	if(!(landscape = igetopt("-L","",0,"",0,"",0)))
		updown = igetopt("-up","",0,"",0,"",0);
	ins_mode = igetopt("-i","",0,"",0,"",0);
	if(igetopt("-sc","",0,"%f",&scale,"",0)) {
		yscale = xscale = scale;
	} else {
		igetopt("-xsc","",0,"%f",&xscale,"",0);
		igetopt("-ysc","",0,"%f",&yscale,"",0);
	}
	if(xscale == 0.) xscale = 1.;
	if(yscale == 0.) yscale = 1.;
	getoffset("-x",&xoff0);
	getoffset("-y",&yoff0);
	if((outf = asciifileopt("-wr",1)) == NULL)
		outf = stdout;
	inr_PSFILE = outf;
	tobw = igetopt0("-conv");
	if(tstoptm())
		iusage_(Ucmd,Udetail);
	nbim = tstnbopts();
	if(nbim > 1 && ins_mode)
		iusage_(Ucmd,Udetail);
	fileopt(in,"i");
	next = 1;
	psprologue(nbim < 0 ? 1 : nbim);
	pg_ctr = 1;
	ier = 0;
	if(nbim <= 0) silent_err();
	while(next) {
	  if(nbim <= 0) c_erreur(ier);
	  if(ier == 0) 
	    psimage1(in,tobw);
	  else
	    break;
	  pg_ctr++;
	  if(nbim > 0) 
	    next = xfileoptg(in,512,"i",0);
	}
	fprintf(outf,"%%EOF\n");
	return 0;

}

getoffset(strg,off)
register char *strg;
register int *off;
{
	char buf[80];
	*off = 0;
	if(igetopt(strg,"",0,"%s",buf,"",0) == 0 ) return 1 /* makes c99 happy */;
	if(ins_mode) iusage_(Ucmd,Udetail);
	if(*buf != 'c') {
		*off = atoi(buf);
		return 1 /* makes c99 happy */;
	}
	*off = -1;
	return 1 /* makes c99 happy */;
}

#define I_BW 0
#define I_VCOL 1
#define I_ZCOL 2
psimage1(in,tobw)
char *in;
{
	register i;
	register nx, ny;
	struct image *nf;
	int i0, nbcol;
	register unsigned char *ptb;
	Fort_int lfmti[9], lfmto[9];
	int type_col;

	nf = image_(in,"e","",lfmti);
	ntabcou = ird_ctrgb(nf,&i0,&nbcol,tabr,tabg,tabb,MAXCOL);
	
	for(i=0;i<9;i++)
		lfmto[i] = lfmti[i];
	type_col = I_BW;
	if (tobw)
	  if(lfmti[I_NDIMV] == 3) 
	    type_col = I_VCOL;
	  else if (lfmti[I_NDIMZ] == 3) 
	    type_col = I_ZCOL;
	nx = lfmti[I_DIMX];
	ny = lfmti[I_DIMY];
	if (type_col == I_VCOL) {
	  nx = lfmti[I_NDIMX];
	  lfmto[I_NDIMV] = 1;
	  lfmto[I_DIMX] = nx;
	} else if (type_col == I_ZCOL){
	  ny = lfmti[I_NDIMY];
	  lfmto[I_NDIMZ] = 1;
	  lfmto[I_DIMY] = ny;
	} 
	h_pixels = echan(nx);
	v_pixels = echan(ny);
	if(lfmti[I_BSIZE] == -1 && lfmti[I_TYPE]< 0 && pas == 1) {
		g_size = 1;
		lglig = (h_pixels +7) / 8;
	} else {
		g_size = 8;
		lglig = h_pixels;
	}
	iwidth = (int) (xscale * h_pixels * COEFCM * CM2PS);
	iheight = (int) (yscale * v_pixels * COEFCM * CM2PS);
	h_pg = PAGE_WIDTH;
	v_pg = PAGE_HEIGHT;
	imwidth = iwidth;
	imheight = iheight;
	if(landscape) {
		EXCHG(iwidth,iheight);
		/* 		EXCHG(h_pg,v_pg); */
	}
	xoff = xoff0;
	yoff = yoff0;
	if(xoff < 0)  /* centrage */
		xoff = (h_pg - iwidth) / 2;
	else
		xoff = xoff * COEFCM * CM2PS;  /* convert to PS points */
	if(yoff < 0)
		yoff = (v_pg - iheight) / 2;
	else
		yoff = yoff * COEFCM * CM2PS;  /* 0 at top of page */
	if(xoff < 0 || yoff < 0 ||
	   (xoff + iwidth) > h_pg || (yoff + iheight) > v_pg) {
		   fprintf(stderr,"Attention! image trop grande pour la page\n");
/*		   return 0; */
	   }
	ix_off = xoff + (int) (x_bot * CM2PS);  /* pas d'arrondi */
	iy_off = yoff + (int) (y_bot * CM2PS);
	width = ((float) iwidth) / CM2PS;
	height = ((float) iheight) / CM2PS;
	iy_off = v_pg - iy_off;
	if(ins_mode) {
		ix_min = ix_off = 0;
		iy_min = iy_off = 0;
		if(landscape)
			iy_off = iheight;
	} else {
		ix_min = ix_off;
		iy_min = iy_off - iheight;
		if(!landscape) 
			iy_off = iy_min;
	}
	ix_max = ix_min + iwidth;
	iy_max = iy_min + iheight;
	if(iy_max < iy_min) {
		fprintf(stderr,"Bad box: ymin > ymax\n");
	}
	if(!khoros_mode)
		fprintf(stderr,"taille %fmm x %fmm\n",10 * width,10 * height);
	fprintf(outf,"%%%%Page: %d %d\n",pg_ctr,pg_ctr);
	fprintf(outf,"%%%%BoundingBox: %d %d %d %d\n",ix_min,iy_min,ix_max,iy_max);
	if(updown)
		iy_off = iy_max;
	if(ix_off != 0 || iy_off != 0) {
		fprintf(outf,"%d %d translate\n",ix_off,iy_off);
	}
	if(landscape)
		fprintf(outf,"-90 rotate\n");
	if(updown)
		fprintf(outf,"%d %d scale\n",imwidth,-imheight);
	else
		fprintf(outf,"%d %d scale\n",imwidth,imheight);
	fprintf(outf,"/DataString %d string def\n",lglig);
	fprintf(outf,"%d %d %d [%d 0 0 -%d 0 %d]\n",h_pixels,v_pixels,g_size,h_pixels,v_pixels,v_pixels);
	fprintf(outf,"{currentfile DataString readhexstring pop} bind image\n");
	tmpbuf = (unsigned char *)i_malloc(2 * OUTLINE);
	if(lfmti[I_BSIZE] == -1 && lfmti[I_TYPE]< 0 && pas == 1) {
		ps_1bit(nf,lfmti);
	} else {
		nvar = 1;
		nbv = 1;
		exec_(&ier,&zero,&zero,&nbv,&nvar,sim2ps,&nf,&type_col,lfmti,lfmto,&ier);
	}
	fprintf(outf,"showpage\n");
	return 0;
}

sim2ps(buf,ndim,nf,type_col,lfmti,lfmto)
char *buf;
Fort_int *ndim;
struct image **nf;
Fort_int *type_col;
Fort_int lfmti[9], lfmto[9];
{
	register i, n, nx, ny;	 /*nx= nb octets/ligne apres unpack */
	int nlig, nxo, nyo, iy;
	int bufsize;
	ny = (*type_col == I_ZCOL)  ? lfmti[I_NDIMY] : lfmti[I_DIMY];
	nx = lfmti[I_DIMX];
	if(lfmti[I_BSIZE] < 0)
		i = (-lfmti[I_BSIZE] +7) >> 3;
	else
		i = lfmti[I_BSIZE];
	nx *= i;
	n = *ndim * sizeof(t_aligned);
	nlig = n / nx;
	if(*type_col == I_ZCOL) nlig /= 3;
	if(nlig <= 0) exit(1);
	if(nlig > ny) nlig = ny;
	/* ?? 	if((i = DIMX) < DIMY) i = DIMY; */
	nxo = echan(lfmto[I_DIMX]);
	nyo = nlig / pas;
	nlig = nyo * pas;
	i_out = 0;
	iy = 1;
	i = ny;
	bufsize = buffer_size(lfmti,nlig);
	while (i > 0) {
	  if(nlig > i) {
	    nlig = i;
	    bufsize = buffer_size(lfmti,nlig);
	  }
	  dops(buf,bufsize,*nf,*type_col,iy,lfmti,nxo,nlig,&pas);
	  i -= nlig;
	  iy += nlig;
	}
}


dops(buf,bufsize,nf,type_col,iy,lfmti,nxo,nlig)
register unsigned char *buf;
struct image *nf;
Fort_int lfmti[9];
int type_col, bufsize;
int nxo, nlig, iy;
{
	register nn, nyo;
	unsigned char *bufr, *bufg, *bufb;
	nyo = echan(nlig);
	nn = nxo * nyo;
	if(type_col == I_ZCOL) {
	  c_lectpl(nf,iy,1,3,nlig,buf);
	  bufr = buf;
	  bufg = bufr + bufsize;
	  bufb = bufg + bufsize;
	  c_cnvvis(bufr,bufr,nlig,lfmti,0,0,0,pas);
	  c_cnvvis(bufg,bufg,nlig,lfmti,0,0,0,pas);
	  c_cnvvis(bufb,bufb,nlig,lfmti,0,0,0,pas);
	  c24_to_bw(bufr,bufg,bufb,buf,nlig * nxo,pcol);
	} else {
	  c_lect(nf,nlig,buf);
	  c_cnvvis(buf,buf,nlig,lfmti,0,0,0,pas);
	  if(type_col == I_VCOL)
	    cvec_to_bw(buf,buf,nlig * nxo,pcol);
	  else if(ntabcou > 1) 
	    c8_to_bw(buf,buf,nlig * nxo,pcol);
	  else
	    apply_1ct(buf,buf,nn,tabr);
	}
	buf2ps(buf,nyo,8 * nxo);
}

/* nbl = nb lignes
nbits : nb bits /ligne  (paddees sur 1 octet 
*/
buf2ps(buf,nbl, nbits)
register unsigned char *buf;
register nbl, nbits;
{
	register j, i, y, v, n;
	register unsigned char *out;
	j = 0;
	out = tmpbuf;
	n = (nbits + 7) / 8;
	for(y=0;y<nbl;y++) {
		for(i=0;i<n;i++) {
			v = *buf++;
			*out++ = hxtab[v >> 4];
			*out++ = hxtab[v & 0xf];
			j++;
			if(j >= OUTLINE) {
				*out = 0;
				fprintf(outf,"%s\n",tmpbuf);
				j = 0;
				out = tmpbuf;
			}
		}
	}
	if(j) {
		*out = 0;
		fprintf(outf,"%s\n",tmpbuf);
		j = 0;
		out = tmpbuf;
	}
}

/* code en Ps une image 1 bit */
ps_1bit(nf,lfmt)
register struct image *nf;
Fort_int lfmt[9];
{
	register n;
	register unsigned char *bufi;
	n = ((DIMX + 7) / 8);
	i_out = 0;
	bufi = (unsigned char *)i_malloc(n * DIMY);
	c_xlect(nf,DIMY,bufi,1);
	buf2ps(bufi,DIMY,n * 8);
}

psprologue(n)
register n;
{
	ps_creator();
	fprintf(outf,"%%%%Pages: %d %d\n",n,n);
	fprintf(outf,"%%%%EndComments\n");
}

