/*!!! ATTENTION PB avec -p */
/* genere Postscript a partir d'une ou plusieurs images
im2ps [-i] [-D] [-k octets] [-p pas] [inputfile] [-x xoff] [-y yoff] [-L | -up] >out
centrage horiz. ou vert. si xoff ou yoff == 'c'
genere une BoundingBox pour Encapsulated PostScript 
 images avec NDIMZ==3 ou NDIMV==3 ou NDIMV,NDIMZ==1
 on aplique les T.C. s'il y en a 3 differentes
*/

#include <stdio.h>
#include <inrimage/image.h>
#include <inrimage/error.h>

static char Ucmd[]= "[-D] [-k octets] [-wr postscript-file-out] [-p pas] [-x xoff] [-y yoff] [-L | -up] [-sc n | [-xsc n][-ysc n]] [-i] [file ... file]";
static char Udetail[] =
"\tproduit du PostScript couleur a partir d'images couleur (DIMZ=3 ou DIMV=3)\n\
\t-wr : fichier de sortie (defaut stdout)\n\
\t-x, -y : offsets (centrage avec -x c -y c)\n\
\t-sc ou -xsc, -ysc : facteurs d'echelle \n\
\t-L : landscape\n\
\t-up : upsidedown = symetrie par rapport axe X\n\
\t-p : pas d'echantillonage\n\
\t-i : pour insertion LaTeX avec \\image ( 1 seul fichier dans ce cas)\n\
";

/*
1cm = 28.3 unites Ps
taille image standard en cm : NDIMX * .016933
*/

	
/* echange de 2 valeurs 32 bits */
#define EXCHG(x,y) {x ^= y; y^= x; x ^= y;}

#define echan(x) ((x -1)/ pas + 1)
/* coeff de conversion de pixel image en cm */
#define COEFCM .016933
/* 72 pts Postscript par pouce = 28.3/cm */
#define IN2PS 72
#define CM2PS 28.346
#define PAGE_WIDTH ((int) ((21 * CM2PS) + .5))
#define PAGE_HEIGHT ((int) ((29.7 * CM2PS) + .5))

#define OUTLINE 39
#define MAXCOL 256
static char tabr[MAXCOL], tabg[MAXCOL], tabb[MAXCOL];
static int ntabcou;

extern FILE *inr_PSFILE;
extern int khoros_mode;

int i_out= 0;
static unsigned char *tmpbuf;
static unsigned char hxtab[]= "0123456789abcdef";
static int pg_ctr;

int g_size;
int lglig;
int ix_min, ix_max, iy_min, iy_max;
int iwidth, iheight;  /* taille image en coord Ps */
int imwidth, imheight;  /* taille image en coord Ps */

float x_bot= 0, y_bot= 0;  /* position en cm du coin bas gauche 0 au lieu de .5 .5*/
float width, height;   /* taille image en cm */
float scale= 1.;  /* scale facteur */
float xscale= 1., yscale= 1.;


int xoff0, yoff0;  /* offsets globaux centrage si < 0 */
int xoff, yoff;  /* offsets de page centrage si < 0 */
int ix_off, iy_off;  /* offsets en points PS */
int pas= 1;


static Fort_int zero= 0;
static int landscape= 0, updown= 0;
static int ins_mode = 0;
static int h_pixels, v_pixels;  /* horiz & vert. bitmap sizes */
static int h_pg, v_pg;      /* horiz & vert size of page in PS points */


static Fort_int lfmt[9];
extern errno;
int sim2psc();
static Fort_int ier, nvar, nbv;
/*extern unsigned char *i_malloc(); */

static FILE *outf;

main(argc,argv)
int argc;
char **argv;
{
	register nbim;
	register i;
	register nx, ny;
	int ier;
	struct image *nf;
	char in[128];
	register next;
	register unsigned char *ptb;

	inr_init(argc,argv,"2.1",Ucmd,Udetail);
	g_size = 0;
	igetopt("-p","",0,"%d",&pas,"",0);
	if(pas <= 0)
		imerror(9,"le pas doit etre >= 1\n");
	if(!(landscape = igetopt("-L","",0,"",0,"",0)))
		updown = igetopt("-up","",0,"",0,"",0);
	ins_mode = igetopt("-i","",0,"",0,"",0);
	if(igetopt("-sc","",0,"%f",&scale,"",0)) {
		yscale = xscale = scale;
	} else {
		igetopt("-xsc","",0,"%f",&xscale,"",0);
		igetopt("-ysc","",0,"%f",&yscale,"",0);
	}
	getoffset("-x",&xoff0);
	getoffset("-y",&yoff0);
	if(xscale == 0.) xscale = 1.;
	if(yscale == 0.) yscale = 1.;
	if((outf = asciifileopt("-wr",1)) == NULL)
		outf = stdout;
	inr_PSFILE = outf;
	if(tstoptm())
		iusage_(Ucmd,Udetail);
	nbim = tstnbopts();
	if(nbim > 1 && ins_mode)
		iusage_(Ucmd,Udetail);
	fileopt(in,"i");
	next = 1;
	psprologue(nbim < 0 ? 1 : nbim);
	pg_ctr = 1;
	ier = 0;
	if(nbim <= 0) silent_err();
	while(next) {
	  if(nbim <= 0) c_erreur(ier);
	  if(ier == 0) 
	    psimage1(in);
	  else
	    break;
	  pg_ctr++;
	  if(nbim > 0) 
	    next = xfileopt(in,"i",0);
	}
	fprintf(outf,"%%EOF\n");
	return 0;

}

getoffset(strg,off)
register char *strg;
register int *off;
{
	char buf[80];
	*off = 0;
	if(igetopt(strg,"",0,"%s",buf,"",0) == 0 ) return 1 /* makes c99 happy */;
	if(ins_mode) iusage_(Ucmd,Udetail);
	if(*buf != 'c') {
		*off = atoi(buf);
		return 1 /* makes c99 happy */;
	}
	*off = -1;
	return 1 /* makes c99 happy */;
}

psimage1(in)
char *in;
{
	register i;
	register nx, ny;
	struct image *nf;
	int i0, nbcol;
	register unsigned char *ptb;
	nf = image_(in,"e","",lfmt);
	ntabcou = ird_ctrgb(nf,&i0,&nbcol,tabr,tabg,tabb,MAXCOL);
	if(ntabcou <= 1)
		ntabcou = 0;
	if(NDIMZ > 1) {
		if(NDIMZ != 3 || NDIMV != 1)
			imerror(9,"bad format for %s: NDIMV  NDIMZ may only be 1 ou 3\n",in);
	} else if(NDIMV > 1) {
		if(NDIMV != 3 || NDIMZ != 1)
			imerror(9,"bad format for %s: NDIMV  NDIMZ may only be 1 ou 3\n",in);
		if(pas != 1)
			imerror(9,"option -p interdite avec NDIMV==3\n");
	}

	nx = NDIMX;
	ny = NDIMY;
	h_pixels = echan(nx);
	v_pixels = echan(ny);
	g_size = 8;
	lglig = h_pixels;
	iwidth = (int) (xscale * h_pixels * COEFCM * CM2PS);
	iheight = (int) (yscale * v_pixels * COEFCM * CM2PS);
	h_pg = PAGE_WIDTH;
	v_pg = PAGE_HEIGHT;
	imwidth = iwidth;
	imheight = iheight;
	if(landscape) {
		EXCHG(iwidth,iheight);
	}
	xoff = xoff0;
	yoff = yoff0;
	if(xoff < 0)  /* centrage */
		xoff = (h_pg - iwidth) / 2;
	else
		xoff = xoff * COEFCM * CM2PS;  /* convert to PS points */
	if(yoff < 0)
		yoff = (v_pg - iheight) / 2;
	else
		yoff = yoff * COEFCM * CM2PS;  /* 0 at top of page */
	if(xoff < 0 || yoff < 0 ||
	   (xoff + iwidth) > h_pg || (yoff + iheight) > v_pg) {
		   fprintf(stderr,"Attention! image trop grande pour la page \n");
/*		   return 0; */
	   }
	ix_off = xoff + (int) (x_bot * CM2PS);  /* pas d'arrondi */
	iy_off = yoff + (int) (y_bot * CM2PS);
	width = ((float) iwidth) / CM2PS;
	height = ((float) iheight) / CM2PS;
	iy_off = v_pg - iy_off;
	if(ins_mode) {
		ix_min = ix_off = 0;
		iy_min = iy_off = 0;
		if(landscape)
			iy_off = iheight;
	} else {
		ix_min = ix_off;
		iy_min = iy_off - iheight;
		if(!landscape) 
			iy_off = iy_min;
	}
	ix_max = ix_min + iwidth;
	iy_max = iy_min + iheight;
	if(iy_max < iy_min) {
		fprintf(stderr,"Bad box: ymin > ymax\n");
/*!!		EXCHG(iy_min,iy_max); */
	}
	if(!khoros_mode)
		fprintf(stderr,"taille %fmm x %fmm\n",10 * width,10 * height);
	fprintf(outf,"%%%%Page: %d %d\n",pg_ctr,pg_ctr);
	fprintf(outf,"%%%%BoundingBox: %d %d %d %d\n",ix_min,iy_min,ix_max,iy_max);
	fprintf(outf,"gsave\n");
	if(updown)
		iy_off = iy_max;
	if(ix_off != 0 || iy_off != 0) {
		fprintf(outf,"%d %d translate\n",ix_off,iy_off);
	}
	if(landscape)
		fprintf(outf,"-90 rotate\n");
	if(updown)
		fprintf(outf,"%d %d scale\n",imwidth,-imheight);
	else
		fprintf(outf,"%d %d scale\n",imwidth,imheight);
	fprintf(outf,"/DataString %d string def\n",3 * lglig);
	fprintf(outf,"%d %d %d [%d 0 0 -%d 0 %d]\n",h_pixels,v_pixels,g_size,h_pixels,v_pixels,v_pixels);
	fprintf(outf,"{currentfile DataString readhexstring pop} bind false 3 colorimage\n");
	tmpbuf = (unsigned char *)i_malloc(2 * OUTLINE);
	nvar = 1;
	nbv = 1;
	exec_(&ier,&zero,&zero,&nbv,&nvar,sim2psc,&nf,lfmt,&ier);

	fprintf(outf,"showpage\ngrestore\n");
	return 0;
}

sim2psc(buf,ndim,nf,lfmt)
char *buf;
Fort_int *ndim;
struct image **nf;
Fort_int lfmt[9];
{
	register i, n, nx, ny;	 /*nx= nb octets/ligne apres unpack */
	int nlig, nxo, nyo;
	int ntot, nbuf;
	int bufsize, iy;
	ny = NDIMY;
	nx = DIMX;
	if(BSIZE < 0)
		i = (-BSIZE +7) >> 3;
	else
		i = BSIZE;
	nx *= i;
	n = *ndim * sizeof(t_aligned);
	nbuf = 1;
	if(NDIMZ == 3)
		nbuf = 3;
	else if((NDIMV == 1) && (ntabcou > 1))
		nbuf = 3;
	bufsize = buffer_split(n,nbuf);
	nlig = bufsize / nx;
	if(nlig <= 0) 
		imerror(8,"not enough memory\n");
	if(nlig > NDIMY) nlig = NDIMY;
	if((i = NDIMX) < NDIMY) i = NDIMY;
	nxo = echan(NDIMX);
	nyo = nlig / pas;
	nlig = nyo * pas;
	i_out = 0;
	iy = 1;
	ntot = NDIMY;
	bufsize = buffer_size(lfmt,nlig);
	while (ntot > 0) {
		if(nlig > ntot) {
			nlig = ntot;
			bufsize = buffer_size(lfmt,nlig);
		}
		dops(*nf,buf,nbuf,bufsize,iy,lfmt,nxo,nlig,&pas);
		iy += nlig;
		ntot -= nlig;
	}
}


dops(nf,buf,nbuf,bufsize,iy,lfmt,nxo,nlig)
register unsigned char *buf;
struct image *nf;
Fort_int lfmt[9];
int nxo, nlig, nbuf, bufsize, iy;
{
	register nn, nyo;
	register unsigned char *buf2, *buf3;
	nyo = echan(nlig);
	nn = nxo * nyo;
	if(nbuf > 1) {
		buf2 = buf + bufsize;
		buf3 = buf2 + bufsize;
	} else {
		buf2 = buf;
		buf3 = buf;
	}
	if(NDIMZ == 3) {
		c_lectpl(nf,iy,1,3,nlig,buf);
		c_cnvvis(buf,buf,nlig,lfmt,0,0,0,pas);
		c_cnvvis(buf2,buf2,nlig,lfmt,0,0,0,pas);
		c_cnvvis(buf3,buf3,nlig,lfmt,0,0,0,pas);
		if(ntabcou) {
			apply_1ct(buf,buf,nn,tabr);
			apply_1ct(buf2,buf2,nn,tabg);
			apply_1ct(buf3,buf3,nn,tabb);
		}
		buf2psc(buf,buf2,buf3,nyo,nxo,3);
		return 1 /* makes c99 happy */;
	}
	c_lptset(nf,iy);
	c_lect(nf,nlig,buf);
/*	cnvcit_(buf,buf,&nlig,lfmt,&zero,&zero,&pas); */
	c_cnvvis(buf,buf,nlig,lfmt,0,0,0,pas);
	if(NDIMV == 1) {
		nbuf = 3;
		if(ntabcou) {
			apply_1ct(buf,buf2,nn,tabg);
			apply_1ct(buf,buf3,nn,tabb);
			apply_1ct(buf,buf,nn,tabr);
		}
	}
	if(NDIMV == 3) {
		nxo *= 3;
		if(ntabcou)
			apply_ctvect(buf,buf,nn,tabr,tabg,tabb);
	}
	buf2psc(buf,buf2,buf3,nyo,nxo,nbuf);
}

/* nbl = nb lignes
lglig : nb octets /ligne
*/
buf2psc(buf,buf2,buf3,nbl, lglig,nbuf)
register unsigned char *buf, *buf2, *buf3;
register nbl, lglig, nbuf;
{
	register j, i, y, v, n;
	register unsigned char *out;
	j = 0;
	out = tmpbuf;
	n = lglig;
	for(y=0;y<nbl;y++) {
		for(i=0;i<n;i++) {
			v = *buf++;
			*out++ = hxtab[v >> 4];
			*out++ = hxtab[v & 0xf];
			j++;
			if(nbuf > 1) {
				v = *buf2++;
				*out++ = hxtab[v >> 4];
				*out++ = hxtab[v & 0xf];
				j++;
				v = *buf3++;
				*out++ = hxtab[v >> 4];
				*out++ = hxtab[v & 0xf];
				j++;
			}
			if(j >= OUTLINE) {
				*out = 0;
				fprintf(outf,"%s\n",tmpbuf);
				j = 0;
				out = tmpbuf;
			}
		}
	}
	if(j) {
		*out = 0;
		fprintf(outf,"%s\n",tmpbuf);
		j = 0;
		out = tmpbuf;
	}
}

psprologue(n)
register n;
{
	ps_creator();
	fprintf(outf,"%%%%Pages: %d %d\n",n,n);
	fprintf(outf,"%%%%EndComments\n");
}

