#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <inrimage/image.h>
#include <inrimage/inrlocal.h>
#include <sys/types.h>
#include <sys/stat.h>


extern int koctet_, errno, debug_;
extern char *get_inr_home();

static char version[]= "1.2";
static char Ucmd[]= "[-cp][-xor][-nobg | -nofg][-c fg/bg][-fn font]\
[-rm n| -lm n][-w textwidth][-h textheight][-cx  | -fx ][-tm n| -bm n][-cy  | -fy ] [-izo n][-nz n][-ivo n][-nv n]{-t text | -tf text-file} [image]";
static char Udetail[]= 
"écriture de texte dans une image\n\
\t-cp (ou pas d'image) : lecture de image(| stdin) et écriture sur stdout\n\
\t-xor : texte écrit avec fonction xor\n\
\t-nobg, -nofg: on n'écrit que les pixels foreground (ou background)\n\
\t-rm, -lm : right / left margin (pixels)\n\
\t-tm, -bm : top / bottom margin (pixels)\n\
\t-w, -h : largeur et hauteur max du texte en pixels\n\
\t-cx, -cy : centrer le texte en x, en y\n\
\t-fx, -fy : mode fill en x ou y (equirépartition)\n\
\t-izo : numéro du plan où écrire (1.. NDIMZ)\n\
\t-nz : répéter écriture dans n plans\n\
\t-ivo : numéro de la 1ère composante vectorielle à modifier (1.. NDIMV)\n\
\t-nv : répéter opération sur n composantes\n\
\t-c : valeurs de foreground et background du texte (def 255/0)\n\
\t-fn : choix de la fonte, si le nom contient '/' c'est un fichier\n\
\t\tsinon on prend le fichier $INR_HOME/fonts/font-xxx.inr.gz\n\
\t-t : texte à écrire\n\
\t-tf : nom du fichier contenant le texte\n\
";


#define ALIGN_R_B 1
#define CENTER 4
#define FILL 8

#define DEF_KOCTET 300

int flg_x= 0, flg_y= 0;
int lgrx= 0, lgry= 0;  /* lgrs pour centrer; 0 = max */
int mrgx= 0, mrgy= 0;  /* marges en X et Y (R ou Left,top ou  bottom*/
int linespace= 3;

int nbz= 1, izo= 0, ivo= 0, nbv= 1;

char font[256]= {0};
char fontname[256];
#define FONT_EXT ".inr.gz"
#define SZ_EXT (sizeof(FONT_EXT) - 1)
#define DEF_FONT "6x13"

char *fntbuf= NULL;
unsigned char *buff= NULL;

Fort_int lfmt[9];
Fort_int lfmt1[9]= {0,0,1,0,0,0,0,0,0};
int nblig, lglig;

int ypos= 0;  /* position courante en y dans le plan*/
int zpos= 0;  /* indice de plan courant (>= 0) */

static int flg_cp= 0, flg_xor=0, flg_nobg= 0, flg_nofg= 0;
int flg_col= 0;
int fgcol= 255, bgcol= 0;

struct image *nfi, *nfo, *nffnt;
int cwidth, csize, cheight;
int xmax;

#define MAXLIG 64
struct tablig {
	int ixdeb, lgr;
	int nsp, lgrsp, nsp1;
};
struct tablig tablig[MAXLIG];
int nbtxtlig;

#define LGR_TEXT 2048
char texte[LGR_TEXT];
int lgrtext;

static void putc_std(), putc_xor(), putc_nobg(), putc_nofg();
static void (*iputc)();
void put_txtlig(), goto_end(), h_advance(), ana_text(), ana_filltext();
int calc_y(), goto_line();
  
int main(argc,argv)
int argc;
char **argv;
{
	register int iy00, iy0, ix0, iz0;
	register int n, i, j;
	register int ny;
	char *pc;
	int fd;
	struct stat statbuf;
	char in[256], out[256];
	char tmp[80];
	inr_init(argc,argv,version,Ucmd,Udetail);
	flg_cp = igetopt0("-cp");
	flg_xor = igetopt0("-xor");
	if(!(flg_nobg = igetopt0("-nobg")))
		flg_nofg = igetopt0("-nofg");
	if(igetopt1("-rm","%d",&mrgx))
		flg_x = ALIGN_R_B;
	else
		igetopt1("-lm","%d",&mrgx);
	if(igetopt0("-cx"))
		flg_x |= CENTER;
	else if(igetopt0("-fx"))
		flg_x |= FILL;
	igetopt1("-w","%d",&lgrx);
	if(igetopt1("-bm","%d",&mrgy))
		flg_y = ALIGN_R_B;
	else
		igetopt1("-tm","%d",&mrgy);
	if(igetopt0("-cy"))
		flg_y |= CENTER;
	else if(igetopt0("-fy"))
		flg_y |= FILL;
	igetopt1("-h","%d",&lgry);
#if 0
	if(flg_x & FILL) {
		fprintf(stderr,"option -fx ignorée pour l'instant\n");
		flg_x ^= FILL;
	}
#endif
	if((flg_col = igetopt1("-c","%s",tmp))) {
		sscanf(tmp,"%d/%d",&fgcol,&bgcol);
		fgcol &= 0xff;
		bgcol &= 0xff;
	}
	if(igetopt1("-izo","%d",&izo))
		nbz = 1;
	if(igetopt1("-nz","%d",&nbz)) {
		if(izo <= 0)
			izo = 1;
	}
	if(igetopt1("-ivo","%d",&ivo))
		nbv = 1;
	if(igetopt1("-nv","%d",&nbv)) {
		if(ivo <= 0)
			ivo = 1;
	}

	if(! igetopt1("-t","%[^\177]",texte)) {
		if(! igetopt1("-tf","%s",texte)) {
			fprintf(stderr,"option -t exigee!\n");
			iusage_(Ucmd,Udetail);
		}
		if((fd = open(texte,0)) < 0)
			imerror(1,"open err %d on text file %s\n",errno,texte);
		lgrtext = read_all(fd,texte,LGR_TEXT);
		close(fd);
	} else
		lgrtext = strlen(texte);
	if(lgrtext <= 0) {
		fprintf(stderr,"Warning! No text\n");
		return 0;
	}
	if(!igetopt1("-fn","%s",font))
		strcpy(font,DEF_FONT);
	if(index(font,'/') != NULL)
		strcpy(fontname,font);
	else {
		if((pc = get_inr_home()) == NULL)
			pc = ".";
		sprintf(fontname,"%s/fonts/font-%s%s",pc,font,FONT_EXT);
	}
	if(! infileopt(in))
		flg_cp= 1;  /* forcer copie sur stdout */
	if(tstopts())
		iusage_(Ucmd,Udetail);
	nfi = image_(in,"e","",lfmt);
	if(flg_cp)
		nfo = image_(">","s","a",lfmt);
	else
		nfo = image_(in,"e","",lfmt);
	nffnt = image_(fontname,"e","t",lfmt1);
	cwidth = lfmt1[I_DIMX];
	cheight = lfmt1[I_NDIMY];
	csize = cwidth * cheight;
	fntbuf = i_malloc(lfmt1[I_DIMX] * lfmt1[I_DIMY]);
	c_lect(nffnt,lfmt1[I_DIMY],fntbuf);
	fermnf_(&nffnt);
	if(mrgx < 0) 
		mrgx = 0;
	else if(mrgx >= NDIMX)
		mrgx = NDIMX - 1;
	if(mrgy < 0) 
		mrgy = 0;
	else if(mrgy >= NDIMY)
		mrgy = NDIMY - 1;
	
	if(lgrx <= 0) 
		lgrx = NDIMX - mrgx;
	else if(lgrx > NDIMX)
		lgrx = NDIMX;
	if(lgry <= 0) 
		lgry = NDIMY - mrgy;
	else if(lgry > NDIMY)
		lgry = NDIMY;
	if(izo > NDIMZ)
		izo = NDIMZ;
	if(izo) {
		if(nbz > (NDIMZ - izo + 1))
			nbz = NDIMZ - izo + 1;
	}
	if(ivo > NDIMV)
		ivo = NDIMV;
	if(ivo) {
		if(nbv > (NDIMV - ivo + 1))
			nbv = NDIMV - ivo + 1;
	} else {
		ivo = 1;
		nbv = NDIMV;
	}
	if(izo > 0) {
		iz0 = izo - 1;
		ny = NDIMY * nbz;
	} else {
		ny = DIMY;
		iz0 = 0;
	}
	if(koctet_ <= 0)
		koctet_= DEF_KOCTET;
	n = BSIZE > 0 ? BSIZE : ((-BSIZE + 7)/8);	
	lglig = n * DIMX;
	nblig = (koctet_ * 1024) / lglig;
	if(flg_cp) {
		if(nblig > ny) nblig = ny;
	} else if(nblig > cheight)
		nblig = cheight;
	if(nblig < cheight)
		imerror(8,"Pas assez de memoire\n");
	buff = (unsigned char *)i_malloc(nblig * lglig);
	if(flg_x & FILL)
		ana_filltext(texte,lgrtext);
	else
		ana_text(texte,lgrtext);
	iputc = putc_std;
	if(flg_nobg)
		bgcol = 0;
	if(flg_nofg)
		fgcol = 0;
	if(flg_xor)
		iputc = putc_xor;
	else if (flg_nobg)
		iputc = putc_nobg;
	else if (flg_nofg)
		iputc = putc_nofg;
	ypos = 0;
	zpos = 0;
	ivo -= 1;
	iy00 = calc_y();
	for(j=0;j<nbz;j++) {
		iy0 = iy00;
		for(i=0;i<nbtxtlig;i++) {
			if((iy0 + cheight) > NDIMY) break;
			if(goto_line(iz0,iy0) < 0) {
				fprintf(stderr,"Attention! dépassement limite de plan\n");
				break;
			}
			put_txtlig(texte,&tablig[i]);
			iy0 = ypos + linespace;
		}
		iz0++;
	}
	goto_end();
	fermnf_(&nfi);
	fermnf_(&nfo);
	return 0;
}

void put_txtlig(txt,ptab)
register unsigned char *txt;
register struct tablig *ptab;
{
	register unsigned char *pc;
	register int n, i, x, c;
	register int nsp, nsp1, lgt, lgsp;
	txt += ptab->ixdeb;
	n = ptab->lgr;
	nsp1 = ptab->nsp1;
	nsp = ptab->nsp;
	lgsp = ptab->lgrsp;
	if(!(flg_x & FILL)) {
		lgt = n * cwidth;
		if(lgt > lgrx) {
			lgt = (n - nsp) * cwidth + (nsp * (cwidth / 2));
			if(lgt <= lgrx)
				ptab->lgrsp = cwidth / 2;
			else
				lgt = lgrx;
		}
	} else {
		lgt = (n - nsp) * cwidth + (nsp1 * lgsp) + ((nsp - nsp1) * (lgsp + 1));
		if(lgt > lgrx) lgt = lgrx;
	}
	if(flg_x & CENTER)
		i = (lgrx - lgt) / 2;
	else
		i = 0;
	if(flg_x & ALIGN_R_B) {
		xmax = NDIMX - mrgx - i;
		x = xmax - lgt;
	} else {
		x = mrgx + i;
		xmax =  x + lgrx;
	}
	lgsp = ptab->lgrsp;
	c_lect(nfi,cheight,buff);
	if(debug_) {
		fprintf(stderr,"**put line at x= %d, y=%d\n",x,ypos);
	}
	nsp = 0;
	for(i=0;i<n;i++) {
		if((c = *txt++) == ' ') {
			if((x + lgsp) >= xmax) break;
			h_advance(buff,x,lgsp);
			nsp++;
			x += lgsp;
			if(nsp == nsp1) lgsp++;
			continue;
		}
		if((x + cwidth) > xmax) break;
		c -= 32;
		if((c >= 0) && (c < lfmt1[I_NDIMZ]))
			iputc(buff,x,&fntbuf[c * csize]);
		x += cwidth;
	}
	c_ecr(nfo,cheight,buff);
	ypos += cheight;
}


int calc_y()
{
	register int y, n, i;
	y = mrgy;
	n = nbtxtlig * cheight;
	if(flg_y & FILL) {
		i = (lgry - n) / (nbtxtlig - 1);
		if(i < 2) i = linespace;
		linespace = i;
	}
	n += (nbtxtlig - 1) * linespace;
	if(n > lgry) n = lgry;
	i = flg_y & CENTER ? (lgry - n) / 2 : 0;
	if(flg_y & ALIGN_R_B) {
		y = NDIMY - mrgy -i - n;
	} else {
		y = mrgy + i;
	}
	return y;
}

void goto_end()
{
	if(!flg_cp) return;
	goto_line(NDIMZ -1,NDIMY);
}

int goto_line(z, y)
register int y, z;
{
	register int dy, n, y1;
	if(y > NDIMY)
		return -1;
	if(z >= NDIMZ) return -1;
	y1 = y + (z * NDIMY);
	if(!flg_cp) {
		c_lptset(nfi,y1 + 1);
		c_lptset(nfo,y1 + 1);
		ypos = y;
		zpos = z;
		return y;
	}
	n = nblig;
	dy = y - ypos + ((z - zpos) * NDIMY);
	while(dy > 0) {
		if(n > dy) n = dy;
		dy -= n;
		c_lect(nfi,n,buff);
		c_ecr(nfo,n,buff);
	}
	ypos = y;
	zpos = z;
	return y;
}

void h_advance(buf,i0,n)
register unsigned char *buf;
register int n, i0;
{
	register int i, j, k;
	register unsigned char *p1;
	register int bg, cw;
	if(flg_nobg) return;
	buf += i0 * NDIMV + ivo;
	bg = bgcol;
	if(NDIMV == nbv) {
		cw = n * NDIMV;
		for(i=0;i<cheight;i++) {
			p1 = buf;
			if(flg_xor) {
				for(j=0;j<cw;j++)
					*p1++ ^=  bg;
			} else {
				for(j=0;j<cw;j++)
					*p1++ =  bg;
			}
			buf += lglig;
		}
		return;
	}
	/* on ne remplit pas toutes les composantes */
	cw = NDIMV - nbv;
	for(i=0;i<cheight;i++) {
		p1 = buf;
		if(flg_xor) {
			for(j=0;j<n;j++) {
				for(k=0;k<nbv;k++)
					*p1++ ^=  bg;
				p1 += cw;
			}
		} else {
			for(j=0;j<n;j++) {
				for(k=0;k<nbv;k++)
					*p1++ =  bg;
				p1 += cw;
			}
			buf += lglig;
		}
	}
}

static void putc_std(buf,i0,pf)
register unsigned char *buf, *pf;
register int i0;
{
	register int i, j, k, v;
	register unsigned char *p1;
	register int bg, fg, dx;
	bg = bgcol;
	fg = fgcol;
	dx = NDIMV - nbv;
	buf += i0 * NDIMV + ivo;
	for(i=0;i<cheight;i++) {
		p1 = buf;
		for(j=0;j<cwidth;j++) {
			v = *pf++  ? fg : bg;
			for(k=0;k<nbv;k++)
				*p1++ = v;
			p1 += dx;
		}
		buf += lglig;
	}
}

static void putc_xor(buf,i0,pf)
register unsigned char *buf, *pf;
register int i0;
{
	register int i, j, k, v;
	register unsigned char *p1;
	register int bg, fg, dx;
	bg = bgcol;
	fg = fgcol;
	dx = NDIMV - nbv;
	buf += i0 * NDIMV + ivo;
	for(i=0;i<cheight;i++) {
		p1 = buf;
		for(j=0;j<cwidth;j++) {
			v = *pf++  ? fg : bg;
			for(k=0;k<nbv;k++)
				*p1++ ^=  v;
			p1 += dx;
		}
		buf += lglig;
	}
}

static void putc_nobg(buf,i0,pf)
register unsigned char *buf, *pf;
register int i0;
{
	register int i, j, k;
	register unsigned char *p1;
	register int fg, dx;
	fg = fgcol;
	dx = NDIMV - nbv;
	buf += i0 * NDIMV + ivo;
	for(i=0;i<cheight;i++) {
		p1 = buf;
		for(j=0;j<cwidth;j++) {
			if(*pf++) {
				for(k=0;k<nbv;k++)
					*p1++ = fg;
				p1 += dx;
			} else 
				p1 += NDIMV;
		}
		buf += lglig;
	}
}

static void putc_nofg(buf,i0,pf)
register unsigned char *buf, *pf;
register int i0;
{
	register int i, j, k;
	register unsigned char *p1;
	register int bg, dx;
	bg = bgcol;
	dx = NDIMV - nbv;
	buf += i0 * NDIMV + ivo;
	for(i=0;i<cheight;i++) {
		p1 = buf;
		for(j=0;j<cwidth;j++) {
			if(*pf++)
				p1 += NDIMV;
			else {
				for(k=0;k<nbv;k++)
					*p1++ = bg;
				p1 += dx;
			}
		}
		buf += lglig;
	}
}

void ana_text(txt,lgtxt)
register unsigned char *txt;
register int lgtxt;
{
	register unsigned char *p2;
	register int i, c, lgsp;
	register int nl, nsp, lg, lgt;
	p2 = txt;
	nl = 0;
	lgt = 0;
	lg = 0;
	nsp = 0;
	tablig[0].ixdeb = 0;
	for(i=0;i<lgtxt;i++) {
		if(!(c = *txt++)) break;
		if(c == '\t')
			c = ' ';
		if(c == '\n') {
			tablig[nl].lgr = lg;
			tablig[nl].nsp = nsp;
			tablig[nl].nsp1 = nsp;
			tablig[nl].lgrsp = cwidth;
			nl++;
			lg = 0;
			nsp = 0;
			if(nl >= MAXLIG) break;
			*p2++ = c;
			lgt++;
			tablig[nl].ixdeb = lgt;
			continue;
		} else if(c == ' ') {
			nsp++;
		}
		*p2++ = c;
		lg++;
		lgt++;
	}
	if(lg) {
		tablig[nl].lgr = lg;
		tablig[nl].nsp = nsp;
		tablig[nl].nsp1 = nsp;
		tablig[nl].lgrsp = cwidth;
		nl++;
	}
	nbtxtlig = nl;
	lgrtext = lgt;
}

void ana_filltext(txt0,lgtxt)
register unsigned char *txt0;
register int lgtxt;
{
	register unsigned char *p2, *txt;
	register int i, c, lgsp;
	register int nl, lg, nsp, nsp1, lgprev, cw2, cwmax;
	register int w, w1, n0, inc0, ixd;
	/* 1e passe : compresser les espaces */
	txt = txt0;
	p2 = txt0;
	lg = 0;
	for(i=0;i<lgtxt;i++) {
		if(!(c = *txt++)) break;
		if((c == '\t') || (c == '\n'))
			c = ' ';
		if(c == ' ') {
			if((lg == 0) || (*(p2 -1) == ' ')) continue;
		}
		*p2++ = c;
		lg++;
	}
	if(c == ' ') lg--;  /* supp last space */
	lgtxt = lg;
	nl = 0;
	lg = 0;
	nsp = 0;
	tablig[0].ixdeb = 0;
	lgprev = 0;
	cw2 = cwidth / 2;
	cwmax = 2 * cwidth;
	w = 0;  /* largeur courante */
	txt = txt0;
	inc0 = 0;
	ixd = 0;
	n0 = 0;
	for(i=1;i<=lgtxt;i++) {
/*		if((c = *txt++) == ' ') nsp++; */
		c = *txt++;
		if((c == ' ') || (i == lgtxt)) {
			w1 = (lg - lgprev - n0) * cwidth + inc0;
			if((w + w1) > lgrx) {
				nsp -= n0;
				if(!lgprev) lgprev = lg;
				tablig[nl].lgr = lgprev;
				tablig[nl].nsp = nsp;
				w = w1 - inc0;
				w1 = lgrx - ((lgprev - nsp) * cwidth);
				if(nsp && (w1 > 0)) {
					lgsp = w1 / nsp;
					nsp1 = nsp - (w1 % nsp);
				} else {
					lgsp = cwidth;
					nsp1 = nsp;
				}
				if(lgsp > cwmax) {
					nsp1 = nsp;
					lgsp = cwidth;
				}
				tablig[nl].nsp1 = nsp1;
				tablig[nl].lgrsp = lgsp;
				nl++;
				lg -= (lgprev + n0);
				lgprev = lg;
				if(c == ' ') {
					lg++;
					nsp = 1;
					n0 = 1;
					inc0 = cw2;
				} else {
					nsp = 0;
					inc0 = 0;
					n0 = 0;
				}
				tablig[nl].ixdeb = ixd;
				ixd = i;
				if(nl >= MAXLIG) break;
				continue;
			} else {
				w += w1;
				ixd = i;
				lgprev = lg;  /* sans le dernier espace */
				inc0 = cw2;
				n0 = 1;
				if(c == ' ')
					nsp++;
				else
					lgprev++;
			}
		}
		lg++;
	}
	if(lg) {
		tablig[nl].lgr = lg;
		w1 = lgrx - ((lgprev - nsp) * cwidth);
		if(nsp && (w1 > 0)) {
			lgsp = w1 / nsp;
			tablig[nl].nsp1 = nsp - (w1 % lgsp);
		} else {
			lgsp = cwidth;
			tablig[nl].nsp1 = nsp;
		}
		tablig[nl].nsp = nsp;
		tablig[nl].lgrsp = lgsp;
		nl++;
	}
	nbtxtlig = nl;
	lgrtext = lgtxt;
}
