/* CRANCH
	SUBROUTINE CRANCH(A,IA,B,IB,N,IND,IDIMX,NBLIG,ISEUIL,IOPT)

 **************************
 * A,B tableaux de donnees et res. commencant en IA et IB (octets)
 *NBLIG lignes de lgr IDIMX (bytesize 1) a traiter
 *WIN fenetre de N paires de Fort_int. 1 paire=j,k
	j = coord. relative dans la ligne (par rapport pt courant)
	k = coord relative dans tableau A
 *le pt courant est remplace par le pnt min de la fenetre (erosion : IOPT(1)=1)
 * ou par le max (dilatation : IOPT(1)=2)
 *on suppose A precede et suivi des lignes necessaires
 * ISEUIL: 2 valeurs (0-255) limites pour les objets : ignorees
 * si IOPT(2)=1 on erode sur les bords (ou on dilate)
 **************************
*/

#include <inrimage/image.h>

cranch_(a,ia,b,ib,n,win,idimx,nblig,iseuil,iopt)
unsigned char *a, *b;
Fort_int *ia, *ib, *n, *idimx, *nblig, *iseuil, *iopt;
Fort_int *win;
{
	if(*iopt == 1) {
		if(*(iopt+2) == 1)
			c_erodeg(a + *ia - 1,b + *ib - 1,(int)*n,win,(int)*idimx,(int)*nblig,(int)*(iopt+1));
		else
			c_erode(a + *ia - 1,b + *ib - 1,(int)*n,win,(int)*idimx,(int)*nblig,iseuil,(int)*(iopt+1));
	} else {
		if(*(iopt+2) == 1)
			c_dilateg(a + *ia - 1,b + *ib - 1,(int)*n,win,(int)*idimx,(int)*nblig);
		else
			c_dilate(a + *ia - 1,b + *ib - 1,(int)*n,win,(int)*idimx,(int)*nblig,iseuil);
	}
}

c_erode(pi,po,n,win,idimx,nblig,iseuil,au_bord)
register unsigned char *pi, *po;
int n, idimx, nblig, *iseuil, au_bord;
Fort_int *win;
{
	register v, w;
	register is1, is2;
	register k, j;
	register i;
	register ki, kj;
	register Fort_int *pwin;
	is1 = *iseuil;
	is2 = *(iseuil +1);
	for(i=0;i<nblig;i++) {
		for(j=0;j<idimx;j++,pi++) {
			v = *pi;
			if(v < is1 || v > is2) {
				*po++ = 0;
				continue;
			}
			v = 255;
			pwin = win;
			k = n;
			while(k-- > 0) {
				kj = *pwin++ + j;  /* indice dans ligne */
				/* test de depassement de ligne*/
				if((kj < 0) || (kj >= idimx)) {
					if(au_bord) { /* erosion  aux bords */
						v = 0;
						break;
					}
					/* erosion sans les bords */
					pwin++;
					continue;
				}
				/* points normaux */
				ki = *pwin++;
				w = *(pi + ki);
				if(w < is1 || w > is2) {
					v = 0;
					break;
				}
			}
			*po++ = v;
		}
	}
}

c_dilate(pi,po,n,win,idimx,nblig,iseuil)
register unsigned char *pi, *po;
int n, idimx, nblig, *iseuil;
Fort_int *win;
{
	register v, w;
	register is1, is2;
	register k, j;
	register i;
	register ki, kj;
	register Fort_int *pwin;
	is1 = *iseuil;
	is2 = *(iseuil +1);
	for(i=0;i<nblig;i++) {
		for(j=0;j<idimx;j++,pi++) {
			v = *pi;
			if(v >= is1 && v <= is2) {
				*po++ = 255;
				continue;
			}
			v = 0;
			pwin = win;
			k = n;
			while(k-- > 0) {
				kj = *pwin++ + j;  /* indice dans ligne */
				/* test de depassement de ligne*/
				if((kj < 0) || (kj >= idimx)) {
					pwin++;
					continue;
				}
				/* points normaux */
				ki = *pwin++;
				w = *(pi + ki);
				if(w >= is1 && w <= is2) {
					v = 255;
					break;
				}
			}
			*po++ = v;
		}
	}
}


c_erodeg(pi,po,n,win,idimx,nblig,au_bord)
register unsigned char *pi, *po;
int n, idimx, nblig, au_bord;
Fort_int *win;
{
	register w;
	register k, j;
	register i;
	register vmin;
	register ki, kj;
	register Fort_int *pwin;
	for(i=0;i<nblig;i++) {
		for(j=0;j<idimx;j++,pi++) {
			if((vmin = *pi) == 0) {
				*po++ = 0;
				continue;
			}
			pwin = win;
			k = n;
			while(k-- > 0) {
				kj = *pwin++ + j;  /* indice dans ligne */
				/* test de depassement de ligne*/
				if((kj < 0) || (kj >= idimx)) {
					if(au_bord) { /* erosion  aux bords */
						vmin = 0;
						break;
					}
					/* erosion sans les bords */
					pwin++;
					continue;
				}
				/* points normaux */
				ki = *pwin++;
				if((w = *(pi + ki)) == 0) {
					vmin = 0;
					break;
				}
				if(w < vmin)
					vmin = w;
			}
			*po++ = vmin;
		}
	}
}

c_dilateg(pi,po,n,win,idimx,nblig)
register unsigned char *pi, *po;
int n, idimx, nblig;
Fort_int *win;
{
	register w;
	register vmax;
	register k, j;
	register i;
	register ki, kj;
	register Fort_int *pwin;
	for(i=0;i<nblig;i++) {
		for(j=0;j<idimx;j++,pi++) {
			if((vmax = *pi) == 255) {
				*po++ = 255;
				continue;
			}
			pwin = win;
			k = n;
			while(k-- > 0) {
				kj = *pwin++ + j;  /* indice dans ligne */
				/* test de depassement de ligne*/
				if((kj < 0) || (kj >= idimx)) {
					pwin++;
					continue;
				}
				/* points normaux */
				ki = *pwin++;
				w = *(pi + ki);
				if(w > vmax)
					vmax = w;
			}
			*po++ = vmax;
		}
	}
}

