/*  
lectures d'un fichier de contours
 (functions)
	irctgp(nfct,numax,nbct,npmax,nblocs,iftyp)	c_irctg
	irctcm(nfct,icom)				c_irctcm
	irctdm(nfct,ixy0,idimxy)			c_irctdm
	irctpa(nfct,num,isens,iext,lgr,par,npmax,np)	c_irctpa
	ircttb(nfct,num,icont,lgmax)	c_ircttb
*/

#include <inrimage/image.h>
#include <inrimage/contour.h>
#include <stdio.h>

extern debug_, errno;

static Fort_int er9= 9, er8= 8;

/*	
	function irctgp(nfct,numax,nbct,npmax,nblocs,iftyp)
lecture de l'info generale du fichier de contours
renvoie dans :
numax : le numero max
nbct : le nb de contours effectifs
npmax : le nb max de param. par contour
nblocs : la taille en blocs du fichier
iftyp : type du fichier (1=> lgr, iext, nbparam de type int4 sur le
fichier, short sinon)
*/

int irctgp_(nfct,numax,nbct,npmax,nblocs,iftyp)
struct fc_cont **nfct;
Fort_int *numax, *nbct, *npmax, *nblocs, *iftyp;
{
	return c_irctgp(*nfct,numax,nbct,npmax,nblocs,iftyp);
}

/**
 * @brief Lecture des informations gnrales d'un fichier de contours
 * @param nfc fichier de contours
 * @param numax le plus haut index de contour
 * @param nbct le nombre de contours effectivement dfinis
 * @param npmax nombre maximum de paramtres par contours
 * @param nblocs taille du fichier (en blocs)
 * @param iftyp format du fichier de contours: si 1, les
 *        longueur, enveloppe et nombre de paramtres d'un contour
 *        sont cods sur int4, sinon cods sur un short.
 * @return ??   
 */

int c_irctgp(nfc,numax,nbct,npmax,nblocs,iftyp)
register struct fc_cont *nfc;
Fort_int *numax, *nbct, *npmax, *nblocs, *iftyp;
{
	if(chk_fc(nfc) < 0) ermesf_(&er9);
	if(!(nfc->fc_access & CT_R)) ermesf_(&er9);
	*numax = nfc->c_nmax;
	*nbct = nfc->c_nb;
	*npmax = nfc->c_npmax;
	*nblocs = nfc->c_nbs + 1;
	*iftyp = nfc->fc_typ;
	return *nbct;
}

/* irctcm(nfct,icom)
renvoie le commentaire associ{ au contour
la valeur de la fonction est egale au nb de car.
*/
int irctcm_(nfct,icom)
struct fc_cont **nfct;
char *icom;
{
	return c_irctcm(*nfct,icom);
}

/**
 * @brief Lecture du commentaire d'un fichier de contours.
 * @param nfc fichier de contours
 * @param icom tableau de 128 caractres.
 * @return nombre de caractres effectivement copis dans icom
 */
int c_irctcm(nfc,icom)
register struct fc_cont *nfc;
char *icom;
{
	register char *p1, *p2;
	register i;
	struct ct_b0 bk0;
	if(chk_fc(nfc) < 0) ermesf_(&er9);
	if(!(nfc->fc_access & CT_R)) ermesf_(&er9);
	xread(nfc->fc_fd,&bk0,0L);
	p1 = bk0.s_com;
	p2 = icom;
	for(i=0;i<128;i++)
		if((*p2++ = *p1++) == '\0') break;
	return i;
}

/* irctdm(nfct,ixy0,idimxy)
renvoie les coordonnees de la fenetre originale dans ixy0, idimxy
code retour 0 si OK -1 si non valide
*/
int irctdm_(nfct,ixy0,idimxy)
struct fc_cont **nfct;
int4 *ixy0, *idimxy;
{
	return c_irctdm(*nfct,ixy0,idimxy);
}

/**
 * @brief Lecture des coordonnes de la fentre originale (s'ils
 *        existent).
 * @param nfc fichier de contours
 * @param ixy0 coordonnes coin haut gauche de la fentre ( partir de 1)
 * @param idimxy taille de la fentre
 * @return 0 si pas d'erreur, -1 sinon.
 */

int c_irctdm(nfc,ixy0,idimxy)
register struct fc_cont *nfc;
int4 *ixy0, *idimxy;
{
	char icom[128];
	long val[4];

	c_irctcm(nfc,icom);	
	int ret = sscanf(icom,"-ix %ld -iy %ld -dx %ld -dy %ld",val,val+1,val+2,val+3);
	*ixy0 = *val;
	ixy0[1] = val[1];
	*idimxy = val[2];
	idimxy[1] = val[3];
	if( ret != 4)
		return -1;
	return 0;
}

/* lecture des parametres 
retour: valeur = isens = 0 si n'existe pas, +1/-1 selon signe 1e param
*/

int irctpa_(nfct,num,isens,iext,lgr,par,npmax,np)
struct fc_cont **nfct;
Fort_int *num, *isens, *iext, *lgr, *par, *npmax, *np;
{
	return c_irctpa(*nfct,(int)*num,isens,iext,lgr,par,(int)*npmax,np);
}

/**
 * @brief Lecture des paramtres d'un contour.
 * @param nfc fichier de contours
 * @param num numro (non sign) de contours commencant  1
 * @param isens sens du contour:
 *        - 1 positif,
 *        - -1 ngatif,
 *        - 0 le contours n'existe pas.
 * @param iext numro d'enveloppe (numero du contour qui le contient directement, 0 si pas d'enveloppe).
 * @param lgr longueur du contour
 * @param par tableau des parametres
 * @param npmax nombre max de parametres
 * @param np  nombre de parametrs
 * @return la valeur du parametre isens.
 *
 * A priori, par contient toujours au moins un parametre
 * qui est la surface du contours.
 * @sa c_irctgp()
 */

int c_irctpa(nfc,num,isens,iext,lgr,par,npmax,np)
register struct fc_cont *nfc;
Fort_int *isens, *iext, *lgr, *np;
int4 *par;
int num, npmax;
{
register lg;
int ist;
union hdrdata hdrd[2];
struct hdrdn hdrdn;
	if(chk_fc(nfc) < 0) ermesf_(&er9);
	if(!(nfc->fc_access & CT_R)) ermesf_(&er9);
	go_indx(nfc,num,0);		/* aller sur sect. index */
		/* aller vers debut param. contour courant */
	if((ist = go_dat(nfc,0,0)) <= 0)
		return (*isens = 0);
	if(nfc->fc_typ) {
		gettab(nfc,&hdrdn,4);
		if(nfc->fc_swap_needed) swap4buf(&hdrdn,sizeof(hdrdn)/4);
		*iext = hdrdn.dn_env;
		*lgr = hdrdn.dn_lgr;
		lg = hdrdn.dn_nbpar;
	} else {
		gettab(nfc,hdrd,2);
		*iext = hdrd[0].d1.d_env;
		*lgr = hdrd[0].d1.d_lgr;
		lg = hdrd[1].d2.d_nbpar;
	}
	if(lg > npmax) ermesf_(&er8);
	if((*np = lg) < 0) ermesf_(&er9);
	if(lg == 0) {
		nfc->ct_loc = 2;
		return (*isens = 1);
	}
	gettab(nfc,par,lg);
	if(nfc->fc_swap_needed) swap4buf(par,lg);
	nfc->ct_loc = 3;
	if((lg = *par) >= 0)
		return (*isens = 1);
	return (*isens = -1);
}
	

/*
	function ircttb(nfct,num,icont,lgmax)
*/

int ircttb_(nfct,num,icont,lgmax)
struct fc_cont **nfct;
Fort_int *num, *icont, *lgmax;
{
	return c_ircttb(*nfct,(int)*num,icont,(int)*lgmax);
}
/**
 * @brief Lit les lments d'un contour.
 * @param nfc fichier de contours
 * @param n numro de contour ( partir de 1)
 * @param icont tableau ou sont copis les lments
 * @param lgmax taille de icont.
 *
 * <i>lgmax</i> doit tre suffisamment grand sinon une erreur
 * Inrimage numro 8 est gnre.
 * Chaque lment de <i>icont</i> code pour une coordonne 
 * 2D. x est code sur les octets de poid faible et y sur
 * les deux octets de poid fort.
 */

int c_ircttb(nfc,n,icont,lgmax)
register struct fc_cont *nfc;
int n, lgmax;
int4 *icont;
{
	register lg;
	int ist;
	if(chk_fc(nfc) < 0) ermesf_(&er9);
	if(!(nfc->fc_access & CT_R)) ermesf_(&er9);
	go_indx(nfc,n,0);		/* aller sur sect. index */
		/* aller vers debut param. contour courant */
	if((ist = go_dat(nfc,2,0)) <= 0)
		return 0;
	if((lg = nfc->ct_lgr) <= 0) return 0;
	if(lg > lgmax) ermesf_(&er8);
	gettab(nfc,icont,lg);
	if(nfc->fc_swap_needed) swap2buf(icont, 2*lg);
	return lg;
}
