#include <stdio.h>
#include <string.h>
#include <inrimage/image.h>
#include "imext.h"

extern errno;
extern debug_;

/* gif structure (les shorts sont swapps)
1 bloc= 1 char longueur, n char data

file:
6 char : GIF87a ou GIF89a

2 short : screen width and height
1 char : flag (0x80= colormap, 0x70= nb -1 color resolution bits, 0x7 = nb bits - 1 => nb colors = 2 ** nbits)
1 char : background
1 char : aspect ratio ,   (float) (byte + 15) / 64.

3 * nbcolors char : colormap si presente (rgb,rgb,...)

1 char type :
	';'  : fin de fichier
	'!' : extension block, suivi de 
		1 char type suivi de blocks, termins par block de taille nulle
		   types: 0xf9 transparence 
		          0xfe comment
			  0x1  plain text
	',' : image separator, suivi des donnes image


Image:
2 short : left offset , top offset
2 short : image width, height
1 char  : flag (0x40 = entrelac, 0x80= colormap, 0x7 = nbits-1)
3*ncolors char si colormap
LZWencoded data


*/


extern errno;
extern debug_;

extern int imext_read1();
extern struct extimage *openimext();


struct gif_ext {
	int nb_alloc;  /* nb de ptrs de data alloues */
	char **pdata;   /* table de ptrs vers sous_images */
};
#define NB_PDATA 8

/* read a count (1 byte) and then the data 
returns the nb of data bytes */
int gif_rdblock(pnf,buf)
register struct extimage **pnf;
register char *buf;
{
	register n, i;
	if((n = getc((FILE *)(*pnf)->file)) == EOF) {
		fprintf(stderr,"GIF block count read err %d\n",errno);
		return -1;
	}
	
	if((i = imext_read1(pnf,buf,n)) != n) {
		fprintf(stderr,"GIF block read (%d) errno %d\n",i,errno);
		return i;
	}
	return n;
}
/* skip n bytes */
static void skipbytes(pnf,n)
struct extimage **pnf;
int n;
{
  int i, c;
  for(i=0;i<n;i++) {
    if((c = getc((FILE *)(*pnf)->file)) == EOF)
      imext_err(pnf,1,"GIF read err %d\n",errno);
  }
}

/* skip blocks */
static int skipblks(pnf)
register struct extimage **pnf;
{
	register n, c;
	char tmp[16];
	while(1) {
		if((n = getc((FILE *)(*pnf)->file)) == EOF)
			imext_err(pnf,1,"GIF read err %d\n",errno);
		if(n == 0) return 0;
		while(n-- > 0) {
			if((c = getc((FILE *)(*pnf)->file)) == EOF)
				imext_err(pnf,1,"GIF read err %d\n",errno);
			if(n < 16) tmp[n] = c;
		}
	}
}

static int rdcmap (pnf)
register struct extimage **pnf;
{
	char tmp[3*256];
	register numcols, n, i;
	numcols = (*pnf)->numcols;
	if(debug_)
		fprintf(stderr,"read GIF colormap (%d colors)\n",numcols);
	if(imext_read1(pnf,tmp,3 * numcols) != (3 * numcols))
		imext_err(pnf,9,"could not read GIF colormap\n");
	n = 0;
	for(i=0;i<numcols;i++) {
		(*pnf)->rbuf[i] = tmp[n++];
		(*pnf)->gbuf[i] = tmp[n++];
		(*pnf)->bbuf[i] = tmp[n++];
	}

}
static void free_pdata(ppnf)
struct extimage **ppnf;
{
	struct gif_ext *pg;
	struct extimage *pnf;
	register i;
	pnf = *ppnf;
	pg = (struct gif_ext *)pnf->x_params;
	if(! pg) return;
	if(! pg->pdata) return;
	for(i=0;i< pnf->ndimz -1;i++) {
		i_Free(pg->pdata + i);
	}
	i_Free((char **)&pg->pdata);
}

/* si flg_cv on veut convertir en 24 bits (=> ndimv = 3) */
struct extimage *opengif(name, flg_cv)
char *name;
int flg_cv;
{
	unsigned char tmp[16];
	register i, n, flgc, c;
	register nbits, numcols;
	int screenw, screenh;
	struct extimage *nfgif;
	struct gif_ext *pg;
	nfgif = openimext(name,0);
	n = 0;
	if(imext_read1(&nfgif,tmp,6) != 6)
		n = -1;
	else if(strncmp(tmp,"GIF8",4) != 0)
		n = -1;
	if(n) {
		imext_err(&nfgif,9,"%s: not a GIF file\n",name);
	}
	if(tmp[4] == '9')
		nfgif->flag = 1;
	/* read gif header */
	if(imext_read1(&nfgif,tmp,7) != 7) 
		imext_err(&nfgif,9,"incorrect GIF file %s\n",name);
	screenw = tmp[0] + 256 * tmp[1];
	screenh = tmp[2] + 256 * tmp[3];
	flgc = tmp[4] & 0x80;
	nbits = (tmp[4] & 7) + 1;
	numcols = flgc ? 1 << nbits : 0;
	nfgif->nbits = nbits;
	nfgif->numcols = numcols;
	if(flgc) { /* read colormap */
		rdcmap(&nfgif);
	}
	/* goto begining of 1st image */
	pg = (struct gif_ext *)i_malloc(sizeof (struct gif_ext));
	pg->pdata = (char **)i_malloc(NB_PDATA * sizeof(char *));
	pg->nb_alloc = NB_PDATA;
	nfgif->x_params = pg;
	nfgif->free_params = free_pdata;
	if(flg_cv)
		nfgif->ndimv = 3;
	else
		nfgif->ndimv = 1;
	if(init_gifimage(&nfgif))
		imext_err(&nfgif,1,"GIF read err %d\n",errno);
	return nfgif;
}

/* return 0 si ok, 1 si plus rien, -1 si pb */
init_gifimage(ppnf)
register struct extimage **ppnf;
{
	register struct extimage *nfgif;
	register struct gif_ext *pg;
	register i, n, flgc, c;
	unsigned char tmp[512];
	FILE *f;
	int ndx, ndy;
	nfgif = *ppnf;
	f = nfgif->file;
	pg = (struct gif_ext *) nfgif->x_params;
	while(1) {
		if((c = getc(f)) == EOF) return 1;
		switch(c) {
		case ';' :  /* terminator */
			return 1;
/*!!			imext_err(ppnf,9,"Terminator char found!!\n"); */
		case '!' : /* extension block */
			if((c = getc(f)) == EOF)
				imext_err(ppnf,1,"GIF read err %d\n",errno);
			switch(c) {
			case 'R':   /* ascpect ratio */
			  if(debug_)
					fprintf(stderr,"%s: R extension block ignored\n",nfgif->name);
			  c = getc(f);
			  skipbytes(ppnf,c);
			  break;
			case 0x01 : /* text extension */
			  c = getc(f);
			  skipbytes(ppnf,c);
			  while((c = getc(f)) > 0)
			    skipbytes(ppnf,c);
			  break;
			case 0xf9 :  /* tranparency */
				if(debug_)
					fprintf(stderr,"%s: transparency extension block ignored\n",nfgif->name);
				skipblks(ppnf);
				break;
			case 0xfe:  /* comment */
				if(debug_)
					fprintf(stderr,"%s: comment extension block ignored\n",nfgif->name);
				skipblks(ppnf);
				break;
			default:
			  c = getc(f);
			  skipbytes(ppnf,c);
			}
			break;
		case ',': /* image separator */
			/* read image parameters */
			if(imext_read1(ppnf,tmp,9) != 9)
				imext_err(ppnf,9,"could not read GIF image format\n");
			ndx = tmp[4] + 256 * tmp[5];
			ndy = tmp[6] + 256 * tmp[7];
			if(nfgif->ndimx == 0) {
				nfgif->ndimx = ndx;
				nfgif->ndimy = ndy;
				nfgif->size = nfgif->ndimv * ndx * ndy * ((nfgif->nbits + 7) / 8);
				nfgif->ndimz = 1;
			} else {
				if((ndx != nfgif->ndimx) || (ndy != nfgif->ndimy)) {
					fprintf(stderr,"!!GIF multi-images : changement de dimension\n");
					return -1;
				}
				nfgif->ndimz++;
			}
			nfgif->interlace = tmp[8] & 0x40;
			if(tmp[8] & 0x80) { /* local colormap */
				if(nfgif->ndimz > 1)
					fprintf(stderr,"Warning, GIF local colormap\n");
				nfgif->numcols = 1 << ((tmp[8] & 7) + 1);
				rdcmap(ppnf);
			}
			nfgif->data = (unsigned char *)i_malloc(nfgif->size);
			if(pg->nb_alloc < nfgif->ndimz) {
				pg->nb_alloc += NB_PDATA;
				pg->pdata = (char **)i_realloc((void *)pg->pdata,pg->nb_alloc * sizeof(char *));
			}
			*(pg->pdata + nfgif->ndimz - 1) = (char *)nfgif->data;
			if(debug_)
				fprintf(stderr,"open GIF file %s, dimx %d, dimy %d, %d bits/pixel,%d colors, interlace: %d\n",
					nfgif->name,nfgif->ndimx,nfgif->ndimy,nfgif->nbits,nfgif->numcols,nfgif->interlace);
			return 0;
#if 0  /* LZWDecode laissait un octet */
		case 0 :
		case 1 :
		case 2 :
		case 0x10 :
			/* if(nfgif->flag && (nfgif->ndimz > 0))*/
				continue;
#endif
		default:
		  strncpy(tmp,nfgif->name,512);
		  imext_err(ppnf,9,"Bad info (0x%x '%c') in GIF file %s\n",c,c,tmp);
		}
	}

}
tst_read(pnf)
struct extimage **pnf;
{
  FILE *f;
  int c, i;
  int codesize;
  f = (FILE *)(*pnf)->file;
  codesize = getc(f);
  while((c = getc(f)) > 0)
    skipbytes(pnf,c);
}

readgif(pnf)
struct extimage **pnf;
{
	register struct extimage *nfgif;
	nfgif = *pnf;
	// tst_read(pnf);
	LZWDecode(pnf);
	/*	getc((FILE *)(*pnf)->file); /* il faut manger un octet suppl. */
	if(nfgif->interlace)
		un_interlace(nfgif);
}

/* image gif entrelacee :
lignes de 8 en 8 a partir de 0 (0,8,16),
puis de 8 en 8 a partir de 4 (4,12,20),
puis de 4 en 4 a partir de 2 (2, 6,10),
enfin de 2 en 2 a partir de 1 (1,3,5)
*/
static char line_incr[]= {8,8,4,2};
static char line_start[] = {0,4,2,1};
un_interlace(nfgif)
register struct extimage *nfgif;
{
	int incr[4], start[4];
	register i, j, nx, ny, dy, y, incp;
	register char *data2;
	register struct gif_ext *pg;
	register char *p1, *p2;
	nx = nfgif->ndimx;
	ny = nfgif->ndimy;
	pg = (struct gif_ext *) nfgif->x_params;
	data2 = (char *)i_malloc(nx * ny);
	for(i=0;i<4;i++) {
		incr[i] = line_incr[i] * nx;
		start[i] = line_start[i] * nx;
	}
	p1 = (char *)nfgif->data;
	for(j=0;j<4;j++) {
		y = line_start[j];
		p2 = data2 + start[j];
		dy = line_incr[j];
		incp = incr[j];
		while(y < ny) {
			memcpy(p2,p1,nx);
			p1 += nx;
			p2 += incp;
			y += dy;
		}
	}
	i_Free((char **)&nfgif->data);
	nfgif->data = (unsigned char *)data2;
	*(pg->pdata + nfgif->ndimz - 1) = data2;
}

gif_ecrinr(nf,ppnf)
struct image *nf;
struct extimage **ppnf;
{
	register struct extimage *pnf;
	register struct gif_ext *pg;
	char *pc;
	register i;
	pnf = *ppnf;
	pg = (struct gif_ext *)pnf->x_params;
	for(i=0;i< pnf->ndimz;i++) {
		pc = *(pg->pdata + i);
		c_ecr(nf,pnf->ndimy,pc);
	}
}
