/**
 * Convertion to Jpeg 2000 using Jasper
 * (c) Dominique Brziat 2006
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <jasper/jasper.h>
#include <inrimage/image.h>

extern int debug_;

char version[]= "1.0.0";
char command[]= "[file][jpeg-file][-c=%s][-o=%s][-r %f][-q %d]";
char details[]= "Encode a Jpeg 2000 image from inrimage.\n"
  "  -lc: list of codecs\n"
  "  -c=: codec (see -lc for available codecs)\n"
  "  -o=: pass option to encoder (see Jasper doc)\n"
  " Some aliases to important jasper decoder options are available:\n"
  "  -r %f: jpeg2000 rate compression,  << 1, 1=no compression\n"
  "  -q %d: jpeg compression quality, 0--100, 100=no compression\n"
  "     -D: print Inrimage traces\n"
  "     -d: print Jasper traces\n"
  ;


void listofcodec( void) {
  jas_image_fmtinfo_t *fmtinf;
  int i=0;
  while( (fmtinf = jas_image_lookupfmtbyid( i++)) != NULL) {
    fprintf( stderr, "codec %s: %s\n", fmtinf->name, 
	     fmtinf->desc);
  }
}

inline
void uchar2long( long *dest, unsigned char *src, size_t count, int v) {
  while( count --) {
    *dest ++ = *src;
    src += v;
  }
}

int select_codec_from_name( char *name, int codec) {
  char *p = strrchr( name, '.');
  jas_image_fmtinfo_t *fmtinf;

  if( p && ((fmtinf=jas_image_lookupfmtbyname( p+1)) != NULL))    
    return  fmtinf->id ;
  else
    return codec;
}
/**
 * retourne le nombre de lignes allouables (au plus dimy)
 * nelem : nombre d'lement par ligne
 * selem : taille d'un element
 */ 
int check_mem( size_t selem, size_t nelem, size_t dimy) {
  void *buf;
  
  do {
    buf = malloc( selem * nelem * dimy);
    if( buf) {
      free( buf);
      return dimy;
    }
    dimy /= 2;
  } while( dimy>1);
  return 1;
}

int main( int argc, char **argv) {
  struct image *nf;
  Fort_int lfmt[9];

  unsigned char *cbuf;
  float *fbuf;
  long  *lbuf;

  char name[256], cod[100];
  int v, y, z, ndimy_bl, ndimv;

  jas_stream_t *out;
  jas_image_t *image;

  int codec;
  char encopts[100] = "";
  float rate;

  jas_image_cmptparm_t *cmptp;

  inr_init( argc, argv, version, command, details); 
  jas_init();

  if( igetopt0( "-lc")) {
    listofcodec();
    return 0;
  }

  igetopt0("-d") && jas_setdbglevel(1);

  /* alias d'options */
  if( igetopt1( "-r", "%g", &rate))
    sprintf( encopts, "rate=%g ", rate);
  if( igetopt1( "-q", "%d", &z))
    sprintf( encopts, "%srate=%g ", encopts, z);
  
  if( igetopt0x( "-o=", "%s", name))
    sprintf( encopts, "%s %s ", encopts, name);
  

  infileopt( name);
  /* only 1-byte image are handled : for other format, use cco ! */
  TYPE = FIXE;  BSIZE = 1; EXP   = 0;
  nf = image_( name, "e", "t", (NF_fmt*)lfmt);

  if( NDIMZ>1 && NDIMV>1)
    imerror( 7, "format not supported");
  ndimv = NDIMV * NDIMZ;


  /* 
   * codec selection  
   */
  /* default codec */
  codec = jas_image_strtofmt("jp2");
  /* by filename */
  outfileopt( name);
  codec = select_codec_from_name( name, codec);
  /* by option */
  if( igetopt0x( "-c=", "%s", cod)) {
    if( jas_image_strtofmt(cod) >= 0)
      codec = jas_image_strtofmt(cod);
  } 
  if( debug_) fprintf( stderr, "Encoder codec: %s\n", jas_image_fmttostr(codec));

  out = (*name=='>')?jas_stream_fdopen(1,"w+b") : jas_stream_fopen( name, "w+b");

  cmptp = (jas_image_cmptparm_t*) malloc( sizeof( jas_image_cmptparm_t) * ndimv);

  for( v=0; v<ndimv; v++)  {
    jas_image_cmptparm_t cmptp_[1] = {0, 0, 1, 1, NDIMX, NDIMY, 8, 0 };
    memcpy( cmptp+v, cmptp_, sizeof( jas_image_cmptparm_t));
  }
  
  image = jas_image_create( ndimv, cmptp, (ndimv>1) ? JAS_CLRSPC_SRGB : JAS_CLRSPC_SGRAY);
 
  for( v=0; v<ndimv; v++)
    // jas_image_setcmpttype(image, v);
    image->cmpts_[v]->type_ = v;

  ndimy_bl = check_mem( sizeof(long)+sizeof(char)*NDIMV, NDIMX, NDIMY);
  if( debug_) fprintf( stderr, "alloc bloc of %d lines\n", ndimy_bl);

  lbuf = (long *)i_malloc(sizeof(long)*NDIMX*ndimy_bl);
  cbuf = (unsigned char*)i_malloc(sizeof(unsigned char)*DIMX*ndimy_bl);
  
  for( z=0; z<NDIMZ; z++) {
    for( y=0; y<NDIMY; y+=ndimy_bl) {
      if( y + ndimy_bl < NDIMY)
	c_lect( nf, ndimy_bl, cbuf);
      else
	c_lect( nf, NDIMY-y, cbuf);
      for( v=0; v<NDIMV; v++) {
	uchar2long( lbuf, cbuf+v, NDIMX*ndimy_bl, NDIMV);
	jas_image_writecmpt2( image, v+z, 0, y, NDIMX, ndimy_bl, lbuf);
      }
    }
  }

  jas_image_encode( image, out, codec, encopts);
  jas_stream_close( out);
  
  jas_image_destroy( image);
  fermnf_(&nf);
  
  return 0;
}
