/**
 * Convert Inrimage to PNG
 * Version alpha. Beaucoup de choses  rgler.
 * (c) Dominique Brziat 2006
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inrimage/image.h>
#include <png.h>

extern int debug_;

char version[] = "1.0.0";
char cmd[] = "[input][png-file]";
char details[] = "Convert an INRIMAGE image in PNG format";

int main( int argc, char **argv) {
  char name[256];
  int bit_depth, color_type, y, convert, has_palette = 0;

  /* Inrimage stuff */
  struct image *nf;
  Fort_int lfmt[9];

  /* Png stuff */
  FILE *fp;
  png_structp png_ptr;
  png_infop info_ptr;
  png_text *text;
  int num_comments;

  /* conversion */
  float* rowbuf_in;
  void* rowbuf_out;
  Fort_int icodi[2], icodo[2], iexpi, iexpo;


  inr_init( argc, argv, version, cmd, details);

  infileopt( name);
  nf = image_( name, "e", "", (void*)lfmt);

  /* 
   * Only fixed images are converted. Float images should
   * be converted by Inrimage tools. Any fixed bsize is
   * handled (by float access & conversion to the best size
   * supported by PNG).
   */
  if( TYPE == REELLE) {
    imerror( 4, 
	     "ERROR: float format are not supported. Please convert to a fixed format "
	     "using the cco command.\n");    
  }

  outfileopt( name);
  fp = *name == '>' ? stdout : fopen( name, "wb");
  if( !fp) {
    imerror( 8, "ERROR: outfile can not be opened.\n");
  }

  png_ptr = png_create_write_struct( PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if( !png_ptr) {
    fclose( fp);
    fermnf_( (void*)&nf);
    imerror( 8, "ERROR: memory allocation error.\n");
  }
  info_ptr = png_create_info_struct( png_ptr);
  if( !info_ptr) {
    png_destroy_read_struct( &png_ptr, NULL, NULL);
    fclose( fp);
    fermnf_( (void*)&nf);
    imerror( 8, "ERROR: memory allocation error.\n");
  }

  png_init_io( png_ptr, fp);

  convert = 0;

  switch( BSIZE) {
  case -4:
    bit_depth = 4;
    break;
  case -3:
    convert = 1;
    bit_depth = 4;
    /* ATTENTION -- MODIFICATION DE BSIZE */
    BSIZE = -4;
    fprintf( stderr, "Pixel bit depth (3) is not supported by PNG. Image is converted to"
	     " the nearest bit depth (4) allowed by PNG.\n");
    break;
  case -2:
    bit_depth = 2;
    break;
  case -1:
    bit_depth = 1;
    break;
  case 1:
    bit_depth = 8;
    break;
  case 2:
    bit_depth = 16;
    break;
  default:
    if( BSIZE < 0 && BSIZE > -16) {
      /* ATTENTION -- MODIFICATION DE BSIZE */
      BSIZE = (7-BSIZE)/8;
      fprintf( stderr, "Pixel bit depth (%d) is not supported by PNG. Image is converted to"
	       " the nearest bit depth (%d) allowed by PNG.\n", bit_depth, BSIZE*8);
      /* nb de bit pour que ca tienne */
      bit_depth = BSIZE * 8;

      convert = 1;
    } else if( BSIZE > 2 || BSIZE < -16) {
      imerror( 6, "ERROR: this format (%d bits/pixel) can not be converted to PNG.\n", 
	       BSIZE>0?BSIZE*8:-BSIZE);
    }
  }

  /* some warning about the conversion */
  if( EXP < 0) {
    fprintf( stderr, "WARNING: Signed images are not supported in PNG. Image is converted to unsigned.\n");
  }

  switch( NDIMV*NDIMZ) {
  case 1: /* case of GRAY image */
    color_type = PNG_COLOR_TYPE_GRAY;
    break;
  case 2: /* case of GRAY_ALPHA image */
    color_type = PNG_COLOR_TYPE_GRAY_ALPHA;
    break;
  case 3: /* case of RGB image */
    color_type = PNG_COLOR_TYPE_RGB;
    break;
  case 4: /* case of RGB_ALPHA image */
    color_type = PNG_COLOR_TYPE_RGB_ALPHA;
    break;
  default:
    imerror( 22, "ERROR: multiframe or vectoriel images are not supported by PNG.\n");
    break;
  }

  /* Palette color */
  {
    int i0, nb;
    unsigned char r[256], g[256], b[256];

    switch( ird_ctrgb( nf, &i0, &nb, (char*)r, (char*)g, (char*)b, 256)) {
    case -1: /* can not occur */
      break;
    case 0:  /* no color map */
      break;
    case 1:  /* the three tables are identicals */
    case 2:  /* two different tables */
    case 3:  /* three different tables */
      if( color_type == PNG_COLOR_TYPE_GRAY ||
	  color_type == PNG_COLOR_TYPE_GRAY_ALPHA ) {
	png_color palette[256];
	
	for( i0=0; i0<nb; i0++) {
	  palette[i0].red = r[i0];
	  palette[i0].green = g[i0];
	  palette[i0].blue = b[i0];
	}
	png_set_PLTE( png_ptr, info_ptr, palette, nb);
	color_type = PNG_COLOR_TYPE_PALETTE;
      }
    }    
  }

  png_set_IHDR( png_ptr, info_ptr, NDIMX, NDIMY, bit_depth, color_type,
		PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, 
		PNG_FILTER_TYPE_BASE);


  /* @FIXME : UNTESTED  */
  if( bit_depth==16 && ismsbfirst()==0) {
    /*    if( bit_depth < 8)
	  png_set_packswap( png_ptr);
	  else */
    png_set_swap( png_ptr);
    if( debug_) fprintf( stderr, "host is big-endian: image is swapped.\n");
  }

  /* Comments */
  {
    int line, i, len;
    char buf[256];

    line = 1;
    num_comments = 1; /* inclure la signature de inr2png */

    while( (line = igethline( nf, "#", line, buf, 255, &len)) != -1)
      num_comments ++;
    text = malloc( sizeof(png_text)*num_comments);
    
    text[0].compression = PNG_TEXT_COMPRESSION_NONE;
    text[0].key = "software";
    text[0].text = "inr2png";
    for( i = 1, line=1; i<num_comments; 
	 i++, line = igethline( nf, "#", line, buf, 255, &len)) {
      char tmp[50];
      
      text[i].compression = PNG_TEXT_COMPRESSION_NONE;
      if( strstr( buf, "*[H]*") == buf) {
	sprintf( tmp, "Historic %d", i);
	text[i].key = strdup( tmp);
	text[i].text = strdup(buf + 5);
      } else {
	sprintf( tmp, "Comment %d", i);
	text[i].key = strdup( tmp);
	text[i].text = strdup(buf);
      }
      if(debug_)
	fprintf(stderr, "key: %s text: %s\n", text[i].key, text[i].text);
    }
  }
  png_set_text( png_ptr, info_ptr, text, num_comments);
  
  /* ? peut-on librer ces ressources maintenant ou apres png_write_info() ? */
  for( y=1; y<num_comments; y++) {
    free( text[y].key);
    free( text[y].text);
  }
  free( text);

  /* @TODO Reorganize planar images ? */
  

  /* writing header */
  png_write_info( png_ptr, info_ptr);
  
  /* allocate one row */
  rowbuf_out = (void *) i_malloc(png_get_rowbytes( png_ptr, info_ptr));
  if( convert)
    rowbuf_in  = (float*) i_malloc(sizeof(float)*DIMX);

  icodi[0] = sizeof(float);
  icodi[1] = REELLE;
  icodo[0] = BSIZE;
  icodo[1] = PACKEE;
  iexpi = iexpo = 0;

  if( debug_)
    fprintf( stderr, "use %s access ...\n", convert ? "float":"direct");
  
  /* @TODO : acces par plan si possible */

  for( y = 0; y < NDIMY; y++) {
    if( convert) {
      c_lecflt( nf, 1, (void*)rowbuf_in);
      /* conversion */
      c_cnvtbg( (void*)rowbuf_in, (void*)rowbuf_out, DIMX, icodi, icodo, iexpi, iexpo);
    } else {
      c_lect( nf, 1, rowbuf_out);
    }

    png_write_rows(png_ptr, (png_bytepp)&rowbuf_out, 1);
  }

  png_write_end(png_ptr, info_ptr);
  fclose(fp);
  fermnf_( &nf);
  
  png_destroy_write_struct(&png_ptr, &info_ptr);
  
  if( convert) i_Free( (void*)&rowbuf_in);
  i_Free( (void*)&rowbuf_out);

  return 0;
}
