/**
 * A TIFF wrapper for Inrimage using libtiff
 * (c) Dominique Brziat 2006
 */
#include <stdlib.h>
#include <tiffio.h>
#include <inrimage/image.h>
#include <assert.h>

/**
 * @TODO switch to set photometric interpretion to RGB with -z 4 or -v 4.   
 * @TODO switch to write in Tiled format (and allow image with slices).
 */

#define SCANLINE 0
#define STRIP    1
#define TILE     2

static char version[] = "1.0.0";
static char command[] = "[input] output-tif [-c=%s][-q %d][-ct=%s] ";
static char details[] = "Options are:\n"
  "  -c= : set compression. Possible values are:\n"
  "         zip,lwz,packbits,jpeg\n"
  "   -q : quality for jpeg compression (00-100)\n"
  "-ct=%s: write color map (if exists) in a separate file\n"
  ;

int main( int argc, char **argv) {
  TIFF *tiff;
  struct image *nf;
  char name[256],compression[20], colormap[256];
  Fort_int lfmt[9], ibs;
  void *buf, *buf2;  int bitsPerSample;
  int l, z;

  int access = SCANLINE;

  inr_init( argc, argv, version, command, details);
  
  infileopt( name);
  nf = image_( name, "e", "", (NF_fmt*)lfmt);
  
  outfileopt( name);
  if( *name == '>') {
    fprintf( stderr, "inr2tiff:error:i can't write to stdout.\n");
    exit(1);
  }
  tiff = TIFFOpen( name, "w");
  TIFFSetField( tiff, TIFFTAG_SOFTWARE, "inr2tiff by D.Bereziat / LIP6 UMPC / INRIA (c) 2006");
		
  /*
   * Header 
   */

  TIFFSetField( tiff, TIFFTAG_IMAGEWIDTH,      NDIMX);
  TIFFSetField( tiff, TIFFTAG_IMAGELENGTH,     NDIMY);

  /* SI NDIMV = 1 et NDIMZ > 1 : PLANAR_SEPARATE 
   * SI NDIMV > 1 et NDIMZ = 1 : PLANAR_CONTIG
   * SINON on transforme les NDIMZ en Directories (ou alors on teste les tile ?)
   */
  if( NDIMV >= 1 && NDIMZ == 1) {
    TIFFSetField( tiff, TIFFTAG_SAMPLESPERPIXEL, NDIMV);
    TIFFSetField( tiff, TIFFTAG_PLANARCONFIG,    PLANARCONFIG_CONTIG);
  } else if( NDIMV == 1 && NDIMZ > 1) {
    TIFFSetField( tiff, TIFFTAG_SAMPLESPERPIXEL, NDIMZ);
    TIFFSetField( tiff, TIFFTAG_PLANARCONFIG,    PLANARCONFIG_SEPARATE);
  } else {
    /* @FIXME : use Directories */
    fprintf( stderr, "NDIMZ > 1 and NDIMV > 1 is not yet supported by inr2tiff\n");
    fermnf_( &nf);
    return 0;
  }

  switch( TYPE) {
  case REELLE:
    TIFFSetField( tiff, TIFFTAG_BITSPERSAMPLE, BSIZE * 8);
    TIFFSetField( tiff, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_IEEEFP);
    break;

  case FIXE:
    /* en codage FIXE, chaque valeur tiens sur un multiple de 8 bit (7-BSIZE)/8). 
     * Comme TIFF ne connat que le format PACKEE. */
    TIFFSetField( tiff, TIFFTAG_BITSPERSAMPLE, BSIZE > 0 ? BSIZE*8 : (7-BSIZE)/8*8 );
    TIFFSetField( tiff, TIFFTAG_SAMPLEFORMAT, EXP>0 ? SAMPLEFORMAT_UINT : SAMPLEFORMAT_INT);
    break;

  case PACKEE:
    TIFFSetField( tiff, TIFFTAG_BITSPERSAMPLE, BSIZE > 0 ? BSIZE*8 : -BSIZE);
    TIFFSetField( tiff, TIFFTAG_SAMPLEFORMAT, EXP>0 ? SAMPLEFORMAT_UINT : SAMPLEFORMAT_INT);
    break;
  }

  /* INCOMPLET */
  if( NDIMV*NDIMZ == 3)
    TIFFSetField( tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
  else
    TIFFSetField( tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);

  /* TODO Image a palette */
  if( igetopt0x("-ct=", "%s", colormap) ) {
    struct image *nf_cm;
    Fort_int cfmt[9];
    uint16 ri[256], gi[256], bi[256];
    nf_cm = image_( colormap, "e", " ", (NF_fmt*)cfmt);
    if( nf_cm->DIMY == 3 && nf_cm->DIMX == 256 && nf_cm->TYPE == FIXE && nf_cm->BSIZE == 1) {
      int i;
      unsigned char cm[256*3];

      c_lect( nf_cm, 3, (void*)cm);
      for( i=0; i<nf_cm->DIMX; i++) {
	ri[i] = cm[i] * 255;
	gi[i] = cm[i+nf_cm->DIMX] * 255;
	bi[i] = cm[i+nf_cm->DIMX*2] * 255;
      }
      TIFFSetField( tiff, TIFFTAG_COLORMAP, &ri, &gi, &bi);
      TIFFSetField( tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_PALETTE);
    } else
      imerror( 100, "%s is not a valid colormap\n", colormap);
    fermnf_( (void*)&nf_cm);
  } else {
    int i0, nb;
    unsigned char r[256], g[256], b[256];
   
    switch( ird_ctrgb( nf, &i0, &nb, (char*)r, (char*)g, (char*)b, 256)) {
    case -1: /* can not occur */
      break;
    case 0:  /* no color map */
      break;
    case 1:  /* the three tables are identical */
    case 2:  /* two tables different */
    case 3:  /* three tables differents */
      {
	uint16 ri[256], gi[256], bi[256];
	for( i0=0; i0<nb; i0++) {
	  ri[i0] = r[i0] * 255;
	  gi[i0] = g[i0] * 255;
	  bi[i0] = b[i0] * 255;
	}
	TIFFSetField( tiff, TIFFTAG_COLORMAP, &ri, &gi, &bi);
	TIFFSetField( tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_PALETTE);
      }
      break;
    }    
  }

  /*
   * Compression options
   */
  if( igetopt0x( "-c=", "%s", compression)) {
    if( !strcmp(compression, "lzw")) 
      TIFFSetField( tiff, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
    else if( !strcmp(compression, "jpeg")) {
      int quality ;
      TIFFSetField( tiff, TIFFTAG_COMPRESSION, COMPRESSION_JPEG);
      access = STRIP;
      if( igetopt1( "-q", "%d", &quality))
	TIFFSetField( tiff, TIFFTAG_JPEGQUALITY, quality);
    }
    else if( !strcmp(compression, "packbits")) 
       TIFFSetField( tiff, TIFFTAG_COMPRESSION, COMPRESSION_PACKBITS);
    else if( !strcmp(compression, "zip")) 
      TIFFSetField( tiff, TIFFTAG_COMPRESSION, COMPRESSION_ADOBE_DEFLATE);
  }
  

  switch( access) {
  case SCANLINE:
    buf = i_malloc( ligbit_( lfmt, &ibs));
    if( TYPE == FIXE && BSIZE < 0) 
      buf2 = i_malloc( ligbit_( lfmt, &ibs));
    else
      buf2 = NULL;

    for( z=0; z < NDIMZ; z ++)
      for( l=0; l < NDIMY; l++) {
	c_lect( nf, 1, buf);
	if( TYPE == FIXE && BSIZE < 0) {
	  /* conversion -b vers -b -p */
	  Fort_int icodi[] = { BSIZE, TYPE};
	  Fort_int icodo[] = { (7-BSIZE)/8, FIXE};
	  c_cnvtbg( buf, buf2, DIMX, icodi, icodo, EXP, EXP);
	} 
	TIFFWriteScanline( tiff, buf2?buf2:buf, l, z);
      }    
    break;

  case STRIP:
    {
      int is;
      uint32 numStrip = TIFFNumberOfStrips(tiff);
      uint32 sizStrip = TIFFStripSize(tiff);
      
      buf = i_malloc( ligbit_( lfmt, &ibs)*DIMY);
      
      c_lect( nf, DIMY, buf);      
      if( TYPE == FIXE && BSIZE < 0) {
	Fort_int icodi[] = { BSIZE, TYPE};
	Fort_int icodo[] = { (7-BSIZE)/8, FIXE};
	buf2 = i_malloc( ligbit_( lfmt, &ibs)*DIMY);
	c_cnvtbg( buf, buf2, DIMX*DIMY, icodi, icodo, EXP, EXP);       
      } else
	buf2 = NULL;
      for( is=0; is<numStrip; is ++)
	TIFFWriteEncodedStrip( tiff, is, buf2?buf2:buf, sizStrip);
    }

  case TILE:
    /* A FAIRE ?? */

    break;
  }
  
  fermnf_( &nf);
  TIFFClose( tiff);
  return EXIT_SUCCESS;
}
