/**
 * A jpeg 2000 wrapper for Inrimage using Jasper
 * (c) Dominique Brziat 2006
 */


#include <jasper/jasper.h>
#include <inrimage/image.h>


#define MIN(a,b) ((a<b)?a:b)
extern int debug_;

char version[] = "1.0.0";
char command[] = "[jpeg-file][file| file1 file2 ...|-i [-i]][-o=%s]";
char details[] = 
  "Decoding a Jpeg2000 image and write it in Inrimage format\n"
  "     -i: print only file information\n"
  "  -i -i: print more file information\n"
  "   file: write one image (one frame per components)\n"
  "  file1 file2 ...: write image components in seperate files\n"
  "  -o=: pass Jasper option to decoder (see Jasper documentation)\n"
  "  -cz: write RGB (NDIMZ=3) image (instead of NDIMV=3)\n"
  "     -D: print Inrimage traces\n"
  "     -d: print Jasper traces\n";

/**
 * @TODO 
 *       option debug de jasper
 *       option -i
 */


void set_fmt( Fort_int *lfmt, int x, int y, int z, int v) {
  NDIMX = x;
  NDIMY = y;
  NDIMV = v;
  NDIMZ = z;
  DIMX = NDIMX * NDIMV;
  DIMY = NDIMY * NDIMZ;
  BSIZE = 1;
  TYPE = FIXE;
  EXP = 0;
}

void print_fmt( Fort_int *lfmt) {
  //  fprintf( stderr, "colorspace : %d\n", jas_image_colorspace(image));
  /* JAS_IMAGE_CS_(RGB|YCBCR|GRAY) */
  Fort_int iopt=3;
  wrfmg_(lfmt,&iopt);
}


inline
void long2uchar( unsigned char *dest, long *src, size_t count, int v) {
  while( count --) {
    *dest = *src++;
    dest += v;
  }
}

/**
 * retourne le nombre de ligne allouable (au plus dimy)
 * nelem : nombre d'lement par ligne
 * selem : taille d'un element
 */ 
int check_mem( size_t selem, size_t nelem, size_t dimy) {
  void *buf;
  
  do {
    buf = malloc( selem * nelem * dimy);
    if( buf) {
      free( buf);
      return dimy;
    }
    dimy /= 2;
  } while( dimy>1);
  return 1;
}

int main( int argc, char **argv) {
  jas_stream_t *in;
  jas_image_t *image;
  char name[256];
  int    nname = 0;
  char **lname = NULL;
  Fort_int lfmt[9];
  jas_matrix_t *matrix;
  struct image *nf;
  char options[100] = "";
  
  long *lbuf;
  unsigned char *cbuf;
  int x,y,z,v;

  
  inr_init( argc, argv, version, command, details);
  jas_init();
  igetopt0("-d") && jas_setdbglevel(1);
  igetopt0x( "-o=", "%s", options);
  
  /* info ? */
  if( igetopt0( "-i")) {
    int decode;
    
    infileopt( name);
    do {
      in = (*name == '<') ? jas_stream_fdopen( 0, "rb") : jas_stream_fopen( name, "rb");
      decode = jas_image_getfmt(in);
      image = jas_image_decode(in, -1, options);
      printf( "-x %d -y %d -z %d\n", jas_image_width(image), jas_image_height( image),
	      jas_image_numcmpts(image));
      
      if( igetopt0( "-i")) {
	int v;

	printf( "colorspace: %d\n", jas_image_clrspc(image)); 
	printf( "bounding box: %d %d %d %d\n", jas_image_tlx(image),
		jas_image_tly(image),jas_image_brx(image),jas_image_bry(image)
		); 
	for ( v=0; v<jas_image_numcmpts(image); v++) {
	  printf( "composante %d: -x %d -y %d\n", v, 
		  jas_image_cmptwidth(image,v),
		  jas_image_cmptheight(image,v));
	  printf( "  sign: %d\n", jas_image_cmptsgnd(image,v));
	  printf( " codec: %s\n", jas_image_fmttostr( decode));
	  printf( "  type: %d\n", jas_image_cmpttype(image,v));
	}
      }
    } while( infileopt(name));
    return 0;
  } 

  /* opening & decoding */
  infileopt( name);
  if( *name == '<') 
    in = jas_stream_fdopen( 0, "rb");
  else
    in = jas_stream_fopen( name, "rb");
  
  if( !in) imerror( 10, "cannot open file\n");

  if( debug_) fprintf( stderr, "decoding image ...");
  image = jas_image_decode(in, -1, options);
  if( !image) imerror( 100, "Jasper Error during decoding image\n");  
  else if( debug_) fprintf ( stderr, "done.\n");

  /* analyse output */
  while( outfileopt( name)) {
    lname = (char**) realloc( lname, sizeof(char*)*(++nname));
    lname[nname-1] = strdup( name);
  }

  if( nname <= 1) {
    int ndimy_bl;
		       
    if( igetopt0( "-cz"))
      set_fmt( lfmt, jas_image_width( image), jas_image_height( image), 
	       jas_image_numcmpts( image), 1);
    else 
      set_fmt( lfmt, jas_image_width( image), jas_image_height( image), 
	       1, jas_image_numcmpts( image));
    
    /* check how many lines one can be allocated */
    ndimy_bl = check_mem( sizeof(long) + NDIMV*sizeof(unsigned char), NDIMX, NDIMY);    
    if( debug_) fprintf( stderr, "Writing bloc of %d lines\n", ndimy_bl);
        
    lbuf = (long *)i_malloc( sizeof(long)*NDIMX*ndimy_bl);
    cbuf = (unsigned char*)i_malloc( sizeof(unsigned char)*DIMX*ndimy_bl);
    
    nf = image_( nname?lname[0]:name, "c", " ", (NF_fmt*)lfmt);
    
    for( z = 0; z < NDIMZ; z ++) {
      for( y = 0; y < NDIMY; y += ndimy_bl) {
	for( v = 0; v < NDIMV; v ++) {
	  jas_image_readcmpt2( image, z+v, 0, y, NDIMX, ndimy_bl, lbuf);
	  long2uchar( cbuf+v, lbuf, NDIMX*ndimy_bl, NDIMV);
	}
	if( y + ndimy_bl < NDIMY)
	  c_ecr( nf, ndimy_bl, cbuf);
	else
	  c_ecr( nf, NDIMY-y, cbuf);
      }
    }

    i_Free( (void*)&lbuf);
    i_Free( (void*)&cbuf);
    fermnf_( &nf);
  } else {
    int ndimy_bl;
    
    /* write one image per component */
    for( z=0; z<MIN(jas_image_numcmpts( image),nname); z++) {
      if( debug_)
	fprintf( stderr, "writing component %d (type=%d)\n", z, image->cmpts_[z]->type_);
      
      set_fmt( lfmt, jas_image_cmptwidth(image,z), jas_image_cmptheight(image,z), 1, 1);   
      nf = image_( lname[z], "c", " ", (NF_fmt*)lfmt);
       
      /* check how many lines one can be allocated */
      ndimy_bl = check_mem( sizeof(long) + sizeof(unsigned char), NDIMX, NDIMY);    
      if( debug_) fprintf( stderr, "Writing bloc of %d lines\n", ndimy_bl);

      lbuf = (long *)i_malloc( sizeof(long)*NDIMX*ndimy_bl);
      cbuf = (unsigned char*)i_malloc( sizeof(unsigned char)*NDIMX*ndimy_bl);
      
      for( y = 0; y < NDIMY; y += ndimy_bl) {
	jas_image_readcmpt2( image, z, 0, y, NDIMX, ndimy_bl, lbuf);
	long2uchar( cbuf, lbuf, NDIMX*ndimy_bl, 1);
	if( y + ndimy_bl < NDIMY)
	  c_ecr( nf, ndimy_bl, cbuf);
	else
	  c_ecr( nf, NDIMY-y, cbuf);
      }
      i_Free( (void*)&lbuf);
      i_Free( (void*)&cbuf);
      fermnf_( &nf);
    }
  
  }  

  if( nname > 0) {
    for( z=0; z<nname; free( lname[z++]));
    free( lname);
  }

  jas_stream_close( in);
  jas_image_destroy( image);
  

  return 0;
}
