/**
 * Direct call of libjpeg to decode Jpeg and convert to Inrimage
 * (c) Dominique Bereziat 2006
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <jpeglib.h>
#include <inrimage/image.h>

#include <setjmp.h>

extern int debug_;

/* version : 1.1 is the old perl script version calling
 *           a patched version of djpeg.
 *           1.2 was a fix on script perl adressing djpeg options.
 *           2.0.0 is the new wrapper using directly libjpeg
 */

char version[] = "2.0.0";
char command[] = "[jpeg-file][outfile|-i][djpeg-like options]";
char details[] = 
  "Converts a Jpeg image to Inrimage format.\n"
  "        -i : do not convert but print informations found in input header.\n"
  "  -colors N: reduce to N colors (with colormap) \n"
  " -grayscale: reduce to 256 gray level image\n"
  "      -fast: combination of -nosmooth,-onepass,-dither none and -dct fast\n"
  " -dither none|ordered|fs: color dithering (needs -colors) method\n"
  "      fs=floyd steinberg, ordered color and none dithering (faster)\n"
  " -dct int|fast|float: select the DCT method; float is the slower\n"
  " -scale M/N: scale image with factor M/N\n"
  "  -nosmooth: don't use high-quality upsampling\n"
  "   -onepass: one pass in color quantification (needs -colors)\n"
  " -maxmemory N: maximum memory for jpeg decoding (in K-bytes)\n"
  ;

int main( int argc, char **argv) {
  FILE *fp;

  /* libjpeg stuff */
  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr jerr;

  int image_height;	/* Number of rows in image */
  int image_width;
  JSAMPARRAY buffer;	
  int row_stride;

  /* inrimage stuff */
  Fort_int lfmt[9], ibs;
  struct image *nf;
  void *ibuf;

  char name[256];
  int value;
  int write_colormap = 0;

  inr_init( argc, argv, version, command, details);
  infileopt( name);
  
  fp = *name == '<' ? stdin : fopen( name, "rb");

  if( fp == NULL) imerror( 1, "...");
  
  cinfo.err = jpeg_std_error(&jerr); /* obligatoire */
  jpeg_create_decompress(&cinfo);
  jpeg_stdio_src(&cinfo, fp);
  jpeg_read_header(&cinfo, TRUE);  
  

  if( igetopt0( "-i")) {
    jpeg_start_decompress(&cinfo);
    char *colorspace[] = {
      "unknown", "grayscale", "rgb", "YCbRc", "cmyk", "ycck"
    };
    printf( "%s -x %d -y %d -v %d\n", name, cinfo.output_width, cinfo.output_height,
	    cinfo.output_components);
    printf( "bit precision: %d\n", cinfo.data_precision);
    printf( "jpeg_color_space: %s\n", colorspace[cinfo.jpeg_color_space]);
    printf( "out_color_space: %s\n", colorspace[cinfo.out_color_space]);
    printf( "num_components: %d\n", cinfo.num_components);
    printf( "dct_method: %d\n", cinfo.dct_method);
  } else {

    /* switches libtiff */

    if( igetopt1( "-dct", "%s", name)) {
      if( !strcmp( name, "int"))
	cinfo.dct_method = JDCT_ISLOW;
      else if( !strcmp( name, "fast"))
	cinfo.dct_method = JDCT_IFAST;
      else if( !strcmp( name, "float"))
	cinfo.dct_method = JDCT_FLOAT;
    }


    if( igetopt1( "-colors", "%d", &cinfo.desired_number_of_colors) ||
	igetopt1( "-quantize", "%d", &cinfo.desired_number_of_colors)) {

      cinfo.quantize_colors = TRUE;
      cinfo.dither_mode = JDITHER_ORDERED;
      write_colormap = TRUE;

      if( igetopt1( "-dither", "%s", name)) {
	if( !strcmp( name, "none"))
	  cinfo.dither_mode = JDITHER_NONE;
	else if( !strcmp( name, "ordered"))
	  cinfo.dither_mode = JDITHER_ORDERED;
	else if( !strcmp( name, "fs"))
	  cinfo.dither_mode = JDITHER_FS;
      }
    
      cinfo.two_pass_quantize=igetopt0( "-onepass")?FALSE:TRUE;
    }

    if( igetopt0( "-grayscale")) 
      cinfo.out_color_space = JCS_GRAYSCALE;
    cinfo.do_fancy_upsampling = igetopt0( "-nosmooth") ? FALSE: TRUE;
    if( igetopt1( "-scale", "%s", name))
      sscanf( name, "%d/%d", &cinfo.scale_num, &cinfo.scale_denom);

    if( igetopt0( "-fast")) {
      /* alias de plusieurs options */
      cinfo.do_fancy_upsampling = FALSE;
      cinfo.two_pass_quantize = FALSE;
      cinfo.dct_method = JDCT_IFAST;
      cinfo.dither_mode = JDITHER_NONE;
    }
    if( igetopt1( "-maxmemory", "%d", &value))
      cinfo.mem->max_memory_to_use = value * 1024;

    /* decompression */
    jpeg_start_decompress(&cinfo);
    
    /* fillin lfmt */

    NDIMX = cinfo.output_width;
    NDIMY = cinfo.output_height;
    NDIMV = cinfo.output_components;
    NDIMZ = 1; 

    DIMX = NDIMX*NDIMV; DIMY = NDIMY*NDIMZ;
    BSIZE = - cinfo.data_precision; 
    TYPE = FIXE; EXP = 0;
    
    outfileopt( name);
    
    if( write_colormap  && cinfo.actual_number_of_colors > 16) {
      int hdr = 8;
      igetopt0x("-hdr","%d",&hdr);
      if( debug_) fprintf( stderr, "Setting header size to %d bytes\n", 256*hdr);
      set_hdr_min( hdr);
    }

    nf = image_ ( name, "c", " ", (NF_fmt*)lfmt);

    if( write_colormap )
      iwr_ctrgb( nf, 0, cinfo.actual_number_of_colors, (void*)cinfo.colormap[0], 
		 (void*)cinfo.colormap[1], (void*)cinfo.colormap[2]);
 
    row_stride = cinfo.output_width * cinfo.output_components;
    buffer = (*cinfo.mem->alloc_sarray)
      ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);
    
    while ( cinfo.output_scanline < cinfo.output_height) {
      jpeg_read_scanlines(&cinfo, buffer, 1);
      c_ecr( nf, 1, buffer[0]);
    }
    
    fermnf_( &nf);
    jpeg_finish_decompress(&cinfo);
  }
  
  jpeg_destroy_decompress(&cinfo);
  fclose( fp);
  return 0;
}
