/**
 * Decoder MPEG1/MPEG2 using libmpeg2.
 * written by D.Bereziat
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <assert.h>

#include <mpeg2.h>
#include <mpeg2convert.h>
#include <inrimage/image.h>


/*
 * Pour avoir accs aux nombres de plans
 * on doit dcoder deux fois la squence.
 * si la squence est lue depuis stdin, on doit
 * tamponner.
 */

int main( int argc, char **argv) {
  /* inrimage side */
  struct image *nf;
  Fort_int lfmt[9];
  char name[256];

  /* mpeg2dec side */
  FILE *fp;
  mpeg2dec_t *decoder;
  mpeg2_info_t *info;
  mpeg2_state_t state;
#define BUFFER_SIZE 4096
  uint8_t buffer[BUFFER_SIZE];

  /* common */
  size_t size, fullsiz = 0;
  int iz, dimz;
  int verbose;
  char *fullbuf = NULL;

  inr_init( argc, argv, "1.0.0", "[mpeg-file][outfile]", "Convert an MPEG1/2 sequence to Inrimage image.");
  verbose = igetopt0( "-V");

  infileopt( name);
  fp = *name == '<' ? stdin : fopen( name, "rb");
  assert( fp);

  /* opening mpeg flow */
  decoder = mpeg2_init ();
  assert( decoder);
  info = mpeg2_info (decoder);
  
  /* nom image de sortie */
  outfileopt( name);

  /* si on lit dans stdin et ecrit dans stdout, on doit tamponner */
  if( fp == stdin && *name == '>') {    
#if 0
    /* ce hack fonctionne, il reste a remplacer les fread() dans
       la suite du code par une fonction qui lit soit dans fp soit
       dans fullbuf */
    do {
      size = fread( buffer, 1, BUFFER_SIZE, fp);
      fullsiz += size;
      fullbuf = realloc( fullbuf, fullsiz);
      if( fullbuf == NULL)
	break; /* echec, pas assez de memoire */
      memcpy( fullbuf + (fullsiz-size), buffer, size);
    } while( size);
    
    if( fullbuf) fprintf( stderr, "sequence tamponnee\n");
#else
    /** @FIXME */
    fprintf( stderr, "Warning: impossible to know the number of frame.\n");
#endif
  }

  /** NOMBRE DE PLANS **/
  if( fp != stdin) {
    dimz = 0;
    size = -1;
    do {
      state = mpeg2_parse( decoder);
      switch( state) {
      case STATE_BUFFER: 
	size = fread( buffer, 1, BUFFER_SIZE, fp);
	mpeg2_buffer( decoder, buffer, buffer + size);
	break;
    case STATE_SLICE:
      case STATE_END:
      case STATE_INVALID_END:
	if(info->display_fbuf) dimz ++;
      break;
      }
    } while (size);
    fseek( fp, 0, SEEK_SET);
    if( verbose) fprintf( stderr, "I found %d frames\n",dimz);
  } else
    dimz = 999;
  /****/

  iz = 0;
  size = -1;
  do {
    state = mpeg2_parse( decoder);
    if( verbose) 
      fprintf( stderr, "TRACE: %d\n", state);
    switch( state) {
    case STATE_BUFFER: 
      size = fread( buffer, 1, BUFFER_SIZE, fp);
      mpeg2_buffer( decoder, buffer, buffer + size);
      break;
    case STATE_SEQUENCE:
      mpeg2_convert (decoder, mpeg2convert_rgb24, NULL);
      break;
    case STATE_SLICE:
    case STATE_END:
    case STATE_INVALID_END:
      if(info->display_fbuf) {
	static int write_header = 1;
	
	if( write_header) {
	  /* Create INRIMAGE Header and write it */
	  NDIMX = info->sequence->width;
	  NDIMY = info->sequence->height;
	  NDIMV = 3;
	  NDIMZ = dimz;
	  DIMX = NDIMX * 3;
	  DIMY = NDIMY * NDIMZ;
	  EXP  = 200;
	  BSIZE = 1;
	  TYPE = FIXE;
	  nf = image_( name, "c", " ", (NF_fmt*) lfmt);
	  if(verbose)
	    fprintf( stderr, "Decoding a %dx%d sequence ...\n", 
		     info->sequence->width, info->sequence->height);
	  write_header = 0;
	}
	iz ++;
	if( verbose) fprintf( stderr, "Writing frame %d\n", iz);
	c_ecr( nf, NDIMY, info->display_fbuf->buf[0]);
      }
      break;
    default:
      break;
    }
  } while (size);
  
  mpeg2_close (decoder); 
  fermnf_( &nf);
  fclose( fp);

  if( *name != '>' && dimz == 999) {
    /* adjust the right NDIMZ value if needed */
    char cmd[256];
    sprintf( cmd, "ifmt -z %d %s", iz, name);
    system( cmd);
  }
 
  return 0;
}
