#include <stdio.h>
#include <inrimage/image.h>
#include "sunras.h"

extern errno;
extern debug_;

extern struct extimage *openimext();

#define swp4(x,y) { x = y;\
y = (x << 24) | ((x & 0xff00) << 8) | ((x >> 8) & 0xff00) | ((x >> 24) & 0xff);}
struct extimage *opensunras(name)
char *name;
{
	FILE *f;
	register flg_swap, rs;
	register x, lg, n;
	register unsigned char *pc;
	struct extimage *nfext;
	struct sunras_hdr *pb;
	nfext = openimext(name,0);
	flg_swap = ismsbfirst() ? 0 : 1;
	lg = sizeof(struct sunras_hdr);
	pb = (struct sunras_hdr *)i_calloc(1,lg);
	nfext->x_params = (void *)pb;
	n = 0;
	/* read header */
	if(imext_read1(&nfext,pb,lg) != lg)
		imext_err(&nfext,9,"rasterfile short header\n");
	if(flg_swap) {
		int4 *pl;
		lg /= 4;
		pl = (int4 *)pb;
		while(lg-- > 0) {
			swp4(rs,*pl);
			pl++;
		}
	}
	if(pb->rst_magic != RST_MAGIC)
		imext_err(&nfext,9,"Not a raster file\n");
	if((pb->rst_codage < RST_COD_MIN) || (pb->rst_codage > RST_COD_MAX))
		imext_err(&nfext,9,"unknown raster file type\n");
	n = pb->rst_bits;
	nfext->nbits = n == 1 ? 1 : 8;
	nfext->ndimz = 1;
	nfext->ndimv = 1;
	nfext->ndimy = pb->rst_dimy;
	nfext->ndimx = pb->rst_dimx;
	if(n > 8) {
		if((n != 24) && (n != 32))
			imext_err(&nfext,9,"bad raster file : incorrect bit size\n");
		n /= 8;
		nfext->ndimv = n;
		/*		nfext->ndimx /= n; */
	}
	if(pb->rst_lgr > 0)
		nfext->size = pb->rst_lgr;
	else
		nfext->size = (pb->rst_dimx * pb->rst_dimy * pb->rst_bits) / 8;

	/* color map */
	if((x = pb->rst_ct_lgr) > 3*256)
		imext_err(&nfext,9,"color table larger than 3*256\n");
	if(pb->rst_ct_type == RST_RAWCT) {
		imext_err(&nfext,9,"unexpected RAW color table \n");
	} else if(pb->rst_ct_type == RST_RGBCT) {
		x /= 3;
		nfext->numcols = x;
		imext_read1(&nfext,nfext->rbuf,x);
		imext_read1(&nfext,nfext->gbuf,x);
		imext_read1(&nfext,nfext->bbuf,x);
	} else if(pb->rst_ct_type != RST_NOCT)
		imext_err(&nfext,9,"unknown color table type\n");
	/* alloc data buffers */
	nfext->data = (unsigned char *)i_malloc(nfext->size + 4);
/*	nfext->cindex = (unsigned char *)i_malloc(pb->lgline); */
	if(debug_) {
		fprintf(stderr,"open RASTER file %s,%d x %d (%d bits), type %d\n",
			nfext->name,nfext->ndimx,nfext->ndimy,pb->rst_bits,pb->rst_codage);
	}
	return nfext;
}

/*  swap red component with blue component */
invert_3v(buf, size)
unsigned char *buf;
int size;
{
  unsigned char swp;
  while ( size) {
    swp = buf[0];
    buf[0]=buf[2];
    buf[2]=swp;
    buf += 3;
    size --;
  }
}
/* read data */
readsunras(ppnf)
struct extimage **ppnf;
{
	register struct extimage *pnf;
	register struct sunras_hdr *pb;
	register unsigned char *buf;
	register n;
	pnf = *ppnf;
	pb = (struct sunras_hdr *)pnf->x_params;
	buf = pnf->data;
	if(pb->rst_codage != RST_RLE) {
	  if((pnf->ndimx * pnf->ndimv) & 1) {
	    /* il y a un nb pair d'octets par ligne */
	    readrst_odd(ppnf);
	  } else {
	    if((n = imext_read1(ppnf,buf,pnf->size) != pnf->size))
	      imext_err(ppnf,1,"raster file read err : %d bytes\n",n);
	  }
	} else {  /* RLE */
	  if(pnf->nbits != 8)
	    imext_err(ppnf,9,"RLE unimplemented for 1 bit raster files\n");
	  readrst_rle(ppnf);
	}
	if(pnf->ndimv == 3)
	  invert_3v(pnf->data,pnf->ndimx*pnf->ndimy);
}

rst_readline(ppnf,buf,n)
     struct extimage **ppnf;
     unsigned char *buf;
     int n;
{
  int i;
  while(n > 0) {
    if((i = fread(buf,1,n,(*ppnf)->file)) < 0) {
      fermimext(ppnf);
      imerror(1,"read err %d\n",errno);
    }
    if(i == 0) imerror(9,"Missing data in rasterfile\n");
    buf += i;
    n -= i;
  }
}

readrst_odd(ppnf)
     struct extimage **ppnf;
{
  unsigned char *buf;
  int n, lgl,lgl1;
  n = (*ppnf)->ndimy;
  buf = (*ppnf)->data;
  lgl = (*ppnf)->ndimx * (*ppnf)->ndimv;
  lgl1 = (lgl & 1) ? lgl + 1 : lgl;
  while(n-- > 0) {
    rst_readline(ppnf,buf,lgl1);
    buf += lgl;
  }

}

readrst_rle(ppnf)
struct extimage **ppnf;
{
	register struct extimage *pnf;
	register unsigned char *buf;
	FILE *f;
	register c, i, j, v,  lgl;
	pnf = *ppnf;
	buf = pnf->data;
	f = pnf->file;
	j = 0;
	lgl = pnf->ndimx * pnf->ndimv *  pnf->ndimy;
	while((c = getc(f)) != EOF) {
	  if(c == 0x80) { /* repetition */
	    c = getc(f);
	    if(c == 0) {*buf++ = 0x80;continue;}
	    if(c == EOF) break;
	    v = getc(f);
	    for(i=0;i<=c;i++)
	      *buf++ = v;
	    j += c;
	    if(j > lgl) imext_err(ppnf,9,"ERR\n");
	    continue;
	  } else {
	    *buf++ = c;
	    j++;
	  }
	}
}

