 /**
 * Wrapper Inrimage TIFF
 * (c) 2006 Dominique Brziat
 *
 * CODAGE DES PIXELS
 * TIFF supporte les codages flottant (IEEE) en simple et double prcision
 * (tag TIFFTAG_SAMPLEFORMAT = SAMPLEFORMAT_IEEEFP).
 * Le codage fixe sign est prvu aussi (tag TIFFTAG_SAMPLEFORMAT = SAMPLEFORMAT_INT).
 * On peut aussi coder au bit prs dans l'image fichier (dans un format 'Packe')
 * Un pixel peut tre organis en composantes (TIFFTAG_SAMPLEPERPIXEL). Attention,
 * l'organisation des composantes est soit contigue 
 * (TIFFTAG_PLANARCONFIG=PLANARCONFIG_CONTIG) soit par plans spars 
 * (TIFFTAG_PLANARCONFIG=PLANARCONFIG_SEPARATE).
 * Finalement, tous ces codages sont pris en compte par Inrimage, donc la 
 * conversion sera toujours possible. 
 * La libtiff (comme la libinrimage) gre le codage petit et grand indian.
 *
 * DEUX MODES D'ACCES
 * Il existe deux modes d'accs : scanline ou strip.
 * Scanline est le plus simple et le plus commun. strip est celui utilise par Metamorph
 * par exemple. Apparament, on peut choisir diffremment l'ou ou l'autre
 * systme d'accs, mais certain encodage (au niveau de la compression) interdise
 * l'accs en scanline. L'accs strip semblerait donc le plus universel.
 *
 * IMAGES TILED
 * Certaines image peuvent tre organises en tuiles (l'image est divise en
 * rgions de mme taille, une tuile). Dans ce mode, il est prvu un codage
 * multiplan (slice en ref  l'image mdicale). Le nombre de plans
 * est donne par le tag TIFFTAG_IMAGEDEPTH (et TIFFTAG_TILEDEPTH). Ces tags
 * sont optionnels, il faut tester leur prsence. 
 * La fonction TIFFIsTiled() indique si l'image est dans ce mode.
 * L encore, les images 'Tiled' peuvent tre lues en scanline ou en strip.
 * 
 *
 * DIRECTORIES
 * Un fichier tiff peut contenir plusieurs images (directories) de taille
 * et codage diffrent. A priori, il faut donc les dcoder en plusieurs
 * images. On peut prevoir une option qui choisisse une image  dcoder (-id %d)
 * et mme une srie d'image (-d %d) en ajoutant un index numerique aux
 * images dcodes.
 * On peut prvoir aussi le cas d'images monoplans ayant le mme codage. Dans
 * ce cas, les directories peuvent jouer le rle de multiplan. On peut prvoir
 * cette option (le cas a dj t rencontr, et c'est ce que fait ImageJ).
 *
 *
 * TAGS INFORMATIFS
 * Les tags relatifs aux conditions d'acquisitions peuvent tre placs dans l'entte
 * d'Inrimage (faire une liste de tags connus ? autoriser par une option
 * les tags a conserver ?).
 *
 *
 *
 * INRIMAGE versus TIFF
 * Inrimage est bien antrieur aux versions tendu de Tiff mais ils ont
 * des caractristiques assez quivalentes.
 *
 * Ce que Inrimage n'a pas : * les directories. Est-ce bien utile ? Une
 * interface via tarball serait tout a fait quivalent et facile  mettre
 * en oeuvre.
 *                           * compression des donnes. Mais Inrimage utilise
 * un compresseur externe (au choix). Ce qui s'implifie tout et le rend facilement
 * adaptable aux nouveauts.
 *                           * pas d'accs par Tile ou par Strip. Inrimage ne
 * connat que la scanline ce qui a le mrite de la simplicit de la mise en oeuvre.
 *
 * 
 *
 * $Id: tif2inr.c,v 1.4 2006/08/14 15:17:58 bereziat Exp $
 */

/** TODO 
 * - eliminate the -iz 4 or -iv 4 with Photointerpretation = RGB ?
 * - TIFFRGBAImageOriented() works correctly from version 2003-12-26 (3.6.1)
 */

#include <tiff.h>
#include <tiffio.h>
#include <inrimage/image.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#define MIN(a,b)  ((a)<(b)?(a):(b))
extern int debug_;
int quiet;

void minmax( double *buf, int size) {
  double min, max;
  min = max = *buf;
  while( size --) {
    if( *buf < min) min = *buf;
    if( *buf > max) max = *buf;
    buf ++;
  }
  fprintf( stderr, "%f %f\n", min, max);
}

void tiff_warning(const char* module, const char* fmt, va_list ap) {
  if( !quiet) {
    fprintf( stderr, "tiff2inr:warning:%s:",module);
    vfprintf( stderr, fmt, ap);
    fprintf( stderr, "\n");
  }
}

#define USE_RGB 0  /* Use TIFFRGBAImage() facilities access */
#define USE_STD 1  /* Use standard libtiff access */

/**
 * Read format (coding and dimensions) from a tiff file and build 
 * an appropriate Inrimage Format.
 *
 * Fixe coding are supposed packed.
 *
 * @param lfmt Inrimage format array filled
 * @return if 0, use TIFFRGB routines to load image.
 */

int read_lfmt( TIFF *tiff, Fort_int lfmt[9]) { 
  uint32 ui32;
  uint16 ui16;
  int ret;

  TIFFGetField( tiff, TIFFTAG_IMAGEWIDTH, &ui32);
  NDIMX = ui32;

  TIFFGetField( tiff, TIFFTAG_IMAGELENGTH, &ui32);
  NDIMY = ui32;

  /* following tags are optional */
  NDIMZ = ( TIFFGetField( tiff, TIFFTAG_IMAGEDEPTH, &ui32))?ui32:1;
  NDIMV = ( TIFFGetField( tiff, TIFFTAG_SAMPLESPERPIXEL, &ui16))?ui16:1;  

  TYPE = FIXE;
  TIFFGetField( tiff, TIFFTAG_BITSPERSAMPLE, &ui16);
  switch( ui16) {
  case  8: BSIZE = 1; break;
  case 16: BSIZE = 2; break;
  case 24: BSIZE = 3; break; /* ? */
  case 32: BSIZE = 4; break; 
  case 64: BSIZE = 8; break;
  default : 
    BSIZE = -ui16;
    TYPE = PACKEE;
  }

  TIFFGetField( tiff, TIFFTAG_SAMPLEFORMAT, &ui16);
  switch( ui16) {
  case SAMPLEFORMAT_UINT:
    EXP = 200;
    break;
  case SAMPLEFORMAT_INT:
    EXP = -200; 
    break;
  case SAMPLEFORMAT_IEEEFP:
    EXP = 0;
    TYPE = REELLE;
    break;
  default:
    /* non support (voir doc tiff), */
    if( debug_) fprintf( stderr, 
			 "Warning : unsupported sample format (%d). "
			 "Suppose format is fixed (or packed).\n", ui16);
    //exit(0);
    EXP = 200;
    break;
  }

  /* Si l'image n'est pas tiled et si les composantes 
   * ne sont pas stockes contiguement (en plan spar), 
   * je n'arrive pas  rorganiser l'image en plan contigue
   * pour les codages en bit (en octet, pas de probleme).
   *
   * Rem : Il existe les fonctions c_mvbyte() et c_mvbits() dans 
   * la libinrimage qui pourrait nous aider.
   *
   * A priori, ces images sont a ImageDepth=1 (pas de profondeur Z)
   * donc on garde les composantes en plan spar (NDIMZ>1) et
   * il faut aussi que NDIMV=1 (voir read_and_write_strip_image()). 
   * Par exemple, une image RGB  plan spar sera code en Inrimage avec -z 3.
   */

  if( !TIFFIsTiled(tiff)) {
    TIFFGetField( tiff, TIFFTAG_PLANARCONFIG, &ui16);
    if( ui16 == PLANARCONFIG_SEPARATE) {
      int d = NDIMV;
      NDIMV = NDIMZ;
      NDIMZ = d;
    }
  }

  /*
   * Depending on PhotometricInterpretation TA, check if image can be 
   * loaded directly (ret = 1) or should be loaded via TIFFRGBA interface
   * (ret = 0).
   */

  TIFFGetField( tiff, TIFFTAG_PHOTOMETRIC, &ui16);
  switch( ui16) {
  case PHOTOMETRIC_MINISWHITE:
  case PHOTOMETRIC_MINISBLACK:
  case PHOTOMETRIC_RGB:
  case PHOTOMETRIC_PALETTE:
  case PHOTOMETRIC_LOGL: 
  case PHOTOMETRIC_SEPARATED:  /* ? i guess it is CMYK*/
    ret = USE_STD;
    break;
    
  case PHOTOMETRIC_YCBCR:
  case PHOTOMETRIC_CIELAB:
#if TIFFLIB_VERSION >= 20020326
  case PHOTOMETRIC_ICCLAB:
#endif
  case PHOTOMETRIC_ITULAB:
  case PHOTOMETRIC_LOGLUV:
    ret = USE_RGB;
    break;
  
  default:
    ret = 1;
  }

  DIMX = NDIMX * NDIMV;
  DIMY = NDIMY * NDIMZ;

#if TIFFLIB_VERSION < 20031226
  fprintf( stderr, 
	   "tiff2inr has been compiled with a version of libtiff which can not\n"
	   "handled this type of image. Please compile tiff2inr with a version \n"
	   "of libtiff prior to 3.6.1\n");
  exit( 1);
#endif
  return ret;
}

/**
 * Test if two image have same format
 */

int test_lfmt( Fort_int lfmt[], Fort_int gfmt[]) {
  int i;
  for( i=0; i<9 && lfmt[i] == gfmt[i]; i++);
  return (i==9);
}

/**
 * Check if each directories have the same coding 
 */

int check_directories( TIFF *tiff, Fort_int lfmt[]) {
  Fort_int gfmt[9];
  read_lfmt( tiff, lfmt);
  do
    read_lfmt( tiff, gfmt);
  while( TIFFReadDirectory( tiff) && test_lfmt( lfmt, gfmt));
  return test_lfmt(lfmt,gfmt);
}

void display_others_stats( TIFF *tiff) {

  uint32 ui32;
  uint16 ui16;
  char *str;

  /* stats for debugging */
  if( TIFFGetField( tiff, TIFFTAG_IMAGEDEPTH, &ui32))
    fprintf( stdout, "ImageDepth present : %d\n", ui32);
  TIFFGetField( tiff, TIFFTAG_PLANARCONFIG, &ui16);
  fprintf( stdout, "Planar Config : %s\n", ui16==1?"contig":"separate");
  fprintf( stdout, "IsTiled : %s\n", TIFFIsTiled(tiff)?"yes":"no");
  if( TIFFIsTiled(tiff)) {
    fprintf( stdout, " | TileSize: %d\n", TIFFTileSize (tiff));
    TIFFGetField( tiff, TIFFTAG_TILEWIDTH, &ui32);
    fprintf( stdout, " | TileWidth: %d\n", ui32);
    TIFFGetField( tiff, TIFFTAG_TILELENGTH, &ui32);
    fprintf( stdout, " | TileLength: %d\n", ui32);
    if( TIFFGetField( tiff, TIFFTAG_TILEDEPTH, &ui32))
      fprintf( stdout, " | TileDepth: %d\n", ui32);
  } else {
    fprintf( stdout, "Info about strips :\n");
    fprintf( stdout, " | StripSize: %d\n", TIFFStripSize( tiff));
    fprintf( stdout, " | NumberOfStrips: %d\n", TIFFNumberOfStrips( tiff));
    TIFFGetField( tiff, TIFFTAG_ROWSPERSTRIP, &ui32);
    fprintf( stdout, " | RowsPerStrip: %d\n", ui32);
  }
  fprintf( stdout, "ScanlineSize: %d\n", TIFFScanlineSize (tiff));    

  TIFFGetField( tiff, TIFFTAG_BITSPERSAMPLE, &ui16);
  fprintf( stdout, "BitsPerSample: %d\n", ui16);

  ui32 = TIFFGetField( tiff, TIFFTAG_SAMPLESPERPIXEL, &ui16);
  fprintf( stdout, "SamplesPerPixel: %d\n", ui32?ui16:1);
  
  {
    char *photo[] = {
      "WhiteiZero",
      "BlackIsZero",
      "RGB",
      "Palette color",
      "Transparency Mask",
      "Seperated (CMYK)",
      "YCbCr",
      "CIE L*a*b",
      "ICC L*a*b",
      "ITU L*a*b",
      
      "CIE Log2(L)",
      "CIE Log2(L) (u',v')"
    };
    TIFFGetField( tiff, TIFFTAG_PHOTOMETRIC, &ui16);
    if( ui16 == PHOTOMETRIC_LOGL) ui16 = 10;
    else if (ui16 == PHOTOMETRIC_LOGLUV) ui16 = 11;
    
    if( ui16 < 12)
      fprintf( stdout, "Photometric Interpretation: %s\n", photo[ui16]);
    else
      fprintf( stdout, "Photometric Interpretation: unknown (%d)\n", ui16);
  }

  /* informative tags */
  if( TIFFGetField( tiff, TIFFTAG_ARTIST, &str)) 
    fprintf( stdout, "   Artist : %s\n", str);
  if( TIFFGetField( tiff, TIFFTAG_COPYRIGHT, &str)) 
    fprintf( stdout, "Copyright : %s\n", str);
  if( TIFFGetField( tiff, TIFFTAG_DATETIME, &str)) 
    fprintf( stdout, "Date Time : %s\n", str);
  if( TIFFGetField( tiff, TIFFTAG_COPYRIGHT, &str)) 
    fprintf( stdout, "Copyright : %s\n", str);
  if( TIFFGetField( tiff, TIFFTAG_DOCUMENTNAME, &str)) 
    fprintf( stdout, "Document name : %s\n", str);
  if( TIFFGetField( tiff, TIFFTAG_HOSTCOMPUTER, &str)) 
    fprintf( stdout, "Host computer : %s\n", str);
  if( TIFFGetField( tiff, TIFFTAG_IMAGEDESCRIPTION, &str)) 
    fprintf( stdout, "Image description : %s\n", str);
  if( TIFFGetField( tiff, TIFFTAG_SOFTWARE, &str)) 
    fprintf( stdout, "Software : %s\n", str);

} 

/**
 * Alloc to store lines taking in account each line
 * should be a multiple of 1 byte (for PACKEE format).
 */

char *inr_alloc_lines( Fort_int lfmt[], int lines) {
  size_t amount;
  
  switch( TYPE) {
  case FIXE:
  case REELLE:
    amount = DIMX*BSIZE; 
    break;
  case PACKEE:
    assert( BSIZE<0); /* normallement, toujours vrai */
    amount = (DIMX * -BSIZE + 7)/ 8 ;
    break;
  }
  return i_malloc( amount * lines);
}

/**
 * Use a high level function from libtiff to decode image with non standard
 * format. Probably it does not work for all cases.
 * I use it to
 *   - convert all tif image into a RGB format (usefull for visualisation)
 *   - convert image encoded with YCK and similar ...
 */

void read_and_write_rgb_image( TIFF *tiff, struct image *nf, Fort_int *lfmt) {
  char *rgba, *prgba, *rgb, *prgb;
  size_t size = nf->NDIMX * nf->NDIMY;

  if( lfmt == NULL) lfmt = nf->lfmt;
  
  assert( NDIMV == 3);
  assert( TYPE == FIXE);
  assert( BSIZE == 1);

  rgba = i_malloc( size * 4);
  rgb = i_malloc( size * 3);
#if TIFFLIB_VERSION >= 20031226
  TIFFReadRGBAImageOriented( tiff, NDIMX, NDIMY, (uint32*)rgba, ORIENTATION_TOPLEFT, 0);
#else
  imerror( 100, "Fatal error : TIFFReadRGBAImageOriented not supported\n");
#endif
  for( prgb = rgb, prgba = rgba; size; size --) {
    *prgb++ = *prgba++;
    *prgb++ = *prgba++;
    *prgb++ = *prgba++;    
    prgba++;
  } 
  i_Free( (void*)&rgba);
  
  c_ecr( nf, NDIMY, (void*)rgb);
  i_Free( (void*)&rgb);

}

/**
 * Read a Tiled image (incomplet).
 */

void read_and_write_tiled_image( TIFF *tiff, struct image *nf, Fort_int *lfmt) {
  uint32 tileWidth, tileLength, tileDepth;
  uint16 planar_config;
  int x, y, z, i, j, k, v;
  char *tile, *buf;
  uint16 bsize;

  if( lfmt == NULL) lfmt = nf->lfmt;

  TIFFGetField( tiff, TIFFTAG_BITSPERSAMPLE, &bsize);
  if( bsize % 8) {
    fprintf( stderr, "tiff2inr:error:I dont know process tiled image with %d bits per sample\n",
	     bsize);
    exit(1);
  }
  bsize /= 8;
  TIFFGetField( tiff, TIFFTAG_PLANARCONFIG, &planar_config);  

  if( planar_config == PLANARCONFIG_CONTIG) {
    tile = (char *)i_malloc(TIFFTileSize( tiff));
    buf = inr_alloc_lines( nf->lfmt, DIMY);

    TIFFGetField( tiff, TIFFTAG_TILEWIDTH, &tileWidth);
    TIFFGetField( tiff, TIFFTAG_TILELENGTH, &tileLength);
    TIFFGetField( tiff, TIFFTAG_TILEDEPTH, &tileDepth);
    
    for (z = 0; z < NDIMZ; z += tileDepth)
      for (y = 0; y < NDIMY; y += tileLength)
	for (x = 0; x < NDIMX; x += tileWidth) {
	  
	  TIFFReadTile( tiff, tile, x, y, z, 0);
	  
	  /* insert tile into the main buffer */
	  for( k = 0; k < tileDepth; k++)
	    for( j = 0; j < tileLength; j++)
	      for( i = 0; i < tileWidth; i++)
		// for( v=0; v<NDIMV; v ++)
		if( i+x < NDIMX && j+y < NDIMY && k+z < NDIMZ) 
		  /* @TODO : not very efficiently, should be enhanced : one can copy the whole line */
		  memcpy( buf  + bsize*((i+x)*NDIMV+((j+y)+(k+z)*NDIMY)*DIMX), 
			  tile + bsize*((i+(j+k*tileDepth)*tileWidth)*NDIMV), bsize*NDIMV);
	}
    
    c_ecr( nf, DIMY, (void*)buf);
  } else {
    /* PLANARCONFIG_SEPARATE, FIXME */
    fprintf( stderr, " TILED IMAGE WITH PLANAR SEPARATE NOT YET SUPPORTED\n");
    exit (1);
    
  }

  i_Free( (void*)&tile);
  i_Free( (void*)&buf);
} 

/**
 * Read an image by Strip access.
 * Strip access is prefered to scanline access because it is more efficient
 * and it supports more compression format.
 */

void read_and_write_strip_image( TIFF *tiff, struct image *nf, Fort_int *lfmt) {  
  uint16 planar_config;
  char *buf, *strip;
  uint16 bitsPerSample;
  uint32 numberOfStrips;
  int  stripSize, readSize;

  tstrip_t is, iy;
  int    inrlineSize;

  if( lfmt == NULL) lfmt = nf->lfmt;
  
  TIFFGetField( tiff, TIFFTAG_BITSPERSAMPLE, &bitsPerSample);
  TIFFGetField( tiff, TIFFTAG_PLANARCONFIG, &planar_config);
  stripSize      = TIFFStripSize(tiff);
  numberOfStrips =  TIFFNumberOfStrips( tiff);

  if( planar_config == PLANARCONFIG_SEPARATE && NDIMV > 1) {
    fprintf( stderr, "I don't know to support planar separated images with multiframes !\n");
    exit( 1);
  }

  strip = (char *) i_malloc( stripSize);
  buf = i_malloc( stripSize * numberOfStrips);
 
  inrlineSize  = (DIMX * bitsPerSample + 7) / 8;
  
  /* read the whole image */
  for( is = 0; is < numberOfStrips; is ++) {      
    readSize = TIFFReadEncodedStrip( tiff, is, strip, (tsize_t) -1);
    if( debug_) 
      fprintf( stderr, "reading %d bytes for strip %d ...\n", readSize, is);
    memcpy( buf + is * stripSize, strip, MIN(readSize, stripSize));
  }
    
  /* then write it */
  if( TYPE == PACKEE) {
    /* buf has to been splited into lines (if TYPE==PACKEE)
     * Inrimage requires buf line multiple of 8 bits, it is not necessarly
     * the case with the strip access */ 
    for( iy=0; iy < DIMY; iy ++) 
      c_ecr( nf, 1, (void*)(buf + inrlineSize*iy));	
  } else 
    c_ecr( nf, DIMY, (void*)buf );
  

#if ATTIC
  {
    /* code non fonctionnel pour rorganiser une image  plan spar en plan contigue
     * lorsque le codage en bit n'est pas un multiple de 8 */
    int ix, iv;
    float *fbuf;
    float *line;
    Fort_int icodi[2] = {BSIZE,TYPE}, icodo[2] = {sizeof(float),REELLE};
    
    
    fbuf = (void*)i_malloc( sizeof(float) * DIMX*2);
    line = (void*)i_malloc( sizeof(float) * DIMX*2);
    
    inrlineSize  = (NDIMX * bitsPerSample + 7) / 8;     
    
    /* convert a dimz-buffer in a dimv-buffer */    
    for( iy = 0; iy < NDIMY; iy ++) {      
      for( iv = 0; iv < NDIMV; iv ++)
	c_cnvtbg( buf + inrlineSize*(iy + NDIMY*iv),
		  //c_cnvtbg( buf + bitsPerSample*NDIMX*(iy + NDIMY*iv)/8,
		  fbuf + NDIMX*iv, NDIMX, icodi, icodo, EXP, 0);
      
      /* fbuf contains NDIMV lines to be reorganized */
      for( ix = 0; ix < NDIMX; ix ++)
	for( iv = 0; iv < NDIMV; iv ++)
	  line[iv + NDIMV * ix] = fbuf[ ix + NDIMX * iv];
      
      c_ecrflt( nf, 1, line);	
    }
    i_Free( (void *)&line);
    i_Free( (void *)&fbuf);
  }
#endif

  
  i_Free( (void*)&strip);
  i_Free( (void*)&buf);
    
}

/**
 * A FAIRE, POUR TEST 
 */

void read_and_write_scanline_image( TIFF *tiff, struct image *nf, Fort_int *lfmt) {  
  Fort_int ibs;
  tdata_t inbuf  = i_malloc(TIFFScanlineSize( tiff));
  int y, z;

  if( lfmt == NULL) lfmt = nf->lfmt;

  for( z=0; z<NDIMZ; z++)
    for( y=0; y<NDIMY; y++) {
      TIFFReadScanline( tiff, inbuf, y, z);
      c_ecr( nf, 1, inbuf);
    }
  
  i_Free( (void*)&inbuf);
}

/**
 * Read a image into its coding format (pixel size and memory organization).
 *
 * @param lfmt if \b NULL use format of nf image. If non \b NULL, lfmt is a valid
 *             Inrimage format array. lfmt may contain a restricted number of
 *             lines or frames (in order to read less than the real format).
 * @param use_strip Use a Strip access (instead of a Scanline access). If image
 *                  is Tile, this parameter has no effect.
 */

void read_and_write_image( TIFF *tiff, struct image *nf, Fort_int *lfmt, int use_strip) {
  uint16 photometric;
  TIFFGetField( tiff, TIFFTAG_PHOTOMETRIC, &photometric);

  switch( photometric) {
   
  case PHOTOMETRIC_MINISWHITE:
  case PHOTOMETRIC_MINISBLACK:
  case PHOTOMETRIC_RGB:
  case PHOTOMETRIC_PALETTE:
  case PHOTOMETRIC_MASK:
  case PHOTOMETRIC_SEPARATED:
  case PHOTOMETRIC_LOGL:
    
    if( TIFFIsTiled( tiff) ) {
      if( debug_) fprintf( stderr, "Access method : use Tile routines !\n");
      read_and_write_tiled_image(tiff, nf, lfmt);
    } else {
      if(  !use_strip) {
	if( debug_) fprintf( stderr, "Access method : use Scanline routines !\n");
	read_and_write_scanline_image( tiff, nf, lfmt);
      } else {
	if( debug_) fprintf( stderr, "Access method : use Strip routines !\n");
	read_and_write_strip_image( tiff, nf, lfmt);
      }
    }
    break;

  default:
    /* for other mode, 
     * we use the universal rgb access of libtiff */
    if( debug_) fprintf( stderr, "Access method : use RBGAImage routines !\n");
    read_and_write_rgb_image( tiff, nf, lfmt);
    break;    
  }
}

char cmname[256] = "";
int  hdr_size = 8;

/**
 * Read a colormap from a TIFF file and write into inrimage file.
 * If colormap is not found, the function does nothing.
 */

void read_and_write_colormap( TIFF *tiff, struct image *nf) {
  uint16 photom;
  uint16 *r, *g, *b; /* color palette values coded on 2-byte */

  if( TIFFGetField( tiff, TIFFTAG_PHOTOMETRIC, &photom) &&
      photom == PHOTOMETRIC_PALETTE &&
      TIFFGetField( tiff, TIFFTAG_COLORMAP, &r, &g, &b)) {
    uint16 bitsPerSample;
    int i, ncolors;
    unsigned char *ri, *gi, *bi;
    
    TIFFGetField( tiff, TIFFTAG_BITSPERSAMPLE, &bitsPerSample);
    ncolors = 1 << bitsPerSample;
    
    ri = (unsigned char *)i_malloc(ncolors);    
    gi = (unsigned char *)i_malloc(ncolors);    
    bi = (unsigned char *)i_malloc(ncolors);    

    if( debug_) fprintf( stderr, "found %d colors in colortab listed as below :\n", ncolors);
    for( i=0; i<ncolors; i++) {
      ri[i] = r[i] / 255;
      gi[i] = g[i] / 255;
      bi[i] = b[i] / 255;
      if( debug_) fprintf( stderr, "%4x %4x %4x\n", r[i], g[i], b[i]);
    }

    if( *cmname ) {
      Fort_int lfmt[9] = { ncolors, 3, 1, FIXE, ncolors, 3, 1, 1, 200};
      nf = image_( cmname,"c"," ", (NF_fmt*)lfmt);
      c_ecr( nf, 1, (void*)ri);
      c_ecr( nf, 1, (void*)gi);
      c_ecr( nf, 1, (void*)bi);
      fermnf_( (void*)&nf);
    } else {
      /* adjust inrimage header size */
      iwr_ctrgb( nf, 0, ncolors, (void*)ri, (void*)gi, (void*)bi);
    }
    i_Free( (void*)&ri);
    i_Free( (void*)&gi);
    i_Free( (void*)&bi);
  }
}
/**
 * Compute and set the header size required to store colormap (if exist)
 * and informative tags (if tags==1) founds in tiff image.
 */

void set_header_size( TIFF *tiff, int tags) {
  char *str;
  int size = 0;
  uint16 photom;

  /* mesure the size for informative tags */
  if( tags) {
    if( TIFFGetField( tiff, TIFFTAG_ARTIST, &str)) 
      size += strlen( str) + strlen("ARTIST");
    if( TIFFGetField( tiff, TIFFTAG_COPYRIGHT, &str)) 
      size += strlen( str) + strlen( "COPYRIGHT");
    if( TIFFGetField( tiff, TIFFTAG_DATETIME, &str)) 
      size += strlen( str) + strlen("DATETIME");
    if( TIFFGetField( tiff, TIFFTAG_DOCUMENTNAME, &str)) 
      size += strlen( str) + strlen( "DOCUMENTNAME");
    if( TIFFGetField( tiff, TIFFTAG_HOSTCOMPUTER, &str)) 
      size += strlen( str) + strlen( "HOSTCOMPUTER");
    if( TIFFGetField( tiff, TIFFTAG_IMAGEDESCRIPTION, &str)) 
      size += strlen( str) + strlen( "IMAGEDESCRIPTION");
    if( TIFFGetField( tiff, TIFFTAG_SOFTWARE, &str)) 
      size += strlen( str) + strlen( "SOFTWARE");
  }

  /* for colormap */
  if( TIFFGetField( tiff, TIFFTAG_PHOTOMETRIC, &photom)
      && photom == PHOTOMETRIC_PALETTE)
    size += 8*256;
  
  size /= 256;
  
  if( size > 1) { 
    /* leave a chance to user to set himself
     * the header size */
    igetopt0x("-hdr=","%d", &size);
    set_hdr_min( size);
    if( debug_) 
      fprintf( stderr, "Setting header size to %d bytes\n", 256*size);
  }
}

void conv_nl( char *str) {
  while( *str++)
    if( *str == '\n') *str ='\t';
}
/**
 * Keep in inrimage header some informative tags (copyright, description, etc...)
 */

void write_info_tags( TIFF *tiff, struct image *nf) {
  char *str;

  /* mesure the size */
  if( TIFFGetField( tiff, TIFFTAG_ARTIST, &str)) {
    conv_nl( str);
    iputhline( nf, "ARTIST=", str);
  }
  if( TIFFGetField( tiff, TIFFTAG_COPYRIGHT, &str)) {
    conv_nl( str);
    iputhline( nf, "COPYRIGHT=", str);
  }
  if( TIFFGetField( tiff, TIFFTAG_DATETIME, &str)) {
    conv_nl( str);
    iputhline( nf, "DATETIME=", str);
  }
  if( TIFFGetField( tiff, TIFFTAG_DOCUMENTNAME, &str)) {
    conv_nl( str);
    iputhline( nf, "DOCUMENTNAME=", str);
  }
  if( TIFFGetField( tiff, TIFFTAG_HOSTCOMPUTER, &str)) {
    conv_nl( str);
    iputhline( nf, "HOSTCOMPUTER=", str);
  }
  if( TIFFGetField( tiff, TIFFTAG_IMAGEDESCRIPTION, &str)) {
    conv_nl( str);
    iputhline( nf, "IMAGEDESCRIPTION=", str);
  }
  if( TIFFGetField( tiff, TIFFTAG_SOFTWARE, &str)) {
    conv_nl( str);
    iputhline( nf, "SOFTWARE=", str);
  }
}

/**
 * Change a format to RGB.
 */
void set_lfmt_rgb( Fort_int lfmt[9]) {
  BSIZE = 1;
  TYPE  = FIXE;
  EXP   = 200;
  NDIMV = 3;
  DIMX = NDIMV * NDIMX;
}

/*  version : 2.1 was the lastest tiff2inr version
 *            3.0.0 is the newest tiff2inr (from scratch)
 *            3.0.1 case of tag TIFFTAG_SAMPLESPERPIXEL is missing  is now handled
 */

char version[] = "3.0.1";
char command[] = "input-tiff [ -i | [output][-d %d -id %d][-rgb][-ct=%s][-fs][-nt]]";
char usage[] = "Convert any tiff image into Inrimage. All switches are optionals.\n"
  "    -i : display information on tiff image in INRIMAGE convention.\n"
  "   -ii : display more information than -i (for debugging)\n"
  
  "  -ct= : write colormap in a separate image (specified by option value) \n"
  
  "  -rgb : convert to 24 bits (NDIMV=3). Should works for all images but one\n"
  "         can loose information. Try this option if conversion failed.\n"
  
  "  -d2z : If image contains directories, try to create one image with several\n"
  "         frames (NDIMZ>1). If it is not possible, this option is ignored.\n"
  
  "   -id : index of first directory to extract, 1 is default\n"
  "    -d : count of directories to extract, -1 means extract all, 1 is default\n"
  "         or -1 if option -d2z is set\n"
  
  "   -fs : use strip access instead of scanline access (if image is not tiled)\n"
  "   -nt : don't write info tags found in Inrimage header\n"
  
  "    -D : print debug information (libtiff and Inrimage)\n"
;

int extern debug_;

int main( int argc, char **argv) {
  char name[256];
  TIFF *tiff;

  inr_init(argc,argv, version, command, usage);
  TIFFSetWarningHandler( tiff_warning);
  quiet = !debug_;
  
  infileopt( name);
  if( *name == '<') {
    fprintf( stderr, "tiff2inr:error:can not read standard input.\n");
    exit( 1);
  }
  tiff = TIFFOpen( name, "r");

  if( tiff) {
    Fort_int lfmt[9];
    int ndirs =  TIFFNumberOfDirectories(tiff);
    int convert_to_rgb, convert_dirs_to_dimz, force_strip, dont_write_info_tags;

    int show_info, show_more_info;
    char generic[256];
    int id=1, d=1; /* index of first directory, count of directories */
    
    show_info = igetopt0( "-i");
    show_more_info = igetopt0( "-ii");
 
    /* force to convert into RGB 24 bits ? */
    convert_to_rgb = igetopt0( "-rgb");

    /* try to convert multiple directories image in one image */
    convert_dirs_to_dimz = igetopt0( "-d2z");

    /* force write access using strip */
    force_strip = igetopt0( "-fs");

    /* write in Inrimage header informative tags found */
    dont_write_info_tags = igetopt0( "-nt");
    
    /* Inrimage header size (global variable, read by write_colormap()) */
    igetopt0x( "-hdr=", "%d", &hdr_size);
    
    /* Colormap in a separate file (global variable, read by write_colormap()) */
    igetopt0x( "-ct=", "%s", &cmname);

    /*
     * Display information on images 
     */
    
    if( show_info || show_more_info) {
      Fort_int iopt = 3, std = 1;
      int print_dir = 0;
      
      fprintf( stdout, "%s  \n", name);
      if( ndirs > 1) {
	fprintf( stdout, "Number of directories:  %d\n", ndirs);
	print_dir = 1;
	ndirs = 0;
      }
      do {
	ndirs ++;
	if( print_dir) fprintf( stdout, "******** directory %d ********\n", ndirs);
        switch( read_lfmt( tiff, lfmt)) { 
	case USE_STD:
	  wrfmgg_( lfmt, &iopt, &std);  
	  break;
	case USE_RGB:
	  wrfmgg_( lfmt, &iopt, &std);  
	  fprintf( stdout, 
		   "### tiff2inr can not handle this kind of photometric interpretation (see\n"
		   "### -ii option). This image will converted in RGB 24bits. So the real format\n"
		   "### will be : \n");
	  set_lfmt_rgb( lfmt);
	  wrfmgg_( lfmt, &iopt, &std);  
	  break;
	}
	if(show_more_info) display_others_stats( tiff);	
      } while ( TIFFReadDirectory( tiff));
    } else if( convert_dirs_to_dimz && ndirs > 1 && check_directories( tiff, lfmt) && NDIMZ==1) {
      int use_rgb;
      /*
       * Try to convert to INRIMAGE multiple directories in one file
       */      
      
      struct image *nf;
      
      /* by default, get all directories */
      d = -1;
      igetopt1( "-id", "%d", &id);
      igetopt1( "-d", "%d", &d);
      
      use_rgb = read_lfmt( tiff, lfmt);
      if( use_rgb == USE_RGB)
	set_lfmt_rgb( lfmt);
      
      NDIMZ = (d==-1) ? ndirs:d;
      DIMY = NDIMY*NDIMZ;
      
      /* FIXME */
      if( debug_) fprintf( stderr, "Setting header size to %d bytes\n", 256*hdr_size);
      set_hdr_min( hdr_size);

      outfileopt( generic);
      set_header_size( tiff, !dont_write_info_tags);
      nf = image_( generic, "c", "", (void*)lfmt);
      if( !dont_write_info_tags)
	write_info_tags( tiff, nf);
      if( !use_rgb)
	read_and_write_colormap( tiff, nf);
      
      /* force read_and_write_image() function to process
       * only one frame. Probably it does not work with tiled image,
       * but it is ok with striped image.  */

      NDIMZ = 1; 
      DIMY = NDIMY;
      
      TIFFSetDirectory( tiff, id-1);	
      do 
	read_and_write_image( tiff, nf, lfmt, force_strip);
      while( TIFFReadDirectory( tiff) && --d);
      fermnf_( (void *)&nf);
      
    } else {
      
      /* diagnostic for -d2z option which failed */
      if( convert_dirs_to_dimz && !quiet) {
	if( ndirs == 1)
	  fprintf( stderr, "tiff2inr -d2z:warning:image has only one directory.\n");
	else if( NDIMZ==1) 
	  fprintf( stderr, "tiff2inr -d2z:warning:directories have not the same format.\n");
	else
	  /* FIXME */
	  fprintf( stderr, "tiff2inr -d2z:warning:i can't process depth image.\n");
      }
    
      /*
       * Convert to INRIMAGE
       */

      outfileopt( generic);      

      /* by defaut, only one directory is extracted 
       * set -d -1 for extract all */
      d = 1;      
      igetopt1( "-id", "%d", &id);
      igetopt1( "-d", "%d", &d);
      if( d != 1) 
	sprintf( name, "%s-%04d.inr", generic, id);
      else
	strcpy( name, generic);

      TIFFSetDirectory( tiff, id-1);
      do {
	struct image *nf;
	
	set_header_size( tiff, !dont_write_info_tags);
	if( read_lfmt( tiff, lfmt) == USE_RGB || convert_to_rgb) {
	  set_lfmt_rgb( lfmt);
	  nf = image_( name, "c", " ", (void *)lfmt);
	  if( !dont_write_info_tags)
	    write_info_tags( tiff, nf);
	  read_and_write_rgb_image( tiff, nf, NULL);
	} else {
	  nf = image_( name, "c", " ", (void *)lfmt);
	  if( !dont_write_info_tags)
	    write_info_tags( tiff, nf);
	  read_and_write_colormap( tiff, nf);
	  read_and_write_image( tiff, nf, NULL, force_strip);
	}
	
	fermnf_( (void *)&nf);
	
	if( d > 0) {
	  d --;
	  if( d == 0) break;
	}
	sprintf( name, "%s-%04d.inr", generic, ++id);
      } while( TIFFReadDirectory( tiff));
    }
    TIFFClose( tiff);    
  }

  return 0;  
}
