/* HISTOOG
	SUBROUTINE HISTOG(TAB,NFIX,A,NDIM,NF1,NF2,ICD,ISEP,AMIN,AMAX)
	
 TAB : tableau de resultat de NFIX elements
 A  : tableau pour les donnees (NDIM elements)
 NF1, NF2 : images de donnees et resultat
 ICD : -2 => histogramme de couleur normalise 
       -1 => histogramme normalise (fraction de pts / niveau),
       0 = histo simple (nb pts / niveau)
       1 = fonct. de repartition (fraction de pts de val <= niveau)
 ISEP : 1 : separer les res. par composante vectorielle
        2 : separer par plan
	3 : separer par composante et plan (res: 256,1,NDIMV,NDIMZ)
	0 : grouper
 AMIN, AMAX : min et max global, pour normaliser les valeurs, si image non 8 bits
Note: histogramme de couleur normalise = cube 4x4x4 pour r,g,b de 0 a 3
H(r,g,b) = prob((r*64) <= xr < ((r+1)*64)
                && (g*64) <= xg < ((g+1)*64)
                && (b*64) <= xb < ((b+1)*64)
  resultat dans image ndimx=4 (r), ndimy=4 (g) et ndimz=4 (b)
*/

#include <inrimage/image.h>

#define SEP_V 1
#define SEP_Z 2

histog_(htab,nfix,buf0,ndim,nfi,nfo,icd,isep,amin,amax)
float *htab;
void *buf0;
Fort_int *nfix, *ndim, *icd, *isep;
float *amin, *amax;
struct image **nfi, **nfo;
{
	Fort_int lfmt[9], lfmth[9];
	Fort_int nrow, lnrow,nb_pixels;
	int i, j, k, n, nb, ntot;
	int nhv, nhz, nv;
	int iy, ndy;
	int flg_isep, flg_int;
	register float *p;
	register float *pf;
	register CHAR *pc;
	flg_isep = *isep;
	if((flg_isep > 3) || (flg_isep < 0))
		imerror(9,"histog_ : mauvais argument ISEP");
	c_zcore(htab,*nfix);
	cpyfmg_(nfi,lfmt);
	if((*icd == -2) && (NDIMV != 3))
		imerror(9,"Images couleur avec NDIMV==3 seulement");
	cpyfmg_(nfo,lfmth);
	if((TYPE == FIXE) && (BSIZE == 1)) {
		flg_int = 1;
	} else {
		flg_int = 0;
		TYPE = REELLE;
		BSIZE = sizeof(float);
	}
	setup_(ndim,lfmt,&nrow,&lnrow,&nb_pixels);
	i = DIMY;
	ntot = DIMX * DIMY;
	nb = nrow * DIMX;
	nhv = lfmth[I_NDIMV];
	nhz = lfmth[I_NDIMZ];
	if((nhv == 1) && (nhz == 1)) flg_isep = 0;  /* rien a separer */
	iy = 0;
	p = htab;
	if(! flg_int && (*amin == *amax)) {
		if(*icd == -2) 
			rhist0c(htab,lfmth);
		else
			rhist0(htab,nhv,nhz,ntot,*icd);
		c_ecr(*nfo,lfmth[I_DIMY],htab);
		return 1 /* makes c99 happy */; 
	}
	while(i > 0) {
		if(flg_int)
			c_lect(*nfi,nrow,buf0);
		else
			c_lecflt(*nfi,nrow,buf0);
		if(! flg_isep) {
			if(*icd == -2) {
				if(flg_int)
					histo1c(buf0,nb,htab);
				else
					rhisto1c(buf0,nb,htab,*amin,*amax);
			} else {
				if(flg_int)
					histo1(buf0,nb,htab);
				else
					rhisto1(buf0,nb,htab,*amin,*amax);
			}
		} else {
			nb = nrow * DIMX;
			if(flg_int)
				pc = (CHAR *)buf0;
			else
				pf = (float *)buf0;
			if(flg_isep & SEP_Z)
				ndy = NDIMY;
			else
				ndy = DIMY;
			ntot = nrow;
			n = ntot;
			while(ntot > 0) {  /* plan par plan */
				if((n + iy) >= ndy) {
					n = ndy - iy;
					iy = ndy;
					nb = n * DIMX;
				}
				iy += n;
				ntot -= n;
				if(flg_int) {
					for(k=0;k<nhv;k++) {
						histog1(pc+k,nb,p+k,nhv);
					}
					pc += nb;
				} else {
					for(k=0;k<nhv;k++) {
						rhistog1(pf+k,nb,p+k,nhv,*amin,*amax);
					}
					pf += nb;
				}
				if(iy >= ndy) {
					iy = 0;
					p += lfmth[I_DIMX];
					n = ntot;
					nb = n * DIMX;
				}
			}
		}
		i -= nrow;
		if(i < nrow) {
			nrow = i;
			nb = nrow * DIMX;
		}
	}
	if(*icd > 0) {  /* f. de repartition */
		p = htab;
		for(k=0;k<nhz;k++) {
			for(j=0;j<nhv;j++) {
				for(i=j+nhv;i< lfmth[I_DIMX];i += nhv)
					*(p + i) += *(p + i - nhv);
			}
			
		}
		p += lfmth[I_DIMX];
	}
	if(*icd) { /* diviser par nb pts */
		n = NDIMX * NDIMY;
		nb = lfmth[I_NDIMX];
		if(flg_isep & SEP_V)
			nb *= NDIMV;
		else if(*icd > -2)
			n *= NDIMV;
		else
			nb *= lfmth[I_DIMY];
		if(flg_isep & SEP_Z)
			nb *= NDIMZ;
		else
			n *= NDIMZ;
		for(i=0;i<nb;i++)
			*(htab + i) /= n;
	}
	c_ecr(*nfo,lfmth[I_DIMY],htab);
}

/* ************************************************************
     pour images en 8 bits 
     */
histo1(buf,n,htab)
register CHAR *buf;
register float *htab;
register n;
{
	register i;
	while(n-- > 0) {
		i = (int)*buf++;
		*(htab + i) += 1;
	}
}

/* histogramme couleur normalise 4x4x4, donnees avec ndimv=3 */
histo1c(buf,n,htab)
register CHAR *buf;
register float *htab;
register n;
{
	register i;
	while(n > 0) {
		n -= 3;
		i = (int)*buf++ >> 6;  /* r */
		i += 4 * ((int)*buf++ >> 6);  /* g */
		i += 16 * ((int)*buf++ >> 6);  /* b */
		*(htab + i) += 1;
	}
}

histog1(buf,n,htab,incr)
register CHAR *buf;
register float *htab;
register n, incr;
{
	register i;
	float tmp[256];
	for(i=0;i<256;i++) {
		tmp[i] = 0;
	}
	while((n -= incr) >= 0) {
		i = (int)*buf;
		buf += incr;
		tmp[i]++;
	}
	for(i=0;i<256;i++) {
		*htab += tmp[i];
		htab += incr;
	}
}

/* ************************************************************
     pour images en flottant
     */

rhisto1(buf,n,htab,min,max)
register float *buf;
register float *htab;
register n;
float min, max;
{
	register i;
	register float d;
	d = max - min;
	while(n-- > 0) {
		i = (int)((255. * (*buf++ - min) / d) + .5);
		*(htab + i) += 1;
	}
}

/* histogramme couleur normalise 4x4x4, donnees avec ndimv=3 */
rhisto1c(buf,n,htab,min,max)
register float *buf;
register float *htab;
register n;
float min, max;
{
	register i;
	register float d;
	d = max - min;
	while(n > 0) {
		n -= 3;
		i = ((int)((255. * (*buf++ - min) / d) + .5)) >> 6;  /* r */
		i += 4 * (((int)((255. * (*buf++ - min) / d) + .5)) >> 6);  /* g */
		i += 16 * (((int)((255. * (*buf++ - min) / d) + .5)) >> 6);  /* b */
		*(htab + i) += 1;
	}
}

rhistog1(buf,n,htab,incr,min,max)
register float *buf;
register float *htab;
register n, incr;
float min, max;
{
	register i;
	float tmp[256];
	register float d;
	d = max - min;
	for(i=0;i<256;i++) {
		tmp[i] = 0;
	}
	while((n -= incr) >= 0) {
		i = (int)((255. * (*buf - min) / d) + .5);
		buf += incr;
		tmp[i]++;
	}
	for(i=0;i<256;i++) {
		*htab += tmp[i];
		htab += incr;
	}
}

rhist0(htab,nv,nz,npts,icd)
float *htab;
int nv, nz, npts, icd;
{
	float val0, val1;
	register i, j, k, n2;
	val1 = 0.;
	val0 = 1.;
	n2 = 255 * nv;
	if(icd == 0) val0 = (float) npts;
	if(icd > 0) val1 = 1.;
	for(i=0;i<nz;i++) {
		for(j=0;j<nv;j++)
			*htab++ = val0;
		for(k=0;k<n2;k++)
			*htab++ = val1;
	}
}

rhist0c(htab,lfmth)
float *htab;
Fort_int lfmth[];
{
	register i;
	*htab++ = 1;
	i = lfmth[I_DIMX] * lfmth[I_DIMY] - 1;
	while (i-- > 0)
		*htab++ = 0;
}
