#include <stdio.h>
#include <inrimage/image.h>

/*
  bwc8_to_c24(bufi,bufr,bufg,bufb,n,pcol)
  c24_to_bw(bufr,bufg,bufb,bufo,n,pcol)
  c8_to_bw(bufi,bufo,n,pcol)
  c24_to_c8(bufr,bufg,bufb,bufo,n,pcol)
  c24_to_cvec(bufr,bufg,bufb,bufo,n,pcol)
  bwc8_to_cvec(bufi,bufo,n,pcol)
  cvec_to_bw(bufi,bufo,n,pcol)
  cvec_to_c8(bufi,bufo,n,pcol)
  cvec_to_c24(bufi,bufr,bufg,bufb,n,pcol)
  set_c8_ct(pcr,pcg,pcb)
  set_c8_1ct(col,numc)
  apply_1ct(bufi,bufo,n,coltb)
  apply_ctvect(bufi,bufo,n,tabr,tabg,tabb)
*/

#define ALIGN4(x) ((x + 3) & ~3)
#define NBCOL 256

/* pour 24 bits on a 3 buffers 
Si le ptr pcol vers la table des coltb est NULL, il n'y a pas de c.t.
*/


/* conversion 8 bits (NB ou couleur) en 24 bits couleur :
on applique les tables donnees par pcol. 
bufo contient 3 tableaux de n octets, alignes sur des long
*/
bwc8_to_c24(bufi,bufr,bufg,bufb,n,pcol)
register unsigned char *bufi, *bufr, *bufg, *bufb, **pcol;
register n;
{
	register i;
	if(pcol == NULL) {
		while(n-- > 0) {
			i = *bufi++;
			*bufr++ = i;
			*bufg++ = i;
			*bufb++ = i;
		}
		return 0;
	}
	while(n-- > 0) {
		i = *bufi++;
		*bufr++ = *(pcol[0] + i);
		*bufg++ = *(pcol[1] + i);
		*bufb++ = *(pcol[2] + i);
	}
	return 0;
}

/* conversion 8 bits (NB ou couleur) en 24 bits couleur , 3 composantes par pixel:
on applique les tables donnees par pcol. 

*/
bwc8_to_cvec(bufi,bufo,n,pcol)
register unsigned char *bufi, *bufo, **pcol;
register n;
{
	register i;
	if(pcol == NULL) {
		while(n-- > 0) {
			i = *bufi++;
			*bufo++ = i;
			*bufo++ = i;
			*bufo++ = i;
		}
		return 0;
	}
	while(n-- > 0) {
		i = *bufi++;
		*bufo++ = *(pcol[0] + i);
		*bufo++ = *(pcol[1] + i);
		*bufo++ = *(pcol[2] + i);
	}
	return 0;
}

/* conversion 24 bits couleur en N&B :
.33R+.5G+.17B 
On applique d'abord les T.C. sur les donnees
*/
c24_to_bw(bufr,bufg,bufb,bufo,n,pcol)
register unsigned char *bufr, *bufg, *bufb;
register unsigned char *bufo, **pcol;
register n;
{
	if(pcol == NULL) {
		while(n-- > 0) {
			*bufo++ = (*bufr++ * 11 +  *bufg++ * 16 + *bufb++ * 5)
				>> 5;
		}
		return 0;
	}
	while(n-- > 0) {
		*bufo++ =  (*(pcol[0] + *bufr++) * 11 +
			   *(pcol[1] + *bufg++) * 16 +
			   *(pcol[2] + *bufb++) * 5) >> 5;
	}
	return 0;
}

/* conversion 24 bits couleur (3 composantes/pixel) en N&B :
.33R+.5G+.17B 
On applique d'abord les T.C. sur les donnees
*/
cvec_to_bw(bufi,bufo,n,pcol)
register unsigned char *bufi;
register unsigned char *bufo, **pcol;
register n;
{
	if(pcol == NULL) {
		while(n-- > 0) {
			*bufo++ =  (*bufi * 11 +  *(bufi+1) * 16 + *(bufi+2) * 5) >> 5;
			bufi += 3;
		}
		return 0;
	}
	while(n-- > 0) {
		*bufo++ =  (*(pcol[0] + *bufi) * 11 +
			   *(pcol[1] + *(bufi+1)) * 16 +
			   *(pcol[2] + *(bufi+2)) * 5) >> 5;
		bufi += 3;
	}
	return 0;
}

/* conversion 8 bits couleur en N&B :
.33R+.5G+.17B 
On applique d'abord les T.C. sur les donnees
*/
c8_to_bw(bufi,bufo,n,pcol)
register unsigned char *bufi;
register unsigned char *bufo, **pcol;
register n;
{
	register i;
	if(pcol == NULL) {
		imerror(9,"c8_to_bw : pointeur de table couleur NULL!\n");
	}
	while(n-- > 0) {
		i = *bufi++;
		*bufo++ = (*(pcol[0] + i) * 11 +
			   *(pcol[1] + i) * 16 +
			   *(pcol[2] + i) * 5) >> 5;
	}
	return 0;
}

/* applique table coltb sur buffer de n pixels */
apply_1ct(bufi,bufo,n,coltb)
register unsigned char *bufi, *bufo, *coltb;
register n;
{
	while(n-- > 0)
		*bufo++ = *(coltb + *bufi++);
}

/* applique tables couleurs tabr,tabg,tabb sur buffer de n pixels a 3 composantes */
apply_ctvect(bufi,bufo,n,tabr,tabg,tabb)
register unsigned char *bufi, *bufo, *tabr, *tabg, *tabb;
register n;
{
	while(n-- > 0) {
		*bufo++ = *(tabr + *bufi++);
		*bufo++ = *(tabg + *bufi++);
		*bufo++ = *(tabb + *bufi++);
	}
}


static unsigned char rvb_shft[]= { 0, 3, 6 };
static unsigned char rvb_msk[]= {0xe0,0xe0,0xc0};
static unsigned char rvb_rshft[]= {5,2,0};
static unsigned char rvb_tc3b[]= {0,0x24,0x49,0x6d,0x92,0xb6,0xdb,0xff};
static unsigned char rvb_tc2b[]= {0,0x49,0x92,0xff}; /* {0,0x55,0xaa,0xff}; */
static unsigned char rvb_lg3b[]= {16,16,16,16,16,16,16,32};
static unsigned char rvb_lg2b[]= {48,48,48,64}; /* {32,32,32,64}; */

static unsigned char cnvr24_8[NBCOL], cnvg24_8[NBCOL], cnvb24_8[NBCOL];
static unsigned char *ptabc24_8[3]= {cnvr24_8, cnvg24_8, cnvb24_8};
static int cnv_init= 0;
static int init_cnv24()
{
	register i, s, msk, cmax, n, n2, c2, ntc;
	register unsigned char *po;
	register unsigned char *pt;
	cnv_init = 1;
	for(ntc=0;ntc<3;ntc++) {
		s = rvb_shft[ntc];
		msk = rvb_msk[ntc] >> s;
		s = rvb_rshft[ntc];
		cmax = msk >> s;  /* indice max  dans la couleur */
		po = ptabc24_8[ntc];
		if(cmax == 7) {
			pt = rvb_lg3b;
			n2 = 32;
		} else {
			pt = rvb_lg2b;
			n2 = 64;
		}

		for(i=0;i<=cmax;i++) {
			c2 = *(pt + i);
			for(n=0;n<c2;n++)
				*po++ = i << s;
			for(;n<n2;n++)
				*po++ = (i+1) << s;
		}
	}
}
c24_to_c8(bufr,bufg,bufb,bufo,n,pcol)
register unsigned char *bufr, *bufg, *bufb;
register unsigned char *bufo, **pcol;
register n;
{
	if(!cnv_init)
		init_cnv24();
	if(pcol == NULL) {
		while(n-- > 0) {
			*bufo++ = cnvr24_8[*bufr++] | cnvg24_8[*bufg++] | cnvb24_8[*bufb++];
		}
		return 0;
	}
	while(n-- > 0) {
		*bufo++ = cnvr24_8[*(pcol[0] + *bufr++)] |
			cnvg24_8[*(pcol[1] + *bufg++)] |
				cnvb24_8[*(pcol[2] + *bufb++)];
	}
	return 0;
}

cvec_to_c8(bufi,bufo,n,pcol)
register unsigned char *bufi;
register unsigned char *bufo, **pcol;
register n;
{
	if(!cnv_init)
		init_cnv24();
	if(pcol == NULL) {
		while(n-- > 0) {
			*bufo++ = cnvr24_8[*bufi] | cnvg24_8[*(bufi+1)] | cnvb24_8[*(bufi+2)];
			bufi += 3;
		}
		return 0;
	}
	while(n-- > 0) {
		*bufo++ = cnvr24_8[*(pcol[0] + *bufi)] |
			cnvg24_8[*(pcol[1] + *(bufi+1))] |
				cnvb24_8[*(pcol[2] + *(bufi+2))];
		bufi += 3;
	}
	return 0;
}

c24_to_cvec(bufr,bufg,bufb,bufo,n,pcol)
register unsigned char *bufr, *bufg, *bufb;
register unsigned char *bufo, **pcol;
register n;
{
	if(pcol == NULL) {
		while(n-- > 0) {
			*bufo++ = *bufr++;
			*bufo++ = *bufg++;
			*bufo++ = *bufb++;
		}
		return 0;
	}
	while(n-- > 0) {
		*bufo++ = *(pcol[0] + *bufr++);
		*bufo++ = *(pcol[1] + *bufg++);
		*bufo++ = *(pcol[2] + *bufb++);
	}
	return 0;
}

cvec_to_c24(bufi,bufr,bufg,bufb,n,pcol)
register unsigned char *bufr, *bufg, *bufb;
register unsigned char *bufi, **pcol;
register n;
{
	if(pcol == NULL) {
		while(n-- > 0) {
			*bufr++ = *bufi++;
			*bufg++ = *bufi++;
			*bufb++ = *bufi++;
		}
		return 0;
	}
	while(n-- > 0) {
		*bufr++ = *(pcol[0] + *bufi++);
		*bufg++ = *(pcol[1] + *bufi++);
		*bufb++ = *(pcol[2] + *bufi++);
	}
	return 0;
}

set_c8_ct(pcr,pcg,pcb)
register unsigned char *pcr, *pcg, *pcb;
{
	set_c8_1ct(pcr,0);
	set_c8_1ct(pcg,1);
	set_c8_1ct(pcb,2);
}

set_c8_1ct(col,numc)
register unsigned char *col;
int numc;
{
	register s, m;
	register i, c2, cmax;
	register unsigned char *pt;
	if(numc < 0) numc = 0;
	else if(numc > 2) numc = 2;
	s = rvb_shft[numc];
	m = rvb_msk[numc] >> s;
	s = rvb_rshft[numc];
	cmax = m >> s;  /* indice max  dans la couleur */
#if 0
	c2 = cmax / 2;
	for(i=0;i<NBCOL;i++) {
		n = (i & m) >> s;  /* numero d'ordre de la couleur */
		n = (255 * n + c2) / cmax;
		*col++ = n;
	}
#else
	pt = cmax == 3 ? rvb_tc2b : rvb_tc3b;
	for(i=0;i<NBCOL;i++) {
		c2 = (i & m) >> s;
		*col++ = *(pt + c2);
	}
#endif
}
		

