/* cnvtbg
	subroutine cnvtbg(a,b,ndim,icodi,icodo,iexpi,iexpo)
	  c_cnvtbg(a,b,ndim,icodi,icodo,iexpi,iexpo)
	     int ndim; Fort_int icodi[2], icodo[2], iexpi, iexpo;
 effectue la conversion de a (ndim points de byte/bit size icodi(1),
  et de type icodi(2)) en b (byte size icodo(1), type icodo(2))
  type=1 si float, 0 ou -1 si virg. fixe
  si type=0, alors b.s. >0= byte size (nb octets/pnt), et b.s.<0=
    bit size (-b.s. bits sur 1 nb entier d'octets, cadres a droite)
  si type=-1, b.s. doit etre <0 et represente une bit size "compacte"
     (bits a la suite les uns des autres)
   pour les formats "compacts", on passe toujours par le format
      "etendu" correspondant (appel a pckbt et unpkbt)
 1 flottant correspond a un entier avec tous les bits a 1 (2**n-1)
  iexpi et iexpo doivent etre 0 pour float. pour v. fix. ils indiquent
    s'il y a un signe (iexp<0) ou non, ainsi que la valeur de l'exposan
    nbs consideres avec virg. devant bit de gauche, multiplies par
    2**n ou n=abs(iexp)-200  (n de -100 a +100)
  si exp=0, nbs compris entre 0 et 1 ou -.5 et +.5

On traite a part le cas 1bit (in ou out)
 **************************
  erreur 9 (niveau 2) si byte size ou type incorrects
 type 1 et b.s. non 4 interdit (pas de dble precision
 *************
  les nbs en dehors des limites permises selon type, signe et nexp.,
   sont forces a ces valeurs limites (par ex. 0 et 1 si nexp=0 sans sign
*/

#include <stdio.h>
#include <math.h>
#include <inrimage/image.h>

extern int debug_;
extern int fl_type();

static int set_type();

cnvtbg_(a,b,ndim,icodi,icodo,iexpi,iexpo)
CHAR *a, *b;
Fort_int *ndim, icodi[2], icodo[2], *iexpi, *iexpo;
{
	c_cnvtbg(a,b,(int)*ndim,icodi,icodo,*iexpi,*iexpo);
}

/* types : 1, 2, 4 octets , signe ou Non Signe */
#define T_FLOAT4 0
#define T_S1 1
#define T_NS1 2
#define T_S2 3
#define T_NS2 4
#define T_S4 5
#define T_NS4 6
#define T_FLOAT8 7
#define IS_PACK 0x100

_PRIVATE static nbits, nbytes, nexp, lgbits;
_PRIVATE static double i_coef, i_coefn, o_coef, o_coefn, i_exp, o_exp;
_PRIVATE static	int i_bits, o_bits, out_pack;
_PRIVATE static	Fort_int i_lgbits, o_lgbits, ndm;

#define p2n(x) (powf((double)2.,(double)x))
#define ABS(x) (x < 0 ? -x : x)

/* on commence par la fin s'il y a expansion des datas et si in == out */
c_cnvtbg(vin,vout,ndim,icodi,icodo,iexpi,iexpo)
register void *vin, *vout;
int ndim;
Fort_int icodi[2], icodo[2], iexpi, iexpo;
{
	register i_type, o_type, i_byte, o_byte;
	register double V, v_coef, v_coefn, o_max, o_min;
	register val;
	register unsigned char *in = vin, *out = vout;

	unsigned long o_imin, o_imax, i_sign;
	long i_sval;
	unsigned char *bufin, *bufout;
	bufin = in;
	bufout = out;
	ndm = (Fort_int)ndim;
	if(iexpi == 0) iexpi = icodi[1] ==  REELLE ? fl_type() : 200;
	if(iexpo == 0) iexpo = icodo[1] ==  REELLE ? fl_type() : 200;
	/* if nothing to do return immediatly */
	if((bufin == bufout) && (iexpi == iexpo) &&
	   (icodi[0] == icodo[0]) && (icodi[1] == icodo[1])) {
		if(debug_)
			fprintf(stderr,"cnvtbg: nothing to do\n");
		return 0;
	}
	i_type = set_type(icodi,iexpi);
	i_bits = nbits;  /* nb bits sans le signe */
	i_byte = nbytes;
	i_lgbits = (Fort_int)lgbits;  /* nb bits de codage */
	if((i_type == T_FLOAT4) || (i_type == T_FLOAT8)) {
		i_coef = 1.;
		i_coefn = 1.;
	} else {
		V = p2n(i_bits);
		i_exp = pow((double)2.,(double)nexp);
		i_sign = 1 << i_bits;
		i_sval = -1 << i_bits;
		i_coef = i_exp / (V - 1.);
		i_coefn = i_exp / V;
	}
	o_type = set_type(icodo,iexpo);
	o_bits = nbits;  /* nb bits sans le signe */
	o_byte = nbytes;
	o_lgbits = (Fort_int)lgbits;  /* nb bits de codage */
	if((o_type == T_FLOAT4) || (o_type == T_FLOAT8)) {
		o_coef = 1.;
		o_coefn = 1.;
	} else {
		V = p2n(o_bits);
		o_max = V - 1.;
		o_min = -V;  /* min pour nb avec signe */
		o_imin = (u_int4) ((int4) -1) << o_bits;
		o_imax = (u_int4)((int4) -1) >> (32 - o_bits);
		o_exp = pow((double)2.,(double)nexp);
		o_coef = o_exp / o_max;
		o_coefn = o_exp / V;
	}
	if((i_type & ~IS_PACK) == (o_type & ~IS_PACK) && iexpi == iexpo
	   && i_bits == o_bits) {
		   if((i_type & IS_PACK) == (o_type & IS_PACK))
			   return c_mvbits(in,out,1,1,lgbits*ndim);
		   if(i_type & IS_PACK)
			return unpkbt_(in,out,&ndm,&i_lgbits);
		   return pckbt_(in,out,&ndm,&i_lgbits);
	   }
	     
	if(in == out && o_byte > i_byte) { /* commencer par fin */
		in = in + (ndim-1) * i_byte;
		out = out + (ndim-1) * o_byte;
		i_byte = -i_byte;
		o_byte = -o_byte;
	}
	out_pack = o_type & IS_PACK ? 1 : 0;
	o_type &= ~IS_PACK;
	if(out_pack && o_bits == 1 && o_lgbits == 1)/* sortie 1 bit */
		if(o_type == T_FLOAT4 || iexpi == iexpo)
/*			return cnvto_1b(bufin,bufout,ndim,i_type,ABS(i_byte)) */ ;
	if(i_type & IS_PACK) {
		if(i_lgbits == 1 && i_bits == 1)
/*!!!			return cnv_1b(bufin,bufout,ndim,o_type,ABS(o_byte)) */;
		unpkbt_(bufin,bufin,&ndm,&i_lgbits);
	}
	i_type &= ~IS_PACK;
	v_coef = i_coef / o_coef;
	v_coefn = i_coefn / o_coefn;
	while(ndim > 0) {
		switch(i_type) {
		case T_FLOAT4:
			if((V = (double) *((float *) in)) < 0. )
				V *= v_coefn;
			else
				V *= v_coef;
			break;
		case T_FLOAT8:
			if((V =  *((double *) in)) < 0. )
				V *= v_coefn;
			else
				V *= v_coef;
			break;
		case T_NS1 :
			V = ((double) *in) * v_coef;
			break;
		case T_S1 :
			if((val = *(char *)in) & i_sign)
				V = ((double)(val | i_sval)) * v_coefn;
			else
				V = (double)val * v_coef;
			break;
		case T_S2 :
			if((val = *(short *)in) & i_sign) 
				V = ((double) (val | i_sval)) * v_coefn;
			else
				V = ((double) val) * v_coef;
			break;
		case T_NS2 :
			V = ((double) *(unsigned short *)in) * v_coef;
			break;
		case T_S4 :
			if((val = *(int4 *)in) & i_sign)
				V = ((double)(val | i_sval)) *  v_coefn;
			else
				V = (double)val * v_coef;
			break;
		case T_NS4 :
			V = ((double) *(u_int4 *)in) * v_coef;
			break;
		}
		switch(o_type) {
		case T_FLOAT4:
			*((float *) out) = V;
			break;
		case T_FLOAT8:
			*((double *) out) = V;
			break;
		case T_S1:
			if(V < 0) {
				if((V -= .5) < o_min)
					*((char *)out) = (char)o_imin;
				else 
					*((char *)out) = (char) V;
			} else {
				if((V += .5) >= o_max)
					*((char *)out) = (char)o_imax;
				else
					*((char *)out) = (char) V;
			}
			break;
		case T_NS1:
			if((V += .5) < 0)
				*out = 0;
			else {
				if(V > o_max)
					*out = (unsigned char)o_imax;
				else
					*out = (unsigned char) V;
			}
			break;
		case T_NS2:
			if((V += .5) < 0)
				*((unsigned short *)out) = 0;
			else {
				if(V > o_max)
					*((unsigned short *)out) = (unsigned short) o_imax;
				else
					*((unsigned short *)out) = (unsigned short) V;
			}
			break;
		case T_S2:
			if(V < 0.) {
				if((V -= .5) < o_min)
					*((short *)out) = (short)o_imin;
				else 
					*((short *)out) = (short) V;
			} else {
				if((V += .5) >= o_max)
					*((short *)out) = (short) o_imax;
				else
					*((short *)out) = (short) V;
			}
			break;
		case T_NS4:
			if((V += .5) < 0)
				*((u_int4 *)out) = 0;
			else {
				if(V > o_max)
					*((u_int4 *)out) = o_imax;
				else
					*((u_int4 *)out) = (u_int4) V;
			}
			break;
		case T_S4:
			if(V < 0.) {
				if((V -= .5) < o_min)
					*((int4 *)out) = (int4) o_imin;
				else 
					*((int4 *)out) = (int4) V;
			} else {
				if((V += .5) >= o_max)
					*((int4 *)out) = (int4) o_imax;
				else
					*((int4 *)out) = (int4) V;
			}
			break;
		}
		ndim--;
		in += i_byte;
		out += o_byte;
	}
	if(out_pack)
		pckbt_(bufout,bufout,&ndm,&o_lgbits);
	return 0;
}	  

static int set_type(icode,iexp)
Fort_int icode[2];
register iexp;
{
	register type;
	if(icode[1] > 0 ) {
		type = T_FLOAT4;
		nbytes = icode[0];
		nbits = 8 * nbytes;
		lgbits = nbits;
		if(nbytes == 8)
			type = T_FLOAT8;
	} else {
		if(icode[1] == 0)
			type = 0;
		else
			type = IS_PACK;
		nbits = icode[0] > 0 ? (icode[0] * 8) : -icode[0];
		nbytes = (nbits + 7) / 8;
		lgbits = nbits;
		if((nexp = iexp) < 0) {
			nbits--;
			nexp = -iexp;
		} else if(iexp == 0)
			nexp = 200;
		if(nexp < 100 || nexp > 300)
			imerror(9,"exposant hors limites\n");
		nexp -= 200;
		switch(nbytes) {
		case 1 :
			type |= iexp < 0 ? T_S1 : T_NS1;
			break;
		case 2 :
			type |= iexp < 0 ? T_S2 : T_NS2;
			break;
		case 4 :
			type |= iexp < 0 ? T_S4 : T_NS4;
			break;
		default:
			imerror(9,"conversion sur 3 octets non acceptee\n");
		}
	}
	return type;
}

#if 0  /* ce code n'a pas ete teste */
#define N_BITS 32
extern koctet_;

cnv_1b(in,out,ndim,type,nbyte)
register unsigned char *out;
register unsigned long *in;
register ndim, type, nbyte;
{
	register unsigned long msk;
	unsigned char *bufout;
	/* on commence par la fin */
	fprintf(stderr,"cnv_1b %d\n",koctet_);
	bufout = out;
	out += (ndim -1) * nbyte;
	in += ((ndim-1) / N_BITS);
	msk = 1 << (N_BITS - ((ndim -1) % N_BITS));
	while(1) {
		while(msk != 0x8000000) {
			switch(type) {
			case T_FLOAT:
				if(*in & msk)
					*((float *)out) = i_exp;
				else
					*((float *)out) = 0.;
				break;
			case T_S1:
				if(*in & msk)
					*((char *)out) = 0x7f;
				else
					*((char *)out) = 0;
				break;
			case T_NS1:
				if(*in & msk)
					 *out = 0xff;
				else
					*out = 0;
				break;
			case T_S2:
				if(*in & msk)
					*((short *)out) = 0x7fff;
				else
					*((short *)out) = 0;
				break;
			case T_NS2:
				if(*in & msk)
					 *((unsigned short *)out) = 0xffff;
				else
					*((unsigned short *)out) = 0;
				break;
			case T_S4:
				if(*in & msk)
					*((long *)out) = 0x7fffffff;
				else
					*((long *)out) = 0;
				break;
			case T_NS4:
				if(*in & msk)
					 *((unsigned long *)out) = 0xffffffff;
				else
					*((unsigned long *)out) = 0;
				break;
			}
			ndim--;
			out -= nbyte;
			if(ndim <= 0) {
				if(out_pack)
					pckbt_(bufout,bufout,&ndm,&o_lgbits);
				return;
			}
			msk <<= 1;
		}
		msk = 1;
		in--;
	}
}

cnvto_1b()
{
}
#endif
