/* interface avec malloc, calloc, realloc et free */
#include <stdio.h>
#include <stdlib.h>
#include <inrimage/image.h>

int inr_nbmalloc= 0, inr_nbcalloc= 0, inr_nbrealloc= 0, inr_nbfree= 0;
int inr_totsize= 0, inr_currsize= 0;  /* nb total alloue, et nb max effectif */
int inr_maxsize=0;  /* maximum atteint par inr_currsize */
int inr_dballoc= 0;  /* mode debug : calcule les statistiques */
static int a_align = 0, a_type = 0;
#define MAGIC 0xef
static int alloc_type(), alloc_size();

extern int debug_;

/* positionee inr_dballoc et renvoie la val precedente */
int i_alloc_debug(idb)
register idb;
{
	register i;
	i = inr_dballoc;
	if(idb)
		inr_dballoc = 1;
	else
		inr_dballoc = 0;
	return i;
}

char *i_malloc(size)
register size;
{
	register char *pc;
	if((pc = (char *)malloc(size)) == (char *)0)
		imerror(8,"i_malloc : alloc error %d\n",errno);
	if(inr_dballoc) {
		inr_nbmalloc++;
		inr_totsize += size;
		inr_currsize += alloc_size(pc);  /* effective size */
		if(inr_maxsize < inr_currsize)
			inr_maxsize = inr_currsize;
		if(debug_)
			fprintf(stderr,"%d i_malloc(%d), total size %d\n",
				inr_nbmalloc,size,inr_currsize);
	} else if(debug_) 
		fprintf(stderr,"i_malloc(%d)\n",size);
	return pc;
}

char *i_calloc(nelem,nbelem)
register nelem, nbelem;
{
	register size;
	register char *pc;
	if((pc = (char *)calloc(nelem,nbelem)) == (char *)0)
		imerror(8,"i_malloc : alloc error %d\n",errno);
	if(inr_dballoc) {
		inr_nbcalloc++;
		size = nelem * nbelem;
		inr_totsize += size;
		inr_currsize += alloc_size(pc);
		if(inr_maxsize < inr_currsize)
			inr_maxsize = inr_currsize;
		if(debug_)
			fprintf(stderr,"%d i_calloc(%d), total size %d\n",
				inr_nbcalloc,size,inr_currsize);
	} else if(debug_)
		fprintf(stderr,"i_calloc(%d)%d\n",nelem * nbelem);
	return pc;
}

char *i_realloc(ptr,size)
register size;
register char *ptr;
{
	register char *pc;
	register n;
	n = alloc_size(ptr);
	if((pc = (char *)realloc(ptr,size)) == (char *)0)
		imerror(8,"i_malloc : alloc error %d\n",errno);
	if(inr_dballoc) {
		inr_nbrealloc++;
		inr_totsize += size;
		inr_currsize += alloc_size(pc) - n;
		if(inr_maxsize < inr_currsize)
			inr_maxsize = inr_currsize;
		if(debug_)
			fprintf(stderr,"%d i_realloc(%d), total size %d\n",
				inr_nbrealloc,size,inr_currsize);
	} else if(debug_)
		fprintf(stderr,"i_realloc(%d)\n",size);
	return pc;
}

i_Free(pptr)
char **pptr;
{
	if(*pptr == NULL) return 1 /* makes c99 happy */; 
	if(inr_dballoc) {
		inr_nbfree++;
		inr_currsize -= alloc_size(*pptr);
		if(debug_)
			fprintf(stderr,"%d i_Free, remaining size %d\n",
				inr_nbfree,inr_currsize);
	} else if(debug_)
		fprintf(stderr,"i_Free\n");
	free(*pptr);
	*pptr = NULL;
}

/* try to guess algorithm */
static int alloc_type(p)
register char *p;
{
	register int4 *pl;
	register unsigned char *pc;
	if(a_align != 0) return 1;  /* already checked */
	pl = (int4 *)p;
	if(*(pl-1) == 0) {
		a_align = 8;
		if(*(pl -2) == 0) {
			a_align = -1;
			return 0;
		}
	} else
		a_align = 4;
	pc = (unsigned char *)(p - a_align);
#ifdef __ksr__
	a_type = -1;
#else
	if(*pc == MAGIC) {
		a_type = 1;
	} else
		a_type = 0;
#endif
	return 1;
}

/* try to guess size of block */
static int alloc_size(p)
register unsigned char *p;
{
	register n;
	if(a_align == 0) alloc_type(p);
	if(a_align < 0) return 0;  
	p -= a_align;
	if(a_type > 0) {
		n = *(p+1);
		n = 1 << (n + 3);  /* size = 2^n+3 */
	} else {
		n = *((int4 *)p);
		if(a_type < 0)
			n = 1 << (((n & 0xfff) >>7) -1);
	}
	return n;
}


/* print debug info on stream (stderr by default) */
i_alloc_info(f)
FILE *f;
{
	if(f == (FILE *)0) f = stderr;
	if(! inr_totsize) {
		fprintf(f,"Pas de statistiques disponibles :\nmode debug non active ou fonctions i_malloc, i_calloc i_realloc non utilisees\n");
		return 1 /* makes c99 happy */; 
	}
	fprintf(f,"%5d malloc, %5d calloc, %5d realloc, %d free\n",inr_nbmalloc, inr_nbcalloc, inr_nbrealloc, inr_nbfree);
	fprintf(f,"\tnb total demande:  %d octets\n",inr_totsize);
	if(inr_currsize < inr_totsize) return 1 /* makes c99 happy */;   /* donnees incorrectes */
	fprintf(f,"\tnb total effectif: %d octets\n",inr_currsize);
	fprintf(f,"\tmaximum effectif:  %d octets\n",inr_maxsize);
	
}
