#include <inrimage/inrlocal.h>
#include <inrimage/image.h>
#include <inrimage/error.h>
#include <stdio.h>
#include <unistd.h>
#if 0 // bereziat 10/2006
#ifdef sun5
#    include <string.h>
#else
#    include <strings.h>
#endif
#else
#include <string.h>
#endif
#include <stdlib.h>

extern int  debug_;
extern char *inr_mymalloc(), *inr_myrealloc(), *inr_mycalloc();
static void inr_rdexts();
#ifdef darwin
#include <crt_externs.h>
char **environ;
#else
extern char **environ;
#endif
extern char *get_inr_extfile();

extern char inr_force_ext[];  /* pour forcer l'ext des images d'entree */

struct inr_exts {
	char *ext; 	/* extension, sans le '.' */
	short lgext;
	short iprog;
};
struct inr_magics {
	short offset;
	short nbchar;
	unsigned char tabmagic[8];
};

#define MAX_XARGS 8
struct inr_progs {
	char is_checked;
	char nb_args;  /* nb arguments par defaut */
	short nb_magic;
	char *defcmd;
	char **pargs;   /* table des arguments */
	char*cmdname;
	char *cmde;
	int imagic;  /* index dans ptab_magics */
};

struct inr_magics tab_magics0[] = {
#define IM_ZCAT 0
	{0,2,{0x1f,0x8b}},	 /* pour zcat */
	{0,2,{0x1f,0x9d}},
#define IM_GIF IM_ZCAT + 2
	{0,6,"GIF87a"},
	{0,6,"GIF89a"},
};
#if 0
struct inr_magics tiff_mag[]= {
	{0,2,"MM"},
	{0,2,"II"},
};
struct inr_magics viff_mag[]= {
	{0,2,{0xab,1}},
};
#endif
static struct inr_progs tab_progs0[]= {
#define I_ZCAT 0
#  ifdef darwin
        {0,0,2,"/usr/local/bin/zcat",(char **)0,"gzcat",(char *)0,IM_ZCAT},  /* 0 */
#  else
        {0,0,2,"/usr/local/bin/zcat",(char **)0,"zcat",(char *)0,IM_ZCAT},  /* 0 */
#  endif  
#define I_GIF (I_ZCAT + 1)
	{0,0,2,(char *)0,(char **)0,"gif2inr",(char *)0,IM_GIF},  /* 1 */
#if 0
#define I_TIFF (I_GIF + 1)
	{0,0,2,(char *)0,(char **)0,"tiff2inr",(char *)0,tiff_mag},  /* 1 */
#define I_VIFF (I_TIFF + 1)
	{0,0,1,(char *)0,(char **)0,"viff2inr",(char *)0,viff_mag},  /* 1 */
#endif
};
struct inr_exts tab_exts0[]= {
	{".Z",2,I_ZCAT},
	{".gz",3,I_ZCAT},
	{".gif",4,I_GIF},
#if 0
	{".tif",4,I_TIFF},
	{".tiff",5,I_TIFF},
	{".vif",4,I_VIFF},
	{".viff",5,I_VIFF},
#endif
};

/* nb d'elements pour chaque realloc */
#define EXT_BLOCK 8
static int nb_progs= 0;  /* nb progs courant */
static int max_progs= 0;  /* nb de slots alloues */
static int nb_exts= 0;  /* nb exts courant */
static int max_exts= 0;  /* nb de slots alloues */
static int nb_magics= 0;  /* nb magics courant */
static int max_magics= 0;  /* nb de slots alloues */
static struct inr_exts *ptab2_exts= NULL;
static struct inr_progs *ptab_progs= NULL;
static struct inr_magics *ptab_magics= NULL;

#define NB_PROGS0 (sizeof(tab_progs0) / sizeof (struct inr_progs))
#define NB_MAGICS0 (sizeof(tab_magics0) / sizeof (struct inr_magics))
#define NB_EXTS (sizeof(tab_exts0) / sizeof (struct inr_exts))

#define LGCMD 128
/* cherche le programme correspondant a pp */
static void inr_findprog(pp)
register struct inr_progs *pp;
{
	char zcatbuf[LGCMD + 2];
	register char *ps, *p2, *env;
	register n, k;
	register flag;
	if(pp->is_checked) return;
	pp->is_checked = -1;
#if 0
	if(env = pp->env_var) {
		if(ps = getenv(env)) {
			if(access(ps,X_OK) == 0) {
				pp->cmde = ps;
				pp->is_checked = 1;
				return;
			}
		}
	}
#endif
	/* privilegier pp->defcmd */
	if(ps = pp->defcmd) {
		if(access(ps,X_OK) == 0) {
			pp->cmde = ps;
			pp->is_checked = 1;
			return;
		}
	}
	/* search PATH */
	if(ps = getenv("PATH")) {
		k = LGCMD - strlen(pp->cmdname);
		while(ps) {
			if(p2 = index(ps,':')) {
				n = p2 - ps;
				p2++;
			} else n = strlen(ps);
			if((n <= 0) || (n > k)) {
				ps = p2;
				continue;  /* nom trop long ou nul */
			}
			strncpy(zcatbuf,ps,n);
			zcatbuf[n] = '/';
			strcpy(&zcatbuf[n+1],pp->cmdname);
			if(access(zcatbuf,X_OK) == 0) {
				pp->cmde = inr_mymalloc(strlen(zcatbuf) + 1,"inr_zcat.c:");
				strcpy(pp->cmde,zcatbuf);
				pp->is_checked = 1;
				return;
			}
			ps = p2;
		}
			
	}
	if(pp->is_checked < 0)
		fprintf(stderr,"Warning! unable to find %s program\n",pp->cmdname);
	
}

static int inr_tstfile(name,nump)
char *name;
register nump;
{
	register n, fd, k, i, j;
	register struct inr_progs *pp;
	register struct inr_magics *pm;
	unsigned char tmp[8];
	nump--;
	if(! ptab_progs) return 0;
	pp = ptab_progs + nump;
	if(!pp->is_checked)
		inr_findprog(pp);
	if(pp->is_checked < 0)
		return 0;
	if(pp->nb_magic <= 0) return 1;
	if((fd = open(name,0)) < 0)
		return 0;
	n = 0;
	pm = ptab_magics + pp->imagic;
	for(j=0;j<pp->nb_magic;j++,pm++) {
		lseek(fd,(off_t)pm->offset,0);
		if((k = pm->nbchar) > 8) continue;
		if(read(fd,tmp,k) == k) {
			n = 1;
			for(i=0;i<k;i++) {
				if(tmp[i] != pm->tabmagic[i]) {
					n = 0;
					break;
				}
			}
		}
		if(n) {
			close(fd);
			return n;
		}
	}
	close(fd);
	return 0;
}

static int file_compiled= 0;

/* teste si le fichier de nom name est d'un type necessitant un filtre
0 si non, indice +1 dans tab_progs si oui */
int inr_tstext(name)
CHAR *name;
{
	register n, k;
	n = strlen(name) - 1;
	if(!file_compiled)
		inr_rdexts();
	if((k = inr_xtstext(name,n,ptab2_exts,nb_exts)) == 0)
		return inr_xtstext(name,n,tab_exts0,NB_EXTS);
	return k;
}

/* teste si le fichier de nom name est d'un type necessitant un filtre
0 si non, indice +1 dans tab_progs si oui */
int inr_xtstext(name,n,ppe,nb)
CHAR *name;
register struct inr_exts *ppe;
register n, nb;
{
	register i, k, nump;
	register CHAR *p, *p0;
	register char *pe;
	
	if(inr_force_ext[0]) {
		n = strlen(inr_force_ext);
		p0 = &inr_force_ext[n-1];
	} else
		p0 = name + n;
	for(i=0;i<nb;i++,ppe++) {
		if(!ppe->ext) return 0;
		if(n < (k = ppe->lgext)) continue;
#if 0
		p = p0 - k;
		if(*p++ != '.') continue;
#else
		p = p0 - k + 1;
#endif
		pe = ppe->ext;
		while(k-- > 0)
			if(*pe++ != *p++) break;
		if(k < 0) {
			nump = ppe->iprog + 1;
			if(inr_tstfile(name,nump))
				return nump;
			return 0;
		}

	}
	return 0;
}


/* lance programme et renvoie le file descriptor du pipe
Ne pas oublier d'appeler wait quand la lecture est finie
*/
int inr_startsubproc(name,num)
register CHAR *name;
int num;
{
	int pp[2];
	char *tabarg[MAX_XARGS + 3];
	register narg, i;
	register struct inr_progs *p;
	register pid;
	if(--num >= nb_progs) 
		imerror(9,"bad call to inr_startsubproc\n");
	p = ptab_progs + num;
	if(p->cmde == NULL)
		inr_findprog(p);
	if(pipe(pp) < 0)
		imerror(9,"image_ : pipe err %d\n",errno);
	if(debug_) {
		if(narg = p->nb_args) {
			fprintf(stderr,"%s: Starting %s",xargv[0],p->cmde);
			for(i=0;i<narg;i++)
				fprintf(stderr," %s",p->pargs[i]);
			fprintf(stderr," %s\n",name);
		} else
			fprintf(stderr,"%s: Starting %s %s\n",xargv[0],p->cmde,name);
	}
	if((pid = fork()) < 0)
		imerror("image_: fork err %d\n",errno);
	if(pid == 0) {  /* fils : exec zcat */
		close(pp[0]);
		if(pp[1] != 1) {
			close(1);
			dup2(pp[1],1);
			close(pp[1]);
		}
/*		execl(p->cmde,p->cmdname,name,0); */
		narg = 1;
		tabarg[0] = p->cmdname;
		if(p->nb_args) {
			for(i=0;i<p->nb_args;i++)
				tabarg[narg++] = p->pargs[i];
		}
		tabarg[narg++] = (char *)name;
		tabarg[narg++] = (char *)0;
#if 0
		execve(p->cmde,tabarg,(char **)0);
#else
#ifdef darwin
		environ = *_NSGetEnviron();
#endif
		execve(p->cmde,tabarg,environ);
#endif
		imerror(9,"image_ : execve err %d\n",errno);
		return 0;
	}
	close(pp[1]);
	return pp[0];
}

#define MAXTOK 5
struct tok {
	char *p1;
	char *p2;
} tabtoks[4];
#define MAX_MAGIC 8
struct inr_magics magtb[MAX_MAGIC];

static int mktokens(p0,pend,tab,n,sep)
register char *p0, *pend;
register struct tok *tab;
register char sep;
register n;
{
	register i, k, j;
	register char *p, *p2;
	p = p0;
	j = 0;
	for(i=0;i<n;i++) {
		while(isspace(*p))
			p++;
		(tab + j)->p1 = p;
		if((p2 = index(p,sep)) == NULL) {
			(tab+j)->p2 = pend;
			j++;
			break;
		}
		if(p == p2) {
			p++;
			*p2 = 0;
		} else {
			p = p2-- + 1;
			while(isspace(*p2))
				p2--;
			*(p2+1) = 0;
		}
		(tab+j)->p2 = p2;
		j++;
	}
	return j;
}

/* lecture du fichier de definition des filtres 
*/
static void inr_rdexts()
{
	register char *ps, *p1;
	char *fname;
	register struct inr_exts *pe;
	register struct inr_progs *pp;
	register struct inr_magics *pm;
	int var;
	int nbext, nbprogs, nbmag, lgc;
	FILE *f;
	register n, i, j;
	register n1, n2, n0;
	char tmp[80];
	struct tok exttoks[MAXTOK], magtoks[MAXTOK], magtok2[MAXTOK];
	struct tok argstoks[MAX_XARGS];

	file_compiled = 1;
	if(!(fname = get_inr_extfile())) return;
	if((f = fopen(fname,"r")) == NULL) return;
	n = NB_PROGS0 > EXT_BLOCK ? NB_PROGS0 : EXT_BLOCK;
	ptab_progs = (struct inr_progs *)inr_mycalloc(sizeof(struct inr_progs),n,"inr_zcat.c :");
	max_progs = n;
	nb_progs = NB_PROGS0;
	pp = ptab_progs + NB_PROGS0;
	memcpy(ptab_progs,tab_progs0,sizeof(tab_progs0));
	n = NB_MAGICS0 > EXT_BLOCK ? NB_MAGICS0 : EXT_BLOCK;
	ptab_magics = (struct inr_magics *)inr_mycalloc(sizeof(struct inr_magics),n,"inr_zcat.c :");
	max_magics = n;
	nb_magics = NB_MAGICS0;
	pm = ptab_magics + NB_MAGICS0;
	memcpy(ptab_magics,tab_magics0,sizeof(tab_magics0));
	pe = ptab2_exts;
	while(fgets(tmp,80,f) != NULL) {
		if(tmp[0] == '#') {
			if(strcmp(tmp,"#EOF#\n") == 0) break;  /* end of file */
			continue;
		}
		n = strlen(tmp) - 1;
		tmp[n--] = 0;
		n = mktokens(tmp,tmp + n,tabtoks,4,':');
		if(n < 3) {
			continue;
		}
		for(i=n;i<4;i++)
			tabtoks[i].p1 = NULL;
		lgc = tabtoks[2].p2 - tabtoks[2].p1 + 1;
		if(lgc <= 0) continue;
		/* chercher si 1 arg (espace) */
		if(ps = index(tabtoks[2].p1,' ')) {
			*ps++ = 0;
			while(isspace(*ps))
				ps++;
		}
		if(! pe) { /* alloc initiaux */
			pe = ptab2_exts = (struct inr_exts *)inr_mycalloc(sizeof(struct inr_exts),EXT_BLOCK,"inr_zcat.c :");
			max_exts = EXT_BLOCK;
		} else if(nb_progs >= max_progs) { /* realloc progs */
			max_progs += EXT_BLOCK;
			ptab_progs = (struct inr_progs *)inr_myrealloc(ptab_progs,max_progs * sizeof(struct inr_progs),"inr_zcat.c:");
			pp = ptab_progs + nb_progs;
		}
		/* separer les extensions */
		n1 = mktokens(tabtoks[0].p1,tabtoks[0].p2,exttoks,MAXTOK,',');
		if(n1 > (max_exts - nb_exts)) { /* realloc exts */
			while(n1 > (max_exts - nb_exts)) max_exts += EXT_BLOCK;
			ptab2_exts = (struct inr_exts *)inr_myrealloc(ptab2_exts,max_exts * sizeof(struct inr_exts),"inr_zcat.c:");
			pe = ptab2_exts + nb_exts;
		}
		nb_exts += n1;
		n2 = mktokens(tabtoks[1].p1,tabtoks[1].p2,magtoks,MAXTOK,',');
		if(n2 > (max_magics - nb_magics)) { /* realloc magics */
			while(n2 > (max_magics - nb_magics)) max_magics += EXT_BLOCK;
			ptab_magics = (struct inr_magics *)inr_myrealloc(ptab_magics,max_magics * sizeof(struct inr_magics),"inr_zcat.c:");
			pm = ptab_magics +  nb_magics;
		}
		pp->nb_magic = n2;
		pp->imagic = nb_magics;
		pp->is_checked = 0;
		nb_magics += n2;
		/* fabriquer les magic numbers */
		for(i=0;i<n2;i++) {
			n0 = mktokens(magtoks[i].p1,magtoks[i].p2,magtok2,MAXTOK,' ');
			if(n0 <= 1) {
				pp->nb_magic--;
				nb_magics--;
				continue;
			}
			pm->offset = atoi(magtok2[0].p1);
			if(--n0 > 8) n0 = 8;
			pm->nbchar = n0;
			for(j=0;j<n0;j++) {
				sscanf(magtok2[j+1].p1,"%x",&var);
				pm->tabmagic[j] = var;
			}
			pm++;
		}
		/* inserer le programme */
		pp->cmdname = inr_mymalloc(lgc + 1,"inr_zcat.c:");
		memcpy(pp->cmdname,tabtoks[2].p1,lgc + 1);
		/* les arguments du programme */
		if(ps) {
			p1 = pp->cmdname + strlen(pp->cmdname) + 1;
			pp->nb_args = mktokens(p1,pp->cmdname + lgc,argstoks,MAX_XARGS,' ');
			pp->pargs = (char **)inr_mymalloc(pp->nb_args * sizeof(char *),"inr_zcat.c:");
			for(i=0;i<pp->nb_args;i++)
				pp->pargs[i] = argstoks[i].p1;
		} else
			pp->nb_args = 0;
		if(ps = tabtoks[3].p1) {
			if(*ps == 0)
				pp->defcmd = NULL;
			else {
				pp->defcmd = inr_mymalloc(tabtoks[3].p2 - tabtoks[3].p1 + 2,"inr_zcat.c:");
				strcpy(pp->defcmd, tabtoks[3].p1);
			}
		} else
			pp->defcmd = NULL;
		/* inserer les extensions */
		for(i=0;i<n1;i++,pe++) {
			ps = exttoks[i].p1;
/*!!			if(*ps == '.') ps++; */
			pe->lgext = strlen(ps);
			pe->ext = inr_mymalloc(pe->lgext + 1,"inr_zcat.c:");
			strcpy(pe->ext,ps);
			pe->iprog = nb_progs;
		}
		pp++;
		nb_progs++;
	}
	fclose(f);
	if(debug_) 
		fprintf(stderr," Extension file %s: %d filters, %d extensions\n",fname,nb_progs,nb_exts);
}
