/* **********************************************************************

	UNPKBT_(IN,OUT,NB,BSIZE)
	c_unpkbt(in,out,nb,bsize)
	decompactent le tableau de NB points
	codes sur BSIZE bits, en un tableau ou les points sont codes
	sur un nombre entier d'octets (justification a droite).
	Adresse de depart: IN, adresse d'arrivee: OUT.

	Le decompactage peut s'effectuer "in-place" (on commence par
	la fin).


	Programme appelable Fortran, passage des parametres par adresse.
*/
#include <stdio.h>
#include <inrimage/image.h>
#include <inrimage/flt_type.h>

unpkbt_(in,out,nb,bsize)
register CHAR *in,*out;
Fort_int *nb;
register Fort_int *bsize;

{
	c_unpkbt(in,out,(int)*nb,(int)*bsize);
}

c_unpkbt(in,out,nb,bsize)
register CHAR *in,*out;
int nb;
register int bsize;
{
	unsigned char tmpb[4];
	register bpix;  /* nb bytes / pixel */
	register val, v;
	register unsigned char msk;  /* high bits mask in out byte */
	register int pad;
	register obits;   /* nb of bits in high output byte */
	register ibits; /* nb bits used in input byte */
	register bits1;  /* nb of input bits going to 1st byte */
	register i, j;
	if((bpix = (bsize + 7) / 8) <= 0)
		imerror(9,"pckbt_ : bad nb of bits\n");
	pad = bsize % 8;
	if(pad)
		pad = 8 - pad;
	/* find end of buffers */
	out += (nb * bpix) -1;
	i = nb * bsize;
	in += (i +7) / 8 -1;
	val = 0;
	if((ibits = i % 8) == 0) ibits = 8;
	v = *in--;
	msk = 0xff >> pad;
	obits = 8 - pad;
	if(bpix == 1) {
		for(i=0;i < nb;i++) {
			if(ibits == 0) {
				v = *in--;
				ibits = 8;
			}
			if((bits1 = obits - ibits) > 0) { /* bits a prendre sur 2 octets */
				val = v >> (8 - ibits);
				v = *in--;
				*out-- = (val | (v << ibits)) & msk;
				ibits = 8 - bits1;
			} else if(bits1 == 0) {  /* complet et changer d'octet */
				*out-- = v >> pad;
				ibits = 0;
			} else {  /* 1er octet non termine */
				*out-- = (v >> (8 - ibits)) & msk;
				ibits = -bits1;
			}
		}
		return 1 /* makes c99 happy */; 
	}
	/* > 1 byte : put most significant bits first */
	if(ISMSBFIRST) {  /* the system has msb first */
		for(i=0;i < nb;i++) {
			j = bpix;
			while(j-- > 0) {
				if(ibits == 0) {
					v = *in--;
					ibits = 8;
				}
				if(j == 0) {
					if((bits1 = obits - ibits) > 0) { /* bits a prendre sur 2 octets */
						val = v >> (8 - ibits);
						v = *in--;
						*out-- = (val | (v << ibits)) & msk;
						ibits = 8 - bits1;
					} else if(bits1 == 0) {  /* complet et changer d'octet */
						*out-- = v >> pad;
						ibits = 0;
					} else {  /* 1er octet non termine */
						*out-- = (v >> (8 - ibits)) & msk;
						ibits = -bits1;
					}
				} else {
					if((bits1 = 8 - ibits) > 0) { /* bits a prendre sur 2 octets */
						val = v >> (8 - ibits);
						v = *in--;
						*out-- = (val | (v << ibits)) & 0xff;
						ibits = 8 - bits1;
					} else if(bits1 == 0) {  /* complet et changer d'octet */
						*out-- = v;
						ibits = 0;
					} else {  /* 1er octet non termine */
						*out-- = (v >> (8 - ibits)) & 0xff;
						ibits = -bits1;
					}
				}
			}
		}
		return 1 /* makes c99 happy */; 
	}

	/* cpu has LSB first : swap ; we must use a temporary buffer */
	for(i=0;i < nb;i++) {
		j = bpix;
		while(j-- > 0) {
			if(ibits == 0) {
				v = *in--;
				ibits = 8;
			}
			if(j == 0) {
				if((bits1 = obits - ibits) > 0) { /* bits a prendre sur 2 octets */
					val = v >> (8 - ibits);
					v = *in--;
					tmpb[0] = (val | (v << ibits)) & msk; /* MSB */
					ibits = 8 - bits1;
				} else if(bits1 == 0) {  /* complet et changer d'octet */
					tmpb[0] = v >> pad;
					ibits = 0;
				} else {  /* 1er octet non termine */
					tmpb[0] = (v >> (8 - ibits)) & msk;
					ibits = -bits1;
				}
			} else {
				if((bits1 = 8 - ibits) > 0) { /* bits a prendre sur 2 octets */
					val = v >> (8 - ibits);
					v = *in--;
					tmpb[j] = (val | (v << ibits)) & 0xff;
					ibits = 8 - bits1;
				} else if(bits1 == 0) {  /* complet et changer d'octet */
					tmpb[j] = v;
					ibits = 0;
				} else {  /* 1er octet non termine */
					tmpb[j] = (v >> (8 - ibits)) & 0xff;
					ibits = -bits1;
				}
			}
		}
		for(j=0;j<bpix;j++)
			*out-- = tmpb[j];
	}

}
